Package: pkgstats
Title: Metrics of R Packages
Version: 0.2.4
Authors@R: 
    c(person("Mark", "Padgham", , "mark.padgham@email.com", role = c("aut", "cre"),
             comment = c(ORCID = "0000-0003-2172-5265")),
      person("Michael", "Sumner", role = c("ctb"), 
             comment = c(ORCID = "0000-0002-2471-7511")),       
      person("Jeffrey", "Hollister", role = c("ctb"), 
             comment = c(ORCID = "0000-0002-9254-9740")),
      person("Egor", "Kotov", role = c("ctb"), 
             comment = c(ORCID = "0000-0001-6690-5345")))
Description: Static code analyses for R packages using the external code-tagging
    libraries 'ctags' and 'gtags'. Static analyses enable packages to be
    analysed very quickly, generally a couple of seconds at most. The package
    also provides access to a database generating by applying the main function
    to the full 'CRAN' archive, enabling the statistical properties of any package
    to be compared with all other 'CRAN' packages.
License: GPL-3
URL: https://docs.ropensci.org/pkgstats/,
        https://github.com/ropensci-review-tools/pkgstats
BugReports: https://github.com/ropensci-review-tools/pkgstats/issues
Imports: brio, checkmate, dplyr, fs, igraph, memoise, methods, readr,
        roxygen2, rprojroot, sys, withr
Suggests: callr, curl, hms, httr2, jsonlite, knitr, parallel,
        piggyback, pkgbuild, Rcpp, rmarkdown, testthat (>= 3.0.0),
        tibble, toml
LinkingTo: cpp11
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/ropensci/maintainer: staff
Encoding: UTF-8
Language: en-US
NeedsCompilation: yes
SystemRequirements: ctags, global
Config/roxygen2/version: 8.0.0
Packaged: 2026-07-21 11:53:46 UTC; smexus
Author: Mark Padgham [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2172-5265>),
  Michael Sumner [ctb] (ORCID: <https://orcid.org/0000-0002-2471-7511>),
  Jeffrey Hollister [ctb] (ORCID:
    <https://orcid.org/0000-0002-9254-9740>),
  Egor Kotov [ctb] (ORCID: <https://orcid.org/0000-0001-6690-5345>)
Maintainer: Mark Padgham <mark.padgham@email.com>
Repository: CRAN
Date/Publication: 2026-07-21 13:50:02 UTC

# Generated by roxygen2: do not edit by hand

export(ctags_install)
export(ctags_test)
export(desc_stats)
export(dl_pkgstats_data)
export(extract_tarball)
export(loc_stats)
export(pkgstats)
export(pkgstats_cran_current_from_full)
export(pkgstats_fn_names)
export(pkgstats_fns_from_archive)
export(pkgstats_fns_update)
export(pkgstats_from_archive)
export(pkgstats_summary)
export(pkgstats_update)
export(plot_network)
export(rd_stats)
export(tags_data)
importFrom(memoise,memoise)
useDynLib(pkgstats, .registration = TRUE)

<!-- badges: start -->

[![R build
status](https://github.com/ropensci-review-tools/pkgstats/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ropensci-review-tools/pkgstats/actions?query=workflow%3AR-CMD-check.yaml)
[![codecov](https://codecov.io/gh/ropensci-review-tools/pkgstats/branch/main/graph/badge.svg)](https://app.codecov.io/gh/ropensci-review-tools/pkgstats)
[![Project Status:
Active](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/pkgstats)](https://cran.r-project.org/package=pkgstats/)
[![CRAN
Downloads](https://cranlogs.r-pkg.org/badges/grand-total/pkgstats?color=orange)](https://cran.r-project.org/package=pkgstats)
<!-- badges: end -->

# pkgstats

Extract summary statistics of R package structure and functionality. Not
all statistics of course, but a good go at balancing insightful
statistics while ensuring computational feasibility. `pkgstats` is a
*static* code analysis tool, so is generally very fast (a few seconds at
most for very large packages). Installation is described in [a separate
vignette](https://docs.ropensci.org/pkgstats/articles/installation.html).

## What statistics?

Statistics are derived from these primary sources:

1.  Numbers of lines of code, documentation, and white space (both
    between and within lines) in each directory and language
2.  Summaries of package `DESCRIPTION` file and related package
    meta-statistics
3.  Summaries of all objects created via package code across multiple
    languages and all directories containing source code (`./R`,
    `./src`, and `./inst/include`).
4.  A function call network derived from function definitions obtained
    from [the code tagging library, `ctags`](https://ctags.io), and
    references (“calls”) to those obtained from [another tagging
    library, `gtags`](https://www.gnu.org/software/global/). This
    network roughly connects every object making a call (as `from`) with
    every object being called (`to`).
5.  An additional function call network connecting calls within R
    functions to all functions from other R packages.

The [primary function,
`pkgstats()`](https://docs.ropensci.org/pkgstats/reference/pkgstats.html),
returns a list of these various components, including full `data.frame`
objects for the final three components described above. The statistical
properties of this list can be aggregated by the [`pkgstats_summary()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats_summary.html),
which returns a `data.frame` with a single row of summary statistics.
This function is demonstrated below, including full details of all
statistics extracted.

## Demonstration

The following code demonstrates the output of the main function,
`pkgstats`, using an internally bundled `.tar.gz` “tarball” of this
package. The `system.time` call demonstrates that the static code
analyses of `pkgstats` are generally very fast.

``` r
library (pkgstats)
tarball <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
system.time (
    p <- pkgstats (tarball)
)
```

    ##    user  system elapsed 
    ##   1.701   0.124   1.802

``` r
names (p)
```

    ## [1] "loc"            "vignettes"      "data_stats"     "desc"          
    ## [5] "translations"   "objects"        "network"        "external_calls"

The result is a list of various data extracted from the code. All except
for `objects` and `network` represent summary data:

``` r
p [!names (p) %in% c ("objects", "network", "external_calls")]
```

    ## $loc
    ## # A tibble: 4 × 12
    ##   langage dir        nfiles nlines ncode  ndoc nempty nspaces nchars nexpr ntabs
    ##   <chr>   <chr>       <int>  <int> <int> <int>  <int>   <int>  <int> <dbl> <int>
    ## 1 C++     src             3    365   277    21     67     933   7002     1     0
    ## 2 R       R              19   3741  2698   536    507   27575  94022     1     0
    ## 3 R       tests           2    146   121     1     24     395   2423     1     0
    ## 4 R       tests/tes…      5    202   145     9     48     375   3738     1     0
    ## # ℹ 1 more variable: indentation <int>
    ## 
    ## $vignettes
    ## vignettes     demos 
    ##         0         0 
    ## 
    ## $data_stats
    ##           n  total_size median_size 
    ##           0           0           0 
    ## 
    ## $desc
    ##    package version                date license
    ## 1 pkgstats     9.9 2022-05-12 11:41:22   GPL-3
    ##                                                                                      urls
    ## 1 https://docs.ropensci.org/pkgstats/,\nhttps://github.com/ropensci-review-tools/pkgstats
    ##                                                       bugs aut ctb fnd rev ths
    ## 1 https://github.com/ropensci-review-tools/pkgstats/issues   1   0   0   0   0
    ##   trl depends                                                        imports
    ## 1   0      NA brio, checkmate, dplyr, fs, igraph, methods, readr, sys, withr
    ##                                                                         suggests
    ## 1 hms, knitr, pbapply, pkgbuild, Rcpp, rmarkdown, roxygen2, testthat, visNetwork
    ##   enhances linking_to
    ## 1       NA      cpp11
    ## 
    ## $translations
    ## [1] NA

The various components of these results are described in further detail
in the [main package
vignette](https://docs.ropensci.org/pkgstats/articles/pkgstats.html).

### Overview of statistics and the `pkgstats_summary()` function

A summary of the `pkgstats` data can be obtained by submitting the
object returned from `pkgstats()` to the [`pkgstats_summary()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats_summary.html):

``` r
s <- pkgstats_summary (p)
```

This function reduces the result of the [`pkgstats()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats_summary.html)
to a single line with 95 entries, represented as a `data.frame` with one
row and that number of columns. This format is intended to enable
summary statistics from multiple packages to be aggregated by simply
binding rows together. While 95 statistics might seem like a lot, the
[`pkgstats_summary()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats_summary.html)
aims to return as many usable raw statistics as possible in order to
flexibly allow higher-level statistics to be derived through combination
and aggregation. These 95 statistics can be roughly grouped into the
following categories (not shown in the order in which they actually
appear), with variable names in parentheses after each description. Some
statistics are summarised as comma-delimited character strings, such as
translations into human languages, or other packages listed under
“depends”, “imports”, or “suggests”. This enables subsequent analyses of
their contents, for example of actual translated languages, or both
aggregate numbers and individual details of all package dependencies, as
demonstrated immediately below.

**Package Summaries**

- name (`package`)
- Package version (`version`)
- Package date, as modification time of `DESCRIPTION` file where not
  explicitly stated (`date`)
- License (`license`)
- Languages, as a single comma-separated character value (`languages`),
  and excluding `R` itself.
- List of translations where package includes translations files, given
  as list of (spoken) language codes (`translations`).

**Information from `DESCRIPTION` file**

- Package URL(s) (`url`)
- URL for BugReports (`bugs`)
- Number of contributors with role of *author* (`desc_n_aut`),
  *contributor* (`desc_n_ctb`), *funder* (`desc_n_fnd`), *reviewer*
  (`desc_n_rev`), *thesis advisor* (`ths`), and *translator* (`trl`,
  relating to translation between computer and not spoken languages).
- Comma-separated character entries for all `depends`, `imports`,
  `suggests`, and `linking_to` packages.

Numbers of entries in each the of the last two kinds of items can be
obtained from by a simple `strsplit` call, like this:

``` r
deps <- strsplit (s$suggests, ", ") [[1]]
length (deps)
print (deps)
```

    ## [1] 9

``` r
print (deps)
```

    ## [1] "hms"        "knitr"      "pbapply"    "pkgbuild"   "Rcpp"      
    ## [6] "rmarkdown"  "roxygen2"   "testthat"   "visNetwork"

**Numbers of files and associated data**

- Number of vignettes (`num_vignettes`)
- Number of demos (`num_demos`)
- Number of data files (`num_data_files`)
- Total size of all package data (`data_size_total`)
- Median size of package data files (`data_size_median`)
- Numbers of files in main sub-directories (`files_R`, `files_src`,
  `files_inst`, `files_vignettes`, `files_tests`), where numbers are
  recursively counted in all sub-directories, and where `inst` only
  counts files in the `inst/include` sub-directory.

**Statistics on lines of code**

- Total lines of code in each sub-directory (`loc_R`, `loc_src`,
  `loc_ins`, `loc_vignettes`, `loc_tests`).
- Total numbers of blank lines in each sub-directory (`blank_lines_R`,
  `blank_lines_src`, `blank_lines_inst`, `blank_lines_vignette`,
  `blank_lines_tests`).
- Total numbers of comment lines in each sub-directory
  (`comment_lines_R`, `comment_lines_src`, `comment_lines_inst`,
  `comment_lines_vignettes`, `comment_lines_tests`).
- Measures of relative white space in each sub-directory (`rel_space_R`,
  `rel_space_src`, `rel_space_inst`, `rel_space_vignettes`,
  `rel_space_tests`), as well as an overall measure for the `R/`,
  `src/`, and `inst/` directories (`rel_space`).
- The number of spaces used to indent code (`indentation`), with values
  of -1 indicating indentation with tab characters.
- The median number of nested expression per line of code, counting only
  those lines which have any expressions (`nexpr`).

**Statistics on individual objects (including functions)**

These statistics all refer to “functions”, but actually represent more
general “objects,” such as global variables or class definitions
(generally from languages other than R), as detailed below.

- Numbers of functions in R (`n_fns_r`)
- Numbers of exported and non-exported R functions (`n_fns_r_exported`,
  `n_fns_r_not_exported`)
- Number of functions (or objects) in other computer languages
  (`n_fns_src`), including functions in both `src` and `inst/include`
  directories.
- Number of functions (or objects) per individual file in R and in all
  other (`src`) directories (`n_fns_per_file_r`, `n_fns_per_file_src`).
- Median and mean numbers of parameters per exported R function
  (`npars_exported_mn`, `npars_exported_md`).
- Mean and median lines of code per function in R and other languages,
  including distinction between exported and non-exported R functions
  (`loc_per_fn_r_mn`, `loc_per_fn_r_md`, `loc_per_fn_r_exp_m`,
  `loc_per_fn_r_exp_md`, `loc_per_fn_r_not_exp_mn`,
  `loc_per_fn_r_not_exp_m`, `loc_per_fn_src_mn`, `loc_per_fn_src_md`).
- Equivalent mean and median numbers of documentation lines per function
  (`doclines_per_fn_exp_mn`, `doclines_per_fn_exp_md`,
  `doclines_per_fn_not_exp_m`, `doclines_per_fn_not_exp_md`,
  `docchars_per_par_exp_mn`, `docchars_per_par_exp_m`).

**Network Statistics**

The full structure of the `network` table is described below, with
summary statistics including:

- Number of edges, including distinction between languages (`n_edges`,
  `n_edges_r`, `n_edges_src`).
- Number of distinct clusters in package network (`n_clusters`).
- Mean and median centrality of all network edges, calculated from both
  directed and undirected representations of network
  (`centrality_dir_mn`, `centrality_dir_md`, `centrality_undir_mn`,
  `centrality_undir_md`).
- Equivalent centrality values excluding edges with centrality of zero
  (`centrality_dir_mn_no0`, `centrality_dir_md_no0`,
  `centrality_undir_mn_no0`, `centrality_undir_md_no`).
- Numbers of terminal edges (`num_terminal_edges_dir`,
  `num_terminal_edges_undir`).
- Summary statistics on node degree (`node_degree_mn`, `node_degree_md`,
  `node_degree_max`)

**External Call Statistics**

The final column in the result of [the `pkgstats_summary()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats_summary.html)
summarises the `external_calls` object detailing all calls make to
external packages (including to base and recommended packages). This
summary is also represented as a single character string. Each package
lists total numbers of function calls, and total numbers of unique
function calls. Data for each package are separated by a comma, while
data within each package are separated by a colon.

``` r
s$external_calls
```

    ## [1] "base:447:78,brio:7:1,dplyr:7:4,fs:4:2,graphics:10:2,hms:1:1,igraph:3:3,pbapply:1:1,pkgstats:99:60,readr:8:5,stats:16:2,sys:13:1,tools:2:2,utils:10:7,visNetwork:3:2,withr:5:1"

This structure allows numbers of calls to all packages to be readily
extracted with code like the following:

``` r
calls <- do.call (
    rbind,
    strsplit (strsplit (s$external_call, ",") [[1]], ":")
)
calls <- data.frame (
    package = calls [, 1],
    n_total = as.integer (calls [, 2]),
    n_unique = as.integer (calls [, 3])
)
print (calls)
```

    ##       package n_total n_unique
    ## 1        base     447       78
    ## 2        brio       7        1
    ## 3       dplyr       7        4
    ## 4          fs       4        2
    ## 5    graphics      10        2
    ## 6         hms       1        1
    ## 7      igraph       3        3
    ## 8     pbapply       1        1
    ## 9    pkgstats      99       60
    ## 10      readr       8        5
    ## 11      stats      16        2
    ## 12        sys      13        1
    ## 13      tools       2        2
    ## 14      utils      10        7
    ## 15 visNetwork       3        2
    ## 16      withr       5        1

The two numeric columns respectively show the total number of calls made
to each package, and the total number of unique functions used within
those packages. These results provide detailed information on numbers of
calls made to, and functions used from, other R packages, including base
and recommended packages.

Finally, the summary statistics conclude with two further statistics of
`afferent_pkg` and `efferent_pkg`. These are package-internal measures
of [afferent and efferent
couplings](https://en.wikipedia.org/wiki/Software_package_metrics)
between the files of a package. The *afferent* couplings (`ca`) are
numbers of *incoming* calls to each file of a package from functions
defined elsewhere in the package, while the *efferent* couplings (`ce`)
are numbers of *outgoing* calls from each file of a package to functions
defined elsewhere in the package. These can be used to derive a measure
of “internal package instability” as the ratio of efferent to total
coupling (`ce / (ce + ca)`).

There are many other “raw” statistics returned by the main `pkgstats()`
function which are not represented in `pkgstats_summary()`. The [main
package
vignette](https://docs.ropensci.org/pkgstats/articles/pkgstats.html)
provides further detail on the full results.

The following sub-sections provide further detail on the `objects`,
`network`, and `external_call` items, which could be used to extract
additional statistics beyond those described here.

## Code of Conduct

Please note that this package is released with a [Contributor Code of
Conduct](https://ropensci.org/code-of-conduct/). By contributing to this
project, you agree to abide by its terms.

## Contributors


<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->
<!-- prettier-ignore-start -->
<!-- markdownlint-disable -->

All contributions to this project are gratefully acknowledged using the [`allcontributors` package](https://github.com/ropensci/allcontributors) following the [allcontributors](https://allcontributors.org) specification. Contributions of any kind are welcome!

### Code

<table>

<tr>
<td align="center">
<a href="https://github.com/mpadge">
<img src="https://avatars.githubusercontent.com/u/6697851?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/commits?author=mpadge">mpadge</a>
</td>
<td align="center">
<a href="https://github.com/jhollist">
<img src="https://avatars.githubusercontent.com/u/5438539?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/commits?author=jhollist">jhollist</a>
</td>
<td align="center">
<a href="https://github.com/jeroen">
<img src="https://avatars.githubusercontent.com/u/216319?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/commits?author=jeroen">jeroen</a>
</td>
<td align="center">
<a href="https://github.com/e-kotov">
<img src="https://avatars.githubusercontent.com/u/8681379?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/commits?author=e-kotov">e-kotov</a>
</td>
<td align="center">
<a href="https://github.com/mcol">
<img src="https://avatars.githubusercontent.com/u/6078177?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/commits?author=mcol">mcol</a>
</td>
<td align="center">
<a href="https://github.com/maelle">
<img src="https://avatars.githubusercontent.com/u/8360597?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/commits?author=maelle">maelle</a>
</td>
<td align="center">
<a href="https://github.com/Bisaloo">
<img src="https://avatars.githubusercontent.com/u/10783929?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/commits?author=Bisaloo">Bisaloo</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/thomaszwagerman">
<img src="https://avatars.githubusercontent.com/u/36264706?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/commits?author=thomaszwagerman">thomaszwagerman</a>
</td>
</tr>

</table>


### Issue Authors

<table>

<tr>
<td align="center">
<a href="https://github.com/helske">
<img src="https://avatars.githubusercontent.com/u/1560448?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+author%3Ahelske">helske</a>
</td>
<td align="center">
<a href="https://github.com/rpodcast">
<img src="https://avatars.githubusercontent.com/u/1043111?u=bb3a363381b39a5172b817ae513c6d10ab1d239a&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+author%3Arpodcast">rpodcast</a>
</td>
<td align="center">
<a href="https://github.com/assignUser">
<img src="https://avatars.githubusercontent.com/u/16141871?u=b8095df6a10813031922a72335bd6579d5494c16&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+author%3AassignUser">assignUser</a>
</td>
<td align="center">
<a href="https://github.com/GFabien">
<img src="https://avatars.githubusercontent.com/u/39873986?u=0c2a28f666efccb0c0b9b23e8584749ab41da789&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+author%3AGFabien">GFabien</a>
</td>
<td align="center">
<a href="https://github.com/pawelru">
<img src="https://avatars.githubusercontent.com/u/12943682?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+author%3Apawelru">pawelru</a>
</td>
<td align="center">
<a href="https://github.com/stitam">
<img src="https://avatars.githubusercontent.com/u/49147718?u=c8736db475d31efad6ebd07b7bc76e4d7241d884&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+author%3Astitam">stitam</a>
</td>
<td align="center">
<a href="https://github.com/willgearty">
<img src="https://avatars.githubusercontent.com/u/7232514?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+author%3Awillgearty">willgearty</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/etiennebacher">
<img src="https://avatars.githubusercontent.com/u/52219252?u=66331618d799d2d4567ecab2812236c9928be368&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+author%3Aetiennebacher">etiennebacher</a>
</td>
</tr>

</table>


### Issue Contributors

<table>

<tr>
<td align="center">
<a href="https://github.com/krlmlr">
<img src="https://avatars.githubusercontent.com/u/1741643?u=caaf26641c159b84fe1b6d506f57fcea302a556c&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+commenter%3Akrlmlr">krlmlr</a>
</td>
<td align="center">
<a href="https://github.com/noamross">
<img src="https://avatars.githubusercontent.com/u/571752?u=49b086850e1716aa25615cea39250c51e085a5d8&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+commenter%3Anoamross">noamross</a>
</td>
<td align="center">
<a href="https://github.com/schneiderpy">
<img src="https://avatars.githubusercontent.com/u/77991319?u=4242d4c5942fced6368dd5c68221e6618092cbf8&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+commenter%3Aschneiderpy">schneiderpy</a>
</td>
<td align="center">
<a href="https://github.com/ScottClaessens">
<img src="https://avatars.githubusercontent.com/u/26170108?u=af7101b58069fc6dae648825e6aeb1bd4be561e0&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+commenter%3AScottClaessens">ScottClaessens</a>
</td>
<td align="center">
<a href="https://github.com/mdsumner">
<img src="https://avatars.githubusercontent.com/u/4107631?u=77e928f4bb904a5c2e8927a02194b86662408329&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+commenter%3Amdsumner">mdsumner</a>
</td>
<td align="center">
<a href="https://github.com/kellijohnson-NOAA">
<img src="https://avatars.githubusercontent.com/u/4108564?u=503d9aecc5fadf069c75e493e5abf72c7537b06f&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/pkgstats/issues?q=is%3Aissue+commenter%3Akellijohnson-NOAA">kellijohnson-NOAA</a>
</td>
</tr>

</table>

<!-- markdownlint-enable -->
<!-- prettier-ignore-end -->
<!-- ALL-CONTRIBUTORS-LIST:END -->

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set (
    collapse = TRUE,
    warning = TRUE,
    message = TRUE,
    width = 120,
    comment = "#>",
    fig.retina = 2,
    fig.path = "README-"
)
options (repos = c (
    ropenscireviewtools = "https://ropensci-review-tools.r-universe.dev",
    CRAN = "https://cloud.r-project.org"
))

## ----install, eval = FALSE----------------------------------------------------
# install.packages ("pkgstats")

## ----options, eval = FALSE----------------------------------------------------
# options (repos = c (
#     ropenscireviewtools = "https://ropensci-review-tools.r-universe.dev",
#     CRAN = "https://cloud.r-project.org"
# ))

## ----remotes, eval = FALSE----------------------------------------------------
# remotes::install_github ("ropensci-review-tools/pkgstats")
# pak::pkg_install ("ropensci-review-tools/pkgstats")

## ----remotes-elsewhere, eval = FALSE------------------------------------------
# remotes::install_git ("https://codeberg.org/ropensci-review-tools/pkgstats")
# remotes::install_git ("https://codefloe.com/ropensci-review-tools/pkgstats")

## ----library, eval = TRUE-----------------------------------------------------
library (pkgstats)

## ----ctags-check, eval = FALSE------------------------------------------------
# ctags_test ()
# #> [1] TRUE

## ----unsetenv, eval = FALSE---------------------------------------------------
# Sys.unsetenv (c ("GTAGSCONF", "GTAGSLABEL"))

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set (
    collapse = TRUE,
    warning = TRUE,
    message = TRUE,
    width = 120,
    comment = "#>",
    fig.retina = 2,
    fig.path = "README-"
)
options (repos = c (
    ropenscireviewtools = "https://ropensci-review-tools.r-universe.dev",
    CRAN = "https://cloud.r-project.org"
))
library (pkgstats)

## ----main-pkgstats-call, eval = FALSE-----------------------------------------
# s <- pkgstats () # run in root directory of `pkgstats` source

## ----pkgstats-fakey, echo = FALSE---------------------------------------------
# These data all have to be faked because they can't be generated on CRAN
# windows machines.
loc <- tibble::tibble (
    language = c ("C++", "R", "R", "Rmd"),
    dir = c ("src", "R", "tests", "vignettes"),
    nfiles = c (3L, 24L, 7L, 2L),
    nlines = c (364L, 4727L, 300L, 347L),
    ncode = c (276L, 345L, 234L, 278L),
    nempty = c (67L, 682L, 61L, 61L),
    nspaces = c (932L, 333L, 511L, 1483L),
    nchars = c (6983L, 114334L, 5543L, 11290L),
    nexpr = c (1L, 1L, 1L, 1L),
    ntabs = c (0L, 0L, 0L, 0L),
    indentation = c (4L, 4L, 4L, 4L)
)
vignettes <- c (vignettes = 2L, demos = 0L)
data_stats <- c (n = 0L, total_size = 0L, median_size = 0L)
desc <- data.frame (
    package = "pkgstats",
    verion = "0.1.1",
    date = date (),
    license = "GPL-3",
    urls = paste0 (c (
        "https://docs.ropensci.org/pkgstats/",
        "nhttps://github.com/ropensci-review-tools/pkgstats"
    ), collapse = ",\n"),
    bugs = "https://github.com/ropensci-review-tools/pkgstats/issues",
    aut = 1L,
    ctb = 0L,
    fnd = 0L,
    rev = 0L,
    ths = 0L,
    trl = 0L,
    depends = NA,
    imports = paste0 (c (
        "brio",
        "checkmate",
        "dplyr",
        "fs",
        "igraph",
        "methods",
        "readr",
        "sys",
        "withr"
    ), collapse = ", "),
    suggests = paste0 (c (
        "curl",
        "hms",
        "jsonlite",
        "knitr",
        "parallel",
        "pkgbuild",
        "Rcpp",
        "rmarkdown",
        "roxygen2",
        "testthat",
        "visNetwork"
    ), collapse = ", "),
    enchances = NA_character_,
    linking_to = "cpp11"
)
translations <- NA_character_

s <- list (
    loc = loc,
    vignettes = vignettes,
    data_stats = data_stats,
    desc = desc,
    translations = translations
)

## -----------------------------------------------------------------------------
s [!names (s) %in% c ("objects", "network", "external_calls")]

## ----loc----------------------------------------------------------------------
s$loc

## ----nested1, eval = FALSE----------------------------------------------------
# x <- myfn ()

## ----nested4, eval = FALSE----------------------------------------------------
# x <- function () { return (myfn ()) }

## ----rel-space----------------------------------------------------------------
index <- which (s$loc$dir %in% c ("R", "src")) # consider source code only
sum (s$loc$nspaces [index]) / sum (s$loc$nchars [index])

## ----desc---------------------------------------------------------------------
s$desc

## ----objects-fakey, echo = FALSE----------------------------------------------
s$objects <- data.frame (
    file_name = c (
        rep ("R/archive-trawl.R", 4L),
        "R/cpp11.R",
        "R/ctags-install.R"
    ),
    fn_name = c (
        "pkgstats_from_archive",
        "list_archive_files",
        "rm_prev_files",
        "pkgstats_fns_from_archive",
        "cpp_loc",
        "clone_ctag"
    ),
    kind = rep ("function", 6L),
    language = rep ("R", 6L),
    loc = c (89, 17, 24, 82, 3, 17),
    npars = c (9, 2, 2, 7, 4, 1),
    has_dots = rep (FALSE, 6L),
    exported = rep (c (TRUE, FALSE, FALSE), 2L),
    param_nchards_md = c (
        133,
        rep (NA_integer_, 2L),
        163,
        rep (NA_integer_, 2L)
    ),
    param_nchards_mn = c (
        159.7778,
        rep (NA_integer_, 2L),
        174.5714,
        rep (NA_integer_, 2L)
    ),
    num_doclines = c (
        77,
        rep (NA_integer_, 2L),
        50,
        rep (NA_integer_, 2L)
    )
)

## ----objects------------------------------------------------------------------
head (s$objects)

## ----network-fakey, echo = FALSE----------------------------------------------
network <- data.frame (
    file = c (
        rep ("R/external_calls.R", 4L),
        rep ("R/pkgstats-summary.R", 2L)
    ),
    line1 = c (11L, 26L, 38L, 326L, 39L, 50L),
    from = c (
        rep ("external_call_network", 3L),
        "add_other_pkgs_to_calls",
        rep ("pkgstats_summary", 2L)
    ),
    to = c (
        "extract_call_content",
        "add_base_recommended_pkgs",
        "add_other_pkgs_to_calls",
        "control_parse",
        "null_stats",
        "loc_summary"
    ),
    language = rep ("R", 6L),
    cluster_dir = rep (1L, 6L),
    centrality_dir = c (9L, 9L, 9L, 1L, 11L, 11L),
    cluster_undir = rep (1L, 6L),
    centrality_undir = c (rep (230.8333, 3L), 6, rep (874, 2L))
)

nrows <- 142 # full number in result
# expand to nrows:
n <- ceiling (nrows / nrow (network))
network <- with (network, data.frame (
    file = rep (file, n),
    line1 = rep (line1, n),
    from = rep (from, n),
    to = rep (to, n),
    language = rep (language, n),
    cluster_dir = rep (cluster_dir, n),
    centrality_dir = rep (centrality_dir, n),
    cluster_undir = rep (cluster_undir, n),
    centrality_undir = rep (centrality_undir, n)
))
s$network <- network [seq (nrows), ]

## -----------------------------------------------------------------------------
head (s$network)
nrow (s$network)

## ----external_calls-fakey, echo = FALSE---------------------------------------
s$external_calls <- data.frame (
    tags_line = 1:6,
    call = c (
        "c",
        rep ("character", 2L),
        "logical",
        "integer",
        "left_join"
    ),
    tag = c (
        "GTAGSLABEL",
        "file_name",
        "fn_name",
        "has_dots",
        "loc",
        "name"
    ),
    file = c (
        "R/ctags-test.R",
        rep ("R/pkgstats.R", 4L),
        "R/plot.R"
    ),
    kind = rep ("nameattr", 6L),
    start = c (109L, 185L, 186L, 189L, 187L, 89L),
    end = c (109L, 185L, 186L, 189L, 187L, 89L),
    package = c (rep ("base", 5L), "dplyr")
)

## ----ext-call-head------------------------------------------------------------
head (s$external_calls)

## ----ext-call-summary, eval = FALSE-------------------------------------------
# s_summ <- pkgstats_summary (s)
# print (s_summ$external_calls)

## ----ext-call-sumamry-fakey, echo = FALSE-------------------------------------
s_summ <- list (external_calls = paste0 (c (
    "base:581:84",
    "brio:11:2",
    "curl:4:3",
    "dplyr:7:4",
    "fs:4:2",
    "graphics:10:2",
    "hms:2:1",
    "igraph:3:3",
    "parallel:2:1",
    "pkgstats:126:73",
    "readr:8:5",
    "stats:19:3",
    "sys:14:1",
    "tools:3:2",
    "utils:22:7",
    "visNetwork:3:2",
    "withr:6:2"
), collapse = ","))

## ----ext-call-details, eval = TRUE--------------------------------------------
x <- strsplit (s_summ$external_calls, ",") [[1]]
x <- do.call (rbind, strsplit (x, ":"))
x <- data.frame (
    pkg = x [, 1],
    n_total = as.integer (x [, 2]),
    n_unique = as.integer (x [, 3])
)
x$n_total_rel <- round (x$n_total / sum (x$n_total), 3)
x$n_unique_rel <- round (x$n_unique / sum (x$n_unique), 3)
print (x)

#' Trawl a local CRAN archive and extract statistics from all packages
#'
#' @param path Path to local archive of R packages, either as source
#' directories, or '.tar.gz' files such as in a CRAN mirror.
#' @param archive If `TRUE`, extract statistics for all packages in the
#' `/Archive` sub-directory, otherwise only statistics for main directory (that
#' is, current packages only).
#' @param prev_results Result of previous call to this function, if available.
#' Submitting previous results will ensure that only newer packages not present
#' in previous result will be analysed, with new results simply appended to
#' previous results. This parameter can also specify a file to be read with
#' `readRDS()`.
#' @param results_file Can be used to specify the name or full path of a `.Rds`
#' file to which results should be saved once they have been generated. The
#' '.Rds' extension will be automatically appended, and any other extensions
#' will be ignored.
#' @param chunk_size Divide large archive trawl into chunks of this size, and
#' save intermediate results to local files. These intermediate files can be
#' combined to generate a single `prev_results` file, to enable jobs to be
#' stopped and re-started without having to recalculate all results. These files
#' will be named `pkgstats-results-N.Rds`, where "N" incrementally numbers each
#' file.
#' @param num_cores Number of machine cores to use in parallel, defaulting to
#' single-core processing.
#' @param save_full If `TRUE`, full \link{pkgstats} results are saved for each
#' package to files in `results_path`.
#' @param save_ex_calls If `TRUE`, the results of the `external_calls` component
#' are saved for each package to files in `results_path` (only if `save_full =
#' FALSE`).
#' @param results_path Path to save intermediate files generated by the
#' `chunk_size` parameter described above.
#'
#' @return A `data.frame` object with one row for each package containing
#' summary statistics generated from the \link{pkgstats_summary} function.
#'
#' @family archive
#' @export
#' @examples
#' # Create fake archive directory with single tarball:
#' f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
#' tarball <- basename (f)
#'
#' archive_path <- file.path (tempdir (), "archive")
#' if (!dir.exists (archive_path)) {
#'     dir.create (archive_path)
#' }
#' path <- file.path (archive_path, tarball)
#' file.copy (f, path)
#' tarball_path <- file.path (archive_path, "tarballs")
#' dir.create (tarball_path, recursive = TRUE)
#' file.copy (path, file.path (tarball_path, tarball))
#' out <- pkgstats_from_archive (tarball_path)
pkgstats_from_archive <- function (path,
                                   archive = TRUE,
                                   prev_results = NULL,
                                   results_file = NULL,
                                   chunk_size = 1000L,
                                   num_cores = 1L,
                                   save_full = FALSE,
                                   save_ex_calls = FALSE,
                                   results_path = fs::path_temp ()) {

    requireNamespace ("hms", quietly = TRUE)
    requireNamespace ("parallel", quietly = TRUE)
    requireNamespace ("callr", quietly = TRUE)

    checkmate::assert_string (path)
    checkmate::assert_directory_exists (path)
    if (!is.null (prev_results)) {
        checkmate::assert_string (prev_results)
        checkmate::assert_file_exists (prev_results)
    }
    if (!is.null (results_file)) {
        checkmate::assert_string (results_file)
    }
    checkmate::assert_int (chunk_size, lower = 1L)
    checkmate::assert_int (num_cores, lower = 1L)
    checkmate::assert_logical (archive)
    checkmate::assert_scalar (archive)
    checkmate::assert_logical (save_full)
    checkmate::assert_scalar (save_full)
    checkmate::assert_logical (save_ex_calls)
    checkmate::assert_scalar (save_ex_calls)
    checkmate::assert_string (results_path)
    checkmate::assert_directory_exists (results_path)

    if (is.character (prev_results)) {
        checkmate::assert_file_exists (prev_results)
        prev_results <- readRDS (prev_results)
    }

    if (!is.null (prev_results)) {
        if (!identical (names (prev_results), names (pkgstats_summary ()))) {
            stop (
                "'prev_results' must contain a ",
                "'data.frame' of 'pkgstats' summaries"
            )
        }
    }

    res <- NULL
    results_files <- NULL
    out <- prev_results

    flist <- list_archive_files (path, archive)
    if (all (grepl ("\\.tar\\.gz$", flist))) {
        flist <- rm_prev_files (flist, prev_results)
    }
    flist <- exclude_these_tarballs (flist)
    flist <- rm_tars_with_different_desc (flist)
    npkgs <- length (flist)

    if (npkgs > 0) {

        n <- ceiling (npkgs / chunk_size)
        n <- factor (rep (seq (n), each = chunk_size)) [seq (npkgs)]
        flist <- split (flist, f = n)

        message (
            "Starting trawl of ", npkgs,
            " files in ", length (flist), " chunks"
        )

        results_path <- expand_path (results_path)
        if (!fs::dir_exists (results_path)) {
            fs::dir_create (results_path, recurse = TRUE)
        }

        index <- 1 # name of temporary files
        pt0 <- proc.time ()

        for (f in flist) {

            fname <- fs::path (
                results_path,
                paste0 ("pkgstats-results-", index, ".Rds")
            )
            if (!fs::file_exists (fname)) {

                if (num_cores > 1L) {

                    res <- parallel::mclapply (f, function (i) {

                        one_summary_from_archive (
                            i,
                            save_full,
                            save_ex_calls,
                            results_path
                        )

                    }, mc.cores = num_cores)

                } else {

                    res <- lapply (f, function (i) {

                        one_summary_from_archive (
                            i,
                            save_full,
                            save_ex_calls,
                            results_path
                        )
                    })
                }

                saveRDS (do.call (rbind, res), fname)
            }

            results_files <- c (results_files, fname)

            archive_trawl_progress_message (index, chunk_size, npkgs, pt0)
            index <- index + 1
        }

        res <- do.call (rbind, lapply (results_files, readRDS))
    }

    out <- rbind (out, res)
    out <- out [which (!is.na (out$package)), ]
    rownames (out) <- NULL

    if (!is.null (results_files)) {
        fs::file_delete (results_files)
    }

    if (!is.null (res) && !is.null (results_file)) {

        results_file <- archive_results_file_name (results_file)
        saveRDS (out, results_file)
    }

    invisible (out)
}

list_archive_files <- function (path, recursive = FALSE) {

    tarballs <- grepl ("tarballs", path)
    if (!tarballs) {
        flist <- fs::dir_ls (path, full.names = TRUE, recurse = FALSE)
        tarballs <- any (grepl ("tarballs", flist))
        if (tarballs) {
            path <- fs::path (path, "tarballs")
        }
    }

    if (!fs::dir_exists (path)) {
        stop ("[", path, "] directory does not exist")
    }

    if (tarballs) {
        flist <- fs::dir_ls (
            path,
            recurse = recursive,
            regexp = "\\.tar\\.gz$"
        )
    } else {
        flist <- fs::dir_ls (
            path,
            recurse = FALSE
        )
        flist <- flist [which (fs::dir_exists (flist))]
        # Reduce to directories with "DESCRIPTION" files
        desc_paths <- fs::path (flist, "DESCRIPTION")
        flist <- flist [which (fs::file_exists (desc_paths))]
    }

    return (expand_path (flist))
}

#' Remove files for which results have already been generated
#' @param flist Full paths to all tarball files to be analysed
#' @param prev_results `data.frame` of previous results
#' @return Modified version of `flist`, after removing any entires present in
#' `prev_results`.
#' @noRd
rm_prev_files <- function (flist, prev_results) {

    if (!is.null (prev_results)) {

        if (is.character (prev_results)) {
            if (length (prev_results) > 1) {
                stop ("prev_results must be a single-length character")
            }
            if (!fs::file_exists (prev_results)) {
                stop ("file [", prev_results, "] does not exist")
            }
            prev_results <- tryCatch (readRDS (prev_results),
                error = function (e) e
            )
            if (methods::is (prev_results, "error")) {
                stop ("Unable to read prev_results: ", prev_results$message)
            }
            prev_results <-
                prev_results [which (!is.na (prev_results$package)), ]
        }

        tars <- basename (flist)

        prev_tars <- paste0 (
            prev_results$package,
            "_",
            prev_results$version,
            ".tar.gz"
        )

        flist <- flist [which (!tars %in% prev_tars)]
    }

    return (flist)
}

#' Some archive tarballs have version numbers in 'DESCRIPTION' files that differ
#' from the tarball numbers. This function re-maps the tarball names to match
#' the internal version numbers, to ensure these tarball file names are removed
#' as having been processed.
#' @noRd
rm_tars_with_different_desc <- function (flist) {

    # first is a grep pattern; second is direct replacement
    dat <- list (
        c ("acepack\\_1\\.1\\.tar\\.gz$", "acepack_1.0.4.tar.gz"),
        c ("bats\\_0\\.1\\-3\\.tar\\.gz$", "bats-0.1-2.tar.gz"),
        c ("HTML\\_0\\.4\\-1\\.tar\\.gz$", "HTML-0.4.tar.gz"),
        c ("survival5\\_1\\.0\\.tar\\.gz$", "survival5-1.0-0.tar.gz"),
        c ("timeslab\\_1\\.0\\-1\\.tar\\.gz$", "timeslab_1.0.tar.gz")
    )

    for (d in dat) {
        index <- grep (d [1], flist)
        if (length (index) == 1L) {
            flist <- flist [-index]
        }
    }

    return (flist)
}

#' These packages fail on main 'pkgstats' call, and can not be processed at all.
#'
#' The 'dse' and 'VR' tarballs have multiple packages in sub-directories with
#' multiple 'DESCRIPTION' files, so can not be processed as single packages, and
#' thus do not return any package names or versions.
#'
#' @noRd
exclude_these_tarballs <- function (flist) {

    exclude <- c (
        "dse_R2000.4-1.tar.gz",
        "dse_R2000.6-1.tar.gz",
        "VR_5.3pl037-1.tar.gz",
        "VR_5.3pl037-2.tar.gz",
        "VR_6.1-4.tar.gz"
    )
    index <- lapply (exclude, function (i) grep (i, flist, fixed = TRUE))
    index <- unlist (index)
    if (length (index) > 0L) {
        flist <- flist [-index]
    }

    return (flist)
}

#' Apply 'pkgstats' to one path and return the summary
#'
#' The archive trawl often stops for reasons which are not reproducible when
#' running individual calls. This therefore uses `callr` to run call processes
#' in the background, and terminate after 5 minutes with an error.
#' @noRd
one_summary_from_archive <- function (path, save_full,
                                      save_ex_calls, results_path) {

    logfiles <- list (
        stdout = fs::path (fs::path_temp (), "pkgstats-stdout"),
        stderr = fs::path (fs::path_temp (), "pkgstats-stderr")
    )
    if (fs::file_exists (logfiles$stdout)) {
        fs::file_delete (logfiles$stdout)
    }
    if (fs::file_exists (logfiles$stderr)) {
        fs::file_delete (logfiles$stderr)
    }

    ps <- callr::r_bg (
        func = pkgstats::pkgstats,
        args = list (path = path),
        stdout = logfiles$stdout,
        stderr = logfiles$stderr,
        package = TRUE
    )

    p0 <- proc.time ()
    elapsed <- proc.time () [3] - p0 [3]
    timeout <- 300
    while (ps$is_alive () && elapsed < timeout) {
        ps$wait (timeout = 10)
        elapsed <- proc.time () [3] - p0 [3]
    }
    if (elapsed < (timeout - 1)) {
        s <- tryCatch (ps$get_result (), error = function (e) NULL)
    } else {
        ps$kill ()
        s <- NULL
    }

    tryCatch (fs::file_delete (logfiles$stdout), error = function (e) NULL)
    tryCatch (fs::file_delete (logfiles$stderr), error = function (e) NULL)

    if (save_full || save_ex_calls) {
        pkg <- utils::tail (decompose_path (path) [[1]], 1L)
        pkg <- gsub ("\\.tar\\.gz$", ".Rds", pkg)
        if (save_full) {
            saveRDS (s, fs::path (results_path, pkg))
        } else if (save_ex_calls) {
            saveRDS (
                s$external_calls,
                fs::path (results_path, pkg)
            )
        }
    }

    summ <- tryCatch (
        pkgstats::pkgstats_summary (s),
        error = function (e) NULL
    )

    if (is.null (summ) || is.na (summ$package)) { # pkgstats failed
        summ <- pkgstats_summary () # null summary
        pkg_vers <- get_pkg_version (path)
        summ ["package"] <- pkg_vers [1]
        summ ["version"] <- pkg_vers [2]
    }

    return (summ)
}

archive_trawl_progress_message <- function (index, chunk_size, npkgs, pt0) {

    prog <- index * chunk_size / npkgs
    prog_fmt <- format (100 * prog, digits = 2)
    pt1 <- as.integer ((proc.time () - pt0) [3])
    t_per_file <- pt1 / (index * chunk_size)
    t_total <- t_per_file * npkgs
    t_rem <- hms::hms (t_total - pt1)

    ndone <- min (c (npkgs, index * chunk_size))

    message (
        "[", ndone, " / ", npkgs,
        "]  = ", prog_fmt, "%; (elapsed, remaining) = (",
        pt1, ", ", t_rem, ")"
    )
}

#' Check and convert 'results_file' to '.Rds' in an existing directory
#' @noRd
archive_results_file_name <- function (results_file) {

    if (!grepl (.Platform$file.sep, results_file)) {
        results_file <- fs::path (".", results_file)
    }
    results_file <- expand_path (results_file)

    results_path <- gsub (
        basename (results_file), "",
        results_file
    )
    results_path <- expand_path (results_path)
    if (!fs::dir_exists (results_path)) {
        stop ("Directory [", results_path, "] does not exist")
    }

    results_file <- basename (results_file)
    results_file <- tools::file_path_sans_ext (results_file)
    results_file <- fs::path (
        results_path,
        paste0 (results_file, ".Rds")
    )

    return (results_file)
}

#' Separate package name and version from a tarball name
#'
#' This is only used to fill in 'package' and 'version' columns of packages
#' which fail main 'pkgstats' call, to fill those columns regardless.
#'
#' @param path Full path to local tarball
#' @noRd
get_pkg_version <- function (path) {

    pkg <- basename (path)
    v_i <- regexpr ("\\_([0-9]|[[:punct:]]).*\\.tar\\.gz$", pkg)
    v <- gsub ("^\\_|\\.tar\\.gz$", "", regmatches (pkg, v_i))
    pkg <- substr (pkg, 1, v_i - 1)

    c (pkg, v)
}

#' test a 'ctags' installation
#'
#' This uses the example from
#' \url{https://github.com/universal-ctags/ctags/blob/master/man/ctags-lang-r.7.rst.in}
#' and also checks the GNU global installation.
#' @param quiet If `TRUE`, display on screen whether or not 'ctags' is correctly
#' installed.
#' @param noerror If `FALSE` (default), this function will error if either
#' 'ctags' or 'gtags' are not installed. If `TRUE`, the function will complete
#' without erroring, and issue appropriate messages regarding required but
#' non-installed system libraries.
#' @return 'TRUE' or 'FALSE' respectively indicating whether or not 'ctags' is
#' correctly installed.
#' @family tags
#' @examples
#' # The function errors if not ctags or gtags found.
#' \donttest{
#' ctags_okay <- !is.null (tryCatch (
#'     ctags_test (),
#'     error = function (e) NULL
#' ))
#' }
#' @export
ctags_test <- function (quiet = TRUE, noerror = FALSE) {

    msg <- "No ctags installation found."
    if (!has_ctags () && !ctags_test_check (msg, noerror = noerror)) {
        return (FALSE)
    }

    msg <- "No GNU global installation found."
    if (!has_gtags () && !ctags_test_check (msg, noerror = noerror)) {
        return (FALSE)
    }

    f_in <- tempfile (fileext = ".R")

    x <- c (
        "G <- 1", # globalVar
        "v <- c(1, 2)", # vector
        "l <- list(3, 4)", # list
        "d <- data.frame(n = v)", # dataframe + nameattr
        "f <- function(a) {", # function
        "    g <- function (b) a + b", # function
        "    w <- c(1, 2)", # vector
        "    m <- list (3, 4)", # list
        "    e <- data.frame(n = w)", # dataframe + nameattr
        "    L <- 2", # functionVar
        "}"
    )
    brio::write_lines (x, path = f_in)

    f_out <- fs::file_temp (ext = ".txt")
    cmd <- paste0 ("ctags --sort=no --fields=+KZ -f ", f_out, " ", f_in)
    chk <- tryCatch (
        system (cmd),
        error = function (e) 127L
    )

    if (chk > 0L || !fs::file_exists (f_out)) {
        msg <- paste0 (
            "Existing ctags installation is incomplete; ",
            "this package requires 'universal-ctags'."
        )
        if (!ctags_test_check (msg, noerror = noerror)) {
            return (FALSE)
        }
    }

    # remove header lines:
    x <- brio::read_lines (f_out)
    x <- x [-which (grepl ("^\\!", x))]
    brio::write_lines (x, path = f_out)

    ctypes <- list (
        readr::col_character (),
        readr::col_character (),
        readr::col_character (),
        readr::col_character (),
        readr::col_character (),
        readr::col_character (),
        readr::col_character ()
    )
    cnames <- c ("tag", "file", "content", "kind", "start", "language", "end")

    suppressWarnings (
        tags <- readr::read_tsv (f_out,
            col_names = cnames,
            col_types = ctypes,
            col_select = cnames,
            progress = FALSE,
            lazy = FALSE
        )
    )

    fs::file_delete (c (f_in, f_out))

    expected_kinds <- c (
        "globalVar",
        "list",
        "dataframe",
        "nameattr",
        "function",
        "function",
        "vector",
        "list",
        "dataframe",
        "nameattr",
        "functionVar"
    )

    ctags_check <- nrow (tags) == length (expected_kinds)
    if (ctags_check) {
        ctags_check <- all (tags$kind == expected_kinds)
    }

    td <- fs::path (fs::path_temp (), "pkgstats-gtags-test")

    if (fs::dir_exists (td)) {
        gtags_check <- TRUE
    } else {
        f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")

        exdir <- fs::path (fs::path_temp (), "pkgstats_ctags_test")
        fs::dir_create (exdir)
        pkgstats_path <- extract_tarball (f, exdir = exdir)
        if (fs::dir_exists (td)) {
            fs::dir_delete (td)
        }
        if (fs::file_exists (td)) {
            chk <- fs::file_delete (td)
        }
        fs::dir_create (td, recurse = TRUE)
        chk <- fs::dir_copy (pkgstats_path, td)
        # unlink (pkgstats_path, recursive = TRUE) # done in unload
        if (.Platform$OS.type == "windows") {
            cmd <- "set GTAGS_LABEL=new-ctags; gtags"
            gtags_test <- withr::with_dir (fs::path (td, "pkgstats"), shell (cmd, intern = TRUE))
        } else {
            cmd <- "export GTAGS_LABEL=new-ctags; gtags"
            gtags_test <- withr::with_dir (fs::path (td, "pkgstats"), system (cmd, intern = TRUE))
        }
        gtags_check <- length (gtags_test) == 0L
        if (!gtags_check) {
            gtags_check <- any (grepl ("error", gtags_test, ignore.case = TRUE))
        }

        fs::file_delete (td)
        fs::dir_delete (exdir)
    }

    check <- ctags_check && gtags_check

    if (!check) {
        if (!quiet) {
            message (
                "ctags does not function as required; you may need to upgrade? see\n", # nolint
                "https://github.com/universal-ctags/ctags/blob/master/man/ctags-lang-r.7.rst.in\n", # nolint
                "for expected output"
            )
        }
        ret <- FALSE
    } else {
        if (!quiet) {
            message ("ctags installation works as expected")
        }
        ret <- TRUE
    }

    return (ret)
}

ctags_test_check <- function (msg = "", noerror = FALSE) {
    if (noerror) {
        message (msg)
        return (FALSE)
    } else {
        stop (msg, call. = FALSE)
    }
}

has_ctags <- function () {

    ctags_path <- dirname (Sys.which ("ctags"))
    nzchar (ctags_path)
}

#' ctags has to be "Universal" and not "Exuberant" (both currently v5.9)
#' @noRd
which_ctags <- function () {

    f <- fs::file_temp (pattern = "ctags-out-", ext = ".txt")
    sys::exec_wait ("ctags", args = "--version", std_out = f)
    which_ctags <- brio::read_lines (f)
    regmatches (which_ctags [1], regexpr ("\\w+", which_ctags [1]))
}

#' universal is required, so this should be TRUE
#' @noRd
ctags_is_universal <- function () {

    which_ctags () == "Universal"
}

#' exuberant is the standard apt install, and insufficient here, so this should
#' be FALSE.
#' @noRd
ctags_is_exuberant <- function () {

    which_ctags () == "Exuberant"
}

#' Get tags for one directory within a package
#' @param d the directory
#' @noRd
get_ctags <- function (d = "R", has_tabs) {

    if (!fs::dir_exists (fs::path (fs::path_wd (), d))) {
        return (NULL)
    }

    path_dir <- expand_path (fs::path (fs::path_wd (), d))

    # tab-characters muck up parsing of tag content so have to be removed.
    # This requires modifying the code, so the whole directory is copied to
    # tempdir() and the new path returned. `path_sub` in the following is the
    # path to substitute out of file names given by ctags
    wd <- path_sub <- fs::path_wd ()
    path_tmp_prepend <- "."
    if (has_tabs) {
        path_tmp_prepend <- fs::path_dir (d)
        path_sub <- path_dir <- rm_tabs (path_dir)
        path_dir <- expand_path (fs::path (path_dir, fs::path_file (d)))
        wd <- setwd (path_dir)
        on.exit ({
            unlink (path_sub, recursive = TRUE)
            setwd (wd)
        })
        # called via withr::with_path anyway, so doesn't really matter
    }

    # ctags fields defines at
    # https://docs.ctags.io/en/latest/man/ctags.1.html#extension-fields
    # fields:
    #   - e: Line number for end of object
    #   - F: Name of source file
    #   - K: Kind of tag as long name
    #   - z: The kind key in kind field
    #   - l: language
    #   - n: Line number where `name` is defined
    #   - N: Name of language object
    #   - S: Language-specific signature of object
    #   - t: type and name of a variable

    if (d == "R") {
        fields <- "eFKlnN"
    } else if (d %in% c ("src", "inst")) {
        fields <- "eFKlnN"
    } else { # Generic for adapation to other non-R-pkg dirs
        fields <- "eFKlnN"
    }

    ptn <- paste0 ("ctags-", Sys.getpid (), "-")
    f <- fs::file_temp (pattern = ptn, ext = ".txt")
    args <- c (
        "-R",
        paste0 ("--fields=", fields),
        paste0 ("-f ", f),
        path_dir
    )
    sys::exec_wait ("ctags", args, std_out = FALSE, std_err = FALSE)
    wait_for_process ("ctags")

    # remove header lines:
    x <- brio::read_lines (f)
    x <- x [-which (grepl ("^\\!", x))]

    if (length (x) == 0L) {
        chk <- rm_file_no_err (f)
        return (NULL)
    } # no ctags

    brio::write_lines (x, path = f)

    ctypes <- list (
        readr::col_character (),
        readr::col_character (),
        readr::col_character (),
        readr::col_character (),
        readr::col_character (),
        readr::col_character (),
        readr::col_character ()
    )
    cnames <- c ("tag", "file", "content", "kind", "start", "language", "end")

    n_flds <- readr::count_fields (f,
        tokenizer = readr::tokenizer_tsv (),
        n_max = 100L
    )
    stop_here <- !any (n_flds == length (cnames)) &&
        any (n_flds < length (cnames))
    # Extra fields thus far have only had them as terminal fields, which get
    # removed anyway.
    if (stop_here) {
        chk <- rm_file_no_err (f)
        return (NULL)
    }

    suppressWarnings (
        tags <- tryCatch (
            readr::read_tsv (
                f,
                col_names = cnames,
                col_types = ctypes,
                col_select = cnames,
                progress = FALSE,
                lazy = FALSE
            ),
            error = function (e) NULL
        )
    )

    chk <- rm_file_no_err (f)

    if (is.null (tags) || nrow (tags) == 0L) {
        return (NULL)
    }

    tags <- tags [which (!grepl (excluded_file_ptn (), tags$file)), ]
    if (nrow (tags) == 0L) {
        return (NULL)
    }
    tags <- rm_vendored_code (tags)
    if (nrow (tags) == 0L) {
        return (NULL)
    }

    tags$start <- suppressWarnings (
        as.integer (gsub ("^line\\:", "", tags$start))
    )

    # end tags may fail, and dump text other than "end:XX", so:
    index0 <- grep ("^end\\:", tags$end)
    index1 <- grep ("^[[:alpha:]]", tags$end)
    index1 <- index1 [which (!index1 %in% index0)]
    tags$end [index1] <- NA

    index <- which (!is.na (tags$end))
    tags$end [index] <- gsub ("^end\\:", "", tags$end [index])
    tags$end [index] <- gsub ("[^0-9.-]", "", tags$end [index])
    # as.integer still triggers warnings for NA values, whereas changing
    # storage.mode does not:
    storage.mode (tags$end) <- "integer"

    tags$file <- fs::path_rel (tags$file, path_sub)
    if (path_tmp_prepend != ".") {
        tags$file <- fs::path (path_tmp_prepend, tags$file)
    }

    attr (tags, "has_tabs") <- has_tabs

    return (tags)
}

rm_vendored_code <- function (tags) {

    is_vendored <- vapply (fs::path_split (tags$file), function (f) {
        any (f == "vendor" | f == "vendored")
    }, logical (1L))
    if (any (is_vendored)) {
        tags <- tags [which (!is_vendored), ]
    }

    return (tags)
}

# Add pkg executable directory to the path
.onLoad <- function (libname, pkgname) { # nolint

    bin <- expand_path (fs::path (libname, pkgname, "bin"))
    Sys.setenv (
        PATH = paste (bin, Sys.getenv ("PATH"), sep = .Platform$path.sep)
    )
}

.onUnload <- function (libname, pkgname) { # nolint

    d <- fs::path (fs::path_temp (), "pkgstats-gtags-test")
    if (fs::dir_exists (d)) {
        fs::file_delete (d)
    }
    d <- fs::path (fs::path_temp (), "pkgstats")
    if (fs::dir_exists (d)) {
        fs::file_delete (d)
    }
}

RELEASE_TAG <- "v0.1.6"

#' Update pkgstats` data on GitHub release
#'
#' This function is intended for internal rOpenSci use only. Usage by any
#' unauthorized users will error and have no effect unless run with `upload =
#' FALSE`, in which case updated data will be created in the sub-directory
#' "pkgstats-results" of R's current temporary directory.
#'
#' @param upload If `TRUE`, upload updated results to GitHub release.
#' @return Local path to directory containing updated results.
#' @family archive
#' @export

# nocov start
pkgstats_update <- function (upload = TRUE) {

    requireNamespace ("callr", quietly = TRUE)
    requireNamespace ("hms", quietly = TRUE)
    requireNamespace ("httr2", quietly = TRUE)
    requireNamespace ("piggyback", quietly = TRUE)

    results_path <- fs::dir_create (fs::path (fs::path_temp (), "pkgstats-results"))

    stats_prev_path <- dl_prev_data (results_path, what = "all")
    stats_prev <- readRDS (stats_prev_path)
    fn_names_prev_path <- dl_prev_data (results_path, what = "fn_names")
    fn_names_prev <- readRDS (fn_names_prev_path)

    check_prev_results (stats_prev)
    check_prev_results (fn_names_prev)

    new_cran_pkgs <- unique (c (
        list_new_cran_updates (stats_prev),
        list_new_cran_updates (fn_names_prev)
    ))

    npkgs <- length (new_cran_pkgs)

    if (npkgs == 0) {
        return (NULL)
    }

    message (
        "Downloading and analysing ", npkgs, " packages."
    )

    pt0 <- proc.time ()

    res <- lapply (seq_along (new_cran_pkgs), function (p) {

        stats <- fn_names <- NULL

        tarball_path <- dl_one_tarball (results_path, new_cran_pkgs [p])
        if (!is.null (tarball_path) && fs::file_exists (tarball_path)) {

            tarball_dir <- extract_tarball (tarball_path)

            stats <- one_summary_from_archive (
                tarball_dir,
                save_full = FALSE,
                save_ex_calls = FALSE,
                results_path
            )
            fn_names <- tryCatch (
                pkgstats::pkgstats_fn_names (tarball_dir),
                error = function (e) NULL
            )

            unlink (tarball_dir, recursive = TRUE)
            unlink (tarball_path, recursive = TRUE)
        }

        archive_trawl_progress_message (p, 1, npkgs, pt0)

        list (stats = stats, fn_names = fn_names)
    })

    stats <- do.call (rbind, lapply (res, function (i) i$stats))
    fn_names <- do.call (rbind, lapply (res, function (i) i$fn_names))

    if (!inherits (stats$date, "POSIXt")) {
        stats$date <- as.POSIXct (stats$date, "%y-%m-%d %H-%M-%S")
    }

    stats <- rbind (stats_prev, stats [which (!is.na (stats$package)), ])
    stats_current <- pkgstats_cran_current_from_full (stats)
    fn_names <- rbind (fn_names_prev, fn_names [which (!is.na (fn_names$package)), ])

    # Reduce fn_names to only current pkgs:
    stats_pkgs_current <- paste0 (stats_current$package, "_", stats_current$version)
    fn_nm_pkgs <- paste0 (fn_names$package, "_", fn_names$version)
    fn_names <- fn_names [which (fn_nm_pkgs %in% stats_pkgs_current), ]

    dat <- list (
        "pkgstats-CRAN-all.Rds" = stats,
        "pkgstats-CRAN-current.Rds" = stats_current,
        "pkgstats-fn-names.Rds" = fn_names
    )

    for (i in seq_along (dat)) {
        fpath <- fs::path (results_path, names (dat) [i])
        saveRDS (dat [[i]], fpath)
        piggyback::pb_upload (
            file = fpath,
            repo = "ropensci-review-tools/pkgstats",
            tag = RELEASE_TAG
        )
    }
}
# nocov end

dl_prev_data <- function (results_path, what = "all") {
    what <- match.arg (what, c ("all", "current", "fn_names"))
    files <- c (
        all = "pkgstats-CRAN-all.Rds",
        current = "pkgstats-CRAN-current.Rds",
        fn_names = "pkgstats-fn-names.Rds"
    )
    f <- files [what]
    path <- piggyback::pb_download (
        file = f,
        repo = "ropensci-review-tools/pkgstats",
        dest = results_path,
        tag = RELEASE_TAG
    )
    if (!is.null (path)) {
        path <- path [[1]]$request$output$path
    } else { # downloaded files already exist
        path <- fs::dir_ls (results_path, regexp = f, fixed = TRUE)
    }
    return (path)
}

get_cran_db <- memoise::memoise (tools::CRAN_package_db)

dl_one_tarball <- function (results_path, tarball) {

    cran_url <- "https://cran.r-project.org/src/contrib/"
    if (!grepl ("\\.tar\\.gz$", tarball)) {
        tarball <- paste0 (tarball, ".tar.gz")
    }
    url <- paste0 (cran_url, tarball)
    path <- fs::path (results_path, tarball)

    if (fs::file_exists (path)) {
        return (NULL)
    }

    # No native pipe here...
    req <- httr2::request (url)
    req <- httr2::req_headers (req, "Accept" = "application/octet-stream")
    resp <- tryCatch (
        httr2::req_perform (req),
        error = function (e) NULL
    )

    if (is.null (resp)) {
        return (NULL)
    }
    if (httr2::resp_is_error (resp)) {
        return (NULL)
    }

    writeBin (httr2::resp_body_raw (resp), path)
    return (path)
}

list_new_cran_updates <- function (prev_results) {

    prev_tarball <- unique (paste0 (prev_results$package, "_", prev_results$version))
    cran_pkgs <- get_cran_db ()
    cran_tarball <- paste0 (cran_pkgs$Package, "_", cran_pkgs$Version)

    cran_pkgs <- cran_pkgs [which (!cran_tarball %in% prev_tarball), ]
    ret <- character (0L)
    if (nrow (cran_pkgs) > 0) {
        ret <- paste0 (cran_pkgs$Package, "_", cran_pkgs$Version)
    }
    return (ret)
}

check_prev_results <- function (prev_results) {
    msg <- paste0 (
        "'prev_results' must be given, and must be a 'data.frame' from a ",
        "previous call to 'pkgstats_from_archive()' or 'pkgstats_update()'."
    )
    if (is.null (prev_results)) stop (msg, call. = FALSE)
    if (!is.null (prev_results)) {
        if (!inherits (prev_results, "data.frame")) stop (msg, call. = FALSE)
        if (nrow (prev_results) < 20000) stop (msg, call. = FALSE)
        if (ncol (prev_results) > 3L) {
            if (ncol (prev_results) != ncol (null_stats ())) stop (msg, call. = FALSE)
            if (!identical (names (prev_results), names (pkgstats_summary ()))) {
                stop (msg, call. = FALSE)
            }
        } else {
            if (ncol (prev_results) != 3L) stop (msg, call. = FALSE)
            nms <- c ("package", "version", "fn_name")
            if (!identical (names (prev_results), nms)) {
                stop (msg, call. = FALSE)
            }
        }
    }
}

test_all <- (identical (Sys.getenv ("MPADGE_LOCAL"), "true") ||
    identical (Sys.getenv ("GITHUB_JOB"), "test-coverage"))

test_that ("pkgstats", {

    if (!test_all) {
        Sys.setenv ("PKGSTATS_CRAN_TESTS" = "true")
    }

    path <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")

    # message is now produced once per session by readr, but can only be
    # suppressed by Suggesting yet another package, `tidyselect`.
    # expect_message (
    s <- pkgstats (path)

    if (!test_all) {
        Sys.unsetenv ("PKGSTATS_CRAN_TESTS")
    }

    expect_type (s, "list")

    nms <- c (
        "loc",
        "vignettes",
        "data_stats",
        "desc",
        "translations",
        "objects",
        "network",
        "external_calls"
    )
    expect_true (all (nms %in% names (s)))

    expect_s3_class (s$loc, "tbl_df")
    # The following 2 tests fail on GitHub windows machines for some reason?
    is_windows <- Sys.info () [["sysname"]] == "Windows"
    if (!is_windows) {
        expect_gte (nrow (s$loc), 3L)
        expect_true (all (c ("R", "src", "tests") %in% s$loc$dir))
    }

    expect_type (s$vignettes, "integer")
    expect_named (s$vignettes)
    expect_length (s$vignettes, 2L)

    expect_type (s$data_stats, "integer")
    expect_length (s$data_stats, 3L)
    expect_named (s$data_stats)

    expect_s3_class (s$desc, "data.frame")
    expect_equal (nrow (s$desc), 1L)
    expect_type (s$translations, "character")

    expect_s3_class (s$objects, "data.frame")

    skip_if (!test_all)

    expect_true (nrow (s$objects) >= 4L)
    nms <- c (
        "file_name",
        "fn_name",
        "kind",
        "language",
        "loc",
        "npars",
        "has_dots",
        "exported",
        "param_nchars_md",
        "param_nchars_mn",
        "num_doclines"
    )
    expect_true (all (nms %in% names (s$objects)))

    expect_s3_class (s$network, "data.frame")
    expect_true (nrow (s$network) > 0L)
    # expect_true (nrow (s$network) < nrow (s$objects))
    nms <- c (
        "file",
        "line1",
        "from",
        "to",
        "language",
        "cluster_dir",
        "cluster_undir",
        "centrality_dir",
        "centrality_undir"
    )
    expect_true (all (nms %in% names (s$network)))
    expect_true (all (c ("R", "C++") %in% s$network$language))

    ext <- s$external_calls
    expect_s3_class (ext, "data.frame")
    expect_true (nrow (ext) > 2L)
    expect_true ("stats" %in% ext$package)
    expect_true (length (unique (ext$package)) > 15L)
})

skip_if (!test_all)

test_that ("pkgstats with tabs", {

    if (!test_all) {
        Sys.setenv ("PKGSTATS_CRAN_TESTS" = "true")
    }

    path <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
    path <- extract_tarball (path)
    flist <- fs::dir_ls (fs::path (path, "R"), type = "file", recurse = TRUE)
    f <- flist [which (basename (flist) == "loc.R")]
    x <- readLines (f)
    x <- gsub ("^\\s{12}", "\\\t\\\t\\\t", x)
    x <- gsub ("^\\s{8}", "\\\t\\\t", x)
    x <- gsub ("^\\s{4}", "\\\t", x)
    writeLines (x, f)

    # message is now produced once per session by readr, but can only be
    # suppressed by Suggesting yet another package, `tidyselect`.
    # expect_message (
    s <- pkgstats (path)

    if (!test_all) {
        Sys.unsetenv ("PKGSTATS_CRAN_TESTS")
    }

    expect_s3_class (s$loc, "tbl_df")
    expect_true (s$loc$ntabs [s$loc$dir == "R"] > 0L)

    fs::dir_delete (path)
})

test_that ("extra_manifest_paths", {

    path <- extract_tarball (
        system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
    )
    s0 <- pkgstats (path)

    expect_null (extra_manifest_paths (path))

    extra_path <- fs::path (path, "extra_dir", "vendor_code")
    fs::dir_create (extra_path)
    cpp_code <- c (
        "// some vendor code",
        "int a () {",
        "  return 1L;",
        "}"
    )
    writeLines (cpp_code, fs::path (extra_path, "test.cpp"))
    toml <- c (
        "[\"vendor_sources\"]",
        "code=\"../extra_dir/vendor_code\""
    )
    writeLines (toml, fs::path (path, "src", "SOURCE_MANIFEST.toml"))

    extra_paths <- extra_manifest_paths (path)
    expect_type (extra_paths, "character")
    expect_length (extra_paths, 1L)
    expect_true (any (grepl ("vendor\\_code", extra_paths)))
    expect_true (all (fs::dir_exists (extra_paths)))

    s1 <- pkgstats (path)

    expect_gt (nrow (s1$loc), nrow (s0$loc))
    expect_gt (sum (s1$loc$nfiles), sum (s0$loc$nfiles))
    expect_gt (sum (s1$loc$nlines), sum (s0$loc$nlines))
    # Should also be one more tagged source object:
    expect_gt (nrow (s1$objects), nrow (s0$objects))
    extra_paths <- fs::path_rel (extra_paths, path)
    expect_true (all (extra_paths %in% fs::path_dir (s1$objects$file)))
    expect_false (any (extra_paths %in% fs::path_dir (s0$objects$file)))

    fs::dir_delete (path)
})

#' My (@mpadge) contribution to Brian Connely's 'ami' package which got removed
#' from CRAN, and so removed as dependency here.
#'
#' \url{https://github.com/briandconnelly/ami/blob/main/R/cran.R}, with logic
#' of function detailed in
#' \url{https://github.com/briandconnelly/ami/issues/14}
#' @noRd
on_cran <- function (cran_pattern = "_R_", n_cran_envvars = 5L) {

    checkmate::assert_string (cran_pattern)
    checkmate::assert_integer (n_cran_envvars, lower = 1L, len = 1L)

    not_cran <- identical (Sys.getenv ("NOT_CRAN", "false"), "true")
    if (!not_cran) { # Not using the envvar
        index <- grep (cran_pattern, names (Sys.getenv ()), fixed = TRUE)
        not_cran <- length (index) < n_cran_envvars
    }
    !not_cran
}

# nocov start
.onAttach <- function (libname, pkgname) { # nolint

    exit_now <- !interactive () || on_cran ()
    if (exit_now) {
        return ()
    }

    pkg_path <- fs::path_abs (fs::path (libname, pkgname))

    os <- Sys.info () ["sysname"]
    if (os == "Linux") {
        return ()
    }
    has_ctags <- pkgstats::ctags_test ()


    if (os == "Darwin" && !has_ctags) {
        packageStartupMessage (
            "This package requires downloading and ",
            "installing binary 'universal-ctags' software from:"
        )
        packageStartupMessage (
            "https://github.com/autobrew/archive/tree/master/high_sierra"
        )
        packageStartupMessage (
            "Run pkgstats::ctags_install() to install universal-ctags."
        )
    } else if (os == "Windows" && !has_ctags) {
        packageStartupMessage (
            "This package requires downloading and installing ",
            "binary 'universal-ctags' software from:"
        )
        packageStartupMessage ("https://github.com/rwinlib/universal-ctags/")
        packageStartupMessage (
            "Run pkgstats::ctags_install() to install universal-ctags."
        )
    }
}

.onLoad <- function (libname, pkgname) {

    os <- Sys.info () ["sysname"]

    if (os == "Windows") {
        pkg_path <- system.file (package = "pkgstats")
        pkg_path <- gsub ("/inst", "", pkg_path)
        ctags_path <- get_ctags_path_win (pkg_path)
        if (!grepl ("universal-ctags-5.9.20210530.0", Sys.getenv ("PATH"))) {
            Sys.setenv (PATH = paste0 (Sys.getenv ("PATH"), ";", ctags_path))
        }
    }

}
# nocov end

test_that ("manifest paths", {

    exdir <- fs::path (fs::path_temp (), "pkgroot", "subdir")
    path <- extract_tarball (
        system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats"),
        exdir = exdir
    )
    extra_path <- fs::path (path, "extra_dir", "vendor_code")
    fs::dir_create (extra_path)

    check_path (exdir)
    check_path (fs::path_dir (exdir))
    check_path (extra_path)

    fs::dir_delete (path)
    expect_error (
        check_path (exdir),
        "Could not find unambiguous project root"
    )

    fs::dir_delete (exdir)
})

#' use ctags and gtags to parse call data
#'
#' @param path Path to local repository
#' @param has_tabs A logical flag indicating whether or not the code contains
#' any tab characters. This can be determined from \link{loc_stats}, which has a
#' `tabs` column. If not given, that value will be extracted from internally
#' calling that function.
#' @param pkg_name Only used for external_call_network, to label
#' package-internal calls.
#' @return A list of three items:
#' \itemize{
#' \item "network" A `data.frame` of relationships between objects, generally as
#' calls between functions in R, but other kinds of relationships in other
#' source languages. This is effectively an edge-based network representation,
#' and the data frame also include network metrics for each edge, calculated
#' through representing the network in both directed (suffix "_dir") and
#' undirected (suffix "_undir") forms.
#' \item "objects" A `data.frame` of statistics on each object (generally
#' functions in R, and other kinds of objects in other source languages),
#' including the kind of object, the language, numbers of lines-of-code,
#' parameters, and lines of documentation, and a binary flag indicating whether
#' or not R functions accept "three-dots" parameters (`...`).
#' \item "external_calls" A `data.frame` of every call from within every R
#' function to any external R package, including base and recommended packages.
#' The location of each calls is recorded, along with the external function and
#' package being called.
#' }
#' @family tags
#' @export
#' @examples
#' f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
#' # have to extract tarball to call function on source code:
#' path <- extract_tarball (f)
#' @examplesIf ctags_test (noerror = TRUE)
#' tags <- tags_data (path)
tags_data <- function (path, has_tabs = NULL, pkg_name = NULL) {

    chk <- tryCatch (ctags_test (noerror = FALSE),
        error = function (e) e
    )
    if (methods::is (chk, "simpleError")) {
        return (dummy_tags_data ())
    }

    if (is.null (has_tabs)) {
        has_tabs <- any (loc_stats (path)$ntabs > 0L)
    } else if (!is.logical (has_tabs) || length (has_tabs) > 1L) {
        stop ("has_tabs must either be NULL or a single logical value")
    }

    if (is.null (pkg_name)) {
        desc <- fs::path (path, "DESCRIPTION")
        checkmate::assert_file (desc)
        d <- data.frame (read.dcf (desc), stringsAsFactors = FALSE)
        pkg_name <- d [["Package"]]
    }

    kind <- start <- NULL # no visible binding messages

    tags_r <- withr::with_dir (path, get_ctags ("R", has_tabs))
    if (!is.null (tags_r)) {
        tags_r <- count_doclines_r (tags_r, path)
    }

    external_calls <- external_call_network (tags_r, path, pkg_name)

    tags_src <- withr::with_dir (path, get_ctags ("src", has_tabs))
    tags_inst <- withr::with_dir (path, get_ctags ("inst", has_tabs))

    extra_paths <- extra_manifest_paths (path)
    if (length (extra_paths) > 0L) {
        path_base <- fs::path_common (c (path, extra_paths))
        extra_paths_rel <- fs::path_rel (extra_paths, path_base)
        extra_tags_src <- lapply (seq_along (extra_paths), function (i) {
            extra_tags <- withr::with_dir (path_base, get_ctags (extra_paths_rel [i], has_tabs))
            path_full <- fs::path (path_base, extra_tags$file)
            extra_tags$file <- fs::path_rel (path_full, path)
            return (extra_tags)
        })
        extra_tags_src <- do.call (rbind, extra_tags_src)
        tags_src <- rbind (tags_src, extra_tags_src)
    }

    tags_src <- count_doclines_src (tags_src, path)
    tags_inst <- count_doclines_src (tags_inst, path)

    gtags <- NULL

    if (!is.null (tags_src) || !is.null (tags_inst)) {

        no_gtags <- withr::with_dir (path, make_gtags ())

        # gtags does not parse R code, so may return NULL if there is no code in
        # other languages
        gtags <- withr::with_dir (path, get_gtags ())
        if (length (extra_paths) > 0) {
            gtags_extra <- lapply (extra_paths, function (p) {
                withr::with_dir (p, {
                    make_gtags ()
                    get_gtags ()
                })
            })
        }

        ctags <- dplyr::arrange (rbind (tags_src, tags_inst), file, start)
        ctags <- dplyr::filter (ctags, kind %in%
            c ("class", "function", "struct"))

        if (!is.null (gtags)) {

            gtags$from <- NA_character_
            for (f in unique (ctags$file)) {
                gtags <- gtags_from_one_file (ctags, gtags, f)
            }
            gtags <- gtags [which (gtags$tag %in% ctags$tag), ]

            langs <- ctags [, c ("tag", "language")]
            langs <- langs [which (!duplicated (langs)), ]
            gtags$language <- gsub (
                "^language\\:", "",
                langs$language [match (gtags$tag, langs$tag)]
            )
        }

        if (no_gtags) {
            chk <- rm_gtags_files (path)
        }
    }

    # Functions defined from namespaces of other packages are tagged as global variables.
    kinds <- c ("function", "globalVar")
    fns_r <- tags_r [which (tags_r$kind %in% kinds & !is.na (tags_r$tag)), ]
    fn_vars_r <- tags_r [which (tags_r$kind == "functionVar" &
        !is.na (tags_r$tag)), ]

    call_graph_r <- fn_var_call_graph_r (
        fns_r,
        fn_vars_r,
        path
    )
    call_graph_src <- fn_var_call_graph_src (gtags)

    network <- rbind (call_graph_r, call_graph_src)
    network <- network [which (!is.na (network$from)), ]

    if (nrow (network) > 0) {

        network <- add_igraph_stats (network, directed = TRUE)
        network <- add_igraph_stats (network, directed = FALSE)
        network$line2 <- NULL
    }

    return (list (
        network = network,
        objects = src_objects (rbind (tags_r, tags_src, tags_inst)),
        external_calls = external_calls
    ))
}

dummy_tags_data <- function () {

    n <- data.frame (matrix (nrow = 0, ncol = 4))
    names (n) <- c ("file", "line1", "from", "to")
    e <- data.frame (matrix (nrow = 0, ncol = 8))
    names (e) <- c (
        "tags_line", "call", "tag", "file", "kind",
        "start", "end", "package"
    )
    return (list (
        network = n,
        external_calls = e
    ))
}

#' Replace tab indentation with a fixed number of spaces
#'
#' This is necessary because the files produced by both ctags and global are
#' unable to be propertly parsed when code contains tab indents. Because this
#' function modifies code, the entire directory, `d`, is copied to `tempdir()`,
#' and the resultant path returned.
#'
#' @param d A directory in which tab indents are to be replaced in all files
#' @param nspaces The equivalent number of spaces with which to replace tab
#' indentations. This parameter has no effect on results.
#' @noRd
rm_tabs <- function (d, nspaces = 2) {

    tmpd <- paste0 (sample (c (letters, LETTERS), size = 8, replace = TRUE),
        collapse = ""
    )
    tmpd <- fs::path (fs::path_temp (), tmpd)
    fs::dir_create (tmpd, recurse = TRUE)
    chk <- fs::dir_copy (d, tmpd) # copies 'd' as sub-dir of tmpd
    if (length (chk) != length (d)) {
        stop ("Unable to copy files from [", d, "] to tempdir()")
    }

    sp <- paste0 (rep (" ", nspaces), collapse = "")

    files <- expand_path (fs::dir_ls (tmpd, recurse = TRUE, type = "file"))

    exts <- file_exts ()
    exts$ext <- gsub ("+", "\\+", exts$ext, fixed = TRUE)
    exts$ext <- gsub ("-", "\\-", exts$ext, fixed = TRUE)
    exts <- paste0 (exts$ext, "$", collapse = "|")
    files <- files [grep (exts, files)]
    files <- files [which (!grepl ("^Makevars", files))]
    # also remove pdf and html files
    index <- which (!grepl ("\\.(pdf|html)$", files, ignore.case = TRUE))
    files <- files [index]

    for (f in files) {
        x <- tryCatch (
            suppressWarnings (brio::read_lines (f)),
            error = function (e) NULL
        )
        if (is.null (x)) {
            next
        }

        has_tabs <- any (grepl ("\\t", x))
        if (has_tabs) {
            # The input string from 'brio' above can be invalid and cause
            # 'gsub' to fail (try on antiword)
            xt <- tryCatch (gsub ("^\\t", sp, x), error = function (e) x)
            # And replace non-leading tabs with single
            xt <- tryCatch (gsub ("\\t", " ", xt), error = function (e) xt)
            if (!identical (x, xt)) {
                brio::write_lines (x, path = f)
            }
        }
    }

    return (expand_path (tmpd))
}

#' convert tags_src into same format as the function summary of R code
#' @param tags tags_src from main fn
#' @noRd
src_objects <- function (tags) {

    res <- data.frame (
        file_name = NA_character_,
        fn_name = NA_character_,
        kind = NA_character_,
        language = NA_character_,
        loc = 0L,
        npars = NA_integer_,
        has_dots = NA,
        num_doclines = NA_integer_,
        stringsAsFactors = FALSE
    )

    if (!is.null (tags)) {

        res <- data.frame (
            file_name = tags$file,
            fn_name = tags$tag,
            kind = tags$kind,
            language = gsub ("^language:", "", tags$language),
            loc = tags$end - tags$start + 1,
            npars = NA_integer_,
            has_dots = NA,
            num_doclines = tags$doclines,
            stringsAsFactors = FALSE
        )
        # ctags does not count loc in Rust:
        res <- res [which (!is.na (res$loc) |
            res$language == "Rust"), ]
    }

    return (res)
}

#' Count non-roxygen documentation lines in R files
#' @noRd
count_doclines_r <- function (tags, path) {

    flist <- unique (tags$file)
    cmt_lines <- lapply (flist, function (f) {
        code <- brio::read_lines (fs::path (path, f))
        grep ("(\\s*)\\#", code)
    })
    names (cmt_lines) <- flist

    tags$doclines <- vapply (seq_len (nrow (tags)), function (i) {
        n <- 0L
        if (is.na (tags$kind [i]) || !tags$kind [i] == "function") {
            return (n)
        }
        if (is.na (tags$start [i]) || is.na (tags$end [i])) {
            return (n)
        }
        if (tags$end [i] > tags$start [i]) {
            doclines <- cmt_lines [[match (tags$file [i], names (cmt_lines))]]
            doclines <- doclines [which (doclines < tags$start [i])]
            dd <- which (diff (doclines) > 1)
            if (length (dd) > 0L) {
                index <- seq (max (dd) + 1L, length (doclines))
                doclines <- doclines [index]
            }
            tags_file <- tags [which (tags$file == tags$file [i]), ]
            index <- which (tags_file$end < tags$start [i])
            if (length (index) > 0L) {
                prev_code <- max (tags_file$end [index])
                if (length (prev_code) > 0L) {
                    doclines <- doclines [which (doclines > prev_code)]
                }
            }
            n <- length (doclines)
        }
        return (n)
    }, integer (1L))

    return (tags)
}

#' Count documentation lines in src files.
#'
#' Currently only implemented for C and C++
#' @param tags The output of `get_ctags("src")` or `get_ctags("inst")`
#' @return Modified version of `tags` with additional `doclines` column
#' counting numbers of documentation lines for each tagged object.
#' @noRd
count_doclines_src <- function (tags, path) {

    if (length (tags) == 0L) {
        return (tags)
    }

    tags$language [which (is.na (tags$language))] <- ""

    files <- unique (tags$file)
    langs <- gsub ("^language\\:", "", tags$language [match (files, tags$file)])
    doclines <- apply (cbind (files, langs), 1, function (i) {

        f <- fs::path (path, i [1])
        code <- brio::read_lines (f)

        exts <- file_exts ()
        ext_i <- exts [match (i [2], exts$type), ]

        out <- grep (ext_i$cmt, code)
        if (nzchar (ext_i$cmt_open) && !is.na (ext_i$cmt_open)) {
            cmt_open <- grep (ext_i$cmt_open, code)
            cmt_close <- grep (ext_i$cmt_close, code)
            if (length (cmt_open) == length (cmt_close)) {
                index <- cbind (cmt_open, cmt_close)
                block_cmts <- apply (index, 1, function (j) {
                    seq (j [1], j [2])
                })
                out <- c (out, unlist (block_cmts))
            }
        }
        return (sort (unique (out)))
    })
    if (length (doclines) > 0L) {
        names (doclines) <- files
    }

    if (length (doclines) == 0L) {
        res <- rep (0L, nrow (tags))
    } else {
        res <- vapply (seq_len (nrow (tags)), function (i) {

            ndoclines <- 0L
            if (is.na (tags$start [i])) {
                return (ndoclines)
            }

            # ctags does not tag end line numbers of Rust objects:
            if (tags$language [i] == "language:Rust") {
                f <- fs::path (path, tags$file [i])
                code <- brio::read_lines (f)
                tag_line_start <- tags$start [i]
                all_ends <- grep ("^\\}", code)
                prev_end <- 1L
                if (any (all_ends < tag_line_start)) {
                    prev_end <- max (all_ends [which (all_ends < tag_line_start)])
                }
                not_code <- code [seq (prev_end + 1L, tag_line_start - 1L)]
                ndoclines <- length (grep ("^\\s*\\/\\/", not_code))
            } else {

                file_tags <- tags [
                    tags$file == tags$file [i],
                    c ("start", "end")
                ]
                file_tags <- file_tags [which (!is.na (file_tags$end)), ]
                index <- which (file_tags$end < tags$start [i])
                if (length (index) > 0) {
                    prev_code <- max (file_tags$end [index], na.rm = TRUE)
                    index <- which (names (doclines) == tags$file [i])
                    doclines_f <- doclines [[index]]
                    if (length (doclines_f) > 0L) {
                        index <- which (
                            doclines_f < tags$start [i] & doclines_f > prev_code
                        )
                        ndoclines <- length (index)
                    }
                }
            }

            return (ndoclines)
        }, integer (1L))
    }

    tags$doclines <- res

    return (tags)
}

#' Reduce `data.frame` of full CRAN archive data to current packages only.
#'
#' @inheritParams pkgstats_from_archive
#' @return A `data.frame` object with one row for each package containing
#' summary statistics generated from the \link{pkgstats_summary} function.
#'
#' @family archive
#' @export
pkgstats_cran_current_from_full <- function (prev_results, results_file = NULL) {

    prev_tarball <- paste0 (prev_results$package, "_", prev_results$version)

    cran_pkgs <- get_cran_db ()
    cran_tarball <- paste0 (cran_pkgs$Package, "_", cran_pkgs$Version)

    index <- which (prev_tarball %in% cran_tarball)
    res <- prev_results [index, ]

    if (!is.null (res) && !is.null (results_file)) {

        results_file <- archive_results_file_name (results_file)
        saveRDS (res, results_file)
    }

    return (res)
}

# Mostly only lightly adapted from
# https://github.com/stan-dev/cmdstanr/blob/master/R/install.R

# nocov start
clone_ctags <- function (destdir = NULL) {

    if (!has_git ()) {
        stop ("Ctags can only be installed with git", call. = FALSE)
    }

    if (is.null (destdir)) {
        stop ("destdir must be specified", call. = FALSE)
    }
    destdir <- expand_path (destdir)
    if (!fs::dir_exists (destdir)) {
        stop ("Directory [", destdir,
            "] does not exist",
            call. = FALSE
        )
    }

    u <- "https://github.com/universal-ctags/ctags.git"

    withr::with_dir (destdir, {
        sys::exec_wait ("git", args = c ("clone", u))
    })

    return (fs::path (destdir, "ctags"))
}

has_git <- function () {

    nzchar (dirname (Sys.which ("git")))
}

#' Make install ctags from extracted .tar.gz in 'ctags_dir', returned from
#' `download_ctags()`.
#' @inheritParams ctags_install
#' @param ctags_dir Directory of 'ctags' repository cloned with
#' \link{clone_ctags}.
#' @noRd
ctags_make <- function (ctags_dir, bin_dir = NULL, sudo = TRUE) {

    if (!sudo && is.null (bin_dir)) {
        stop ("A value for 'bin_dir' must be specified when 'sudo = FALSE'")
    }

    f <- fs::file_temp (pattern = "ctags-make-", ext = ".txt")

    confargs <- NULL
    if (!is.null (bin_dir)) {
        if (!fs::dir_exists (bin_dir)) {
            stop ("bin_dir [", bin_dir, "] does not exist.", call. = FALSE)
        }
        confargs <- paste0 ("--prefix=", bin_dir)
    }

    if (sudo) {
        cmd <- "sudo"
        arg <- c ("make", "install")
    } else {
        cmd <- "make"
        arg <- "install"
    }

    withr::with_dir (ctags_dir, {
        sys::exec_wait ("./autogen.sh", std_out = f)
        sys::exec_wait ("./configure", args = confargs, std_out = f)
        sys::exec_wait ("make", std_out = f)
        sys::exec_wait (cmd, args = arg, std_out = f)
    })
}

#' Install 'ctags' from a clone of the 'git' repository
#'
#' 'ctags' is installed with this package on both Windows and macOS systems;
#' this is an additional function to install from source on Unix systems.
#'
#' @param bin_dir This parameter only has an effect on *nix-type operating
#' systems (such as Linux), on which it's a prefix to pass to the
#' `autoconf` configure command defining location to install the binary, with
#' default of `/usr/local`.
#' @param sudo Set to `FALSE` if `sudo` is not available, in which case a
#' value for `bin_dir` will also have to be explicitly specified, and be a
#' location where a binary is able to be installed without `sudo` privileges.
#' @return Nothing; the function will fail if installation fails, otherwise
#' returns nothing.
#'
#' @family tags
#' @examples
#' \dontrun{
#' ctags_install (bin_dir = "/usr/local") # default Linux location.
#' }
#' @export
ctags_install <- function (bin_dir = NULL, sudo = TRUE) {

    pkg_path <- system.file (package = "pkgstats")
    pkg_path <- gsub ("/inst", "", pkg_path)

    os <- Sys.info () ["sysname"]

    if (os == "Darwin") {
        install_ctags_macos (pkg_path)
    }

    if (os == "Windows") {
        install_ctags_windows (pkg_path)
    }

    if (.Platform$OS.type == "unix" & os != "Darwin") {
        ctags_dir <- clone_ctags (destdir = tempdir ())
        ctags_make (ctags_dir, bin_dir, sudo)

        if (!has_gtags ()) {
            gtags_install (sudo = sudo)
        }

        if (!sudo) {
            # system ("hash -d ctags")
            system ("hash ctags")
            system ("hash gtags")
        }
    }

    if (ctags_test (noerror = TRUE)) {
        # already installed and okay
        message ("ctags installed and passed test.")
        return (NULL)
    }
}

install_ctags_macos <- function (pkg_path) {

    ctags_dir <- fs::path (pkg_path, "inst", "bin")
    if (!fs::dir_exists (ctags_dir)) {
        fs::dir_create (ctags_dir, recurse = TRUE)
    }
    ctags_path <- fs::path (ctags_dir, "ctags")
    if (!file.exists (ctags_path)) {
        u <- "https://autobrew.github.io/archive/high_sierra/universal-ctags-p5.9.20210530.0.tar.xz" # nolint
        f <- basename (u)
        utils::download.file (u, f, quiet = TRUE)
        utils::untar (f, extras = "--strip-components=1", exdir = ctags_dir)
        fs::file_delete (f)
    }
}

get_ctags_path_win <- function (pkg_path) {
    u <- "https://github.com/rwinlib/universal-ctags/archive/refs/tags/v5.9.20210530.0.zip" # nolint
    ctags_path <- normalizePath (fs::path (
        pkg_path,
        "windows",
        paste0 (
            "universal-ctags-",
            tools::file_path_sans_ext (gsub ("^v", "", basename (u)))
        ),
        "bin"
    ), mustWork = FALSE)
    ctags_path
}

install_ctags_windows <- function (pkg_path) {
    u <- "https://github.com/rwinlib/universal-ctags/archive/refs/tags/v5.9.20210530.0.zip" # nolint
    ctags_path <- get_ctags_path_win (pkg_path)
    ctag_win_path <- fs::path_abs (fs::path (pkg_path, "windows"))
    if (!fs::dir_exists (ctag_win_path)) {
        fs::dir_create (ctag_win_path, recurse = TRUE)
    }
    if (!fs::file_exists (ctags_path)) {
        f <- tempfile (fileext = "pkgstats_ctags_lib.zip")
        utils::download.file (u, f, quiet = TRUE)
        utils::unzip (f, exdir = ctags_path, junkpaths = TRUE)
        fs::file_delete (f)
    }
    if (!grepl ("universal-ctags-5.9.20210530.0", Sys.getenv ("PATH"))) {
        Sys.setenv (PATH = paste0 (Sys.getenv ("PATH"), ";", ctags_path))
    }
}
# nocov end

test_all <- (identical (Sys.getenv ("MPADGE_LOCAL"), "true") ||
    identical (Sys.getenv ("GITHUB_JOB"), "test-coverage"))

test_that ("desc stats", {

    if (!test_all) {
        Sys.setenv ("PKGSTATS_CRAN_TESTS" = "true")
    }

    path <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
    expect_error (
        desc_stats (path),
        "path must be directory containing package source"
    )
    path <- extract_tarball (path)
    expect_silent (d <- desc_stats (path))

    expect_s3_class (d, "data.frame")
    expect_equal (nrow (d), 1L)
    expect_equal (ncol (d), 17L)
    expect_identical (
        names (d),
        c (
            "package",
            "version",
            "date",
            "license",
            "urls",
            "bugs",
            "aut",
            "ctb",
            "fnd",
            "rev",
            "ths",
            "trl",
            "depends",
            "imports",
            "suggests",
            "enhances",
            "linking_to"
        )
    )
})

test_that ("fn_names", {

    path <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
    s <- pkgstats_fn_names (path)
    expect_s3_class (s, "data.frame")
    expect_equal (ncol (s), 3L)
    expect_named (s, c ("package", "version", "fn_name"))

    expect_type (s$package, "character")
    expect_type (s$version, "character")
    expect_type (s$fn_name, "character")
    expect_equal (class (s$fn_name), "noquote")

    expect_length (unique (s$package), 1L)
    expect_length (unique (s$version), 1L)
    expect_length (unique (s$fn_name), nrow (s))
})

#' Types of 'person' to be extracted from 'Authors@R' field of 'DESCRIPTION'
#' file. See '?person' for details.
#' - aut = Author
#' - ctb = Contributor
#' - fnd = Funder
#' - rev = Reviewer
#' - ths = Thesis (Advisor)
#' - trl = Translator (usually from another language)
#' @noRd
aut_types <- function () {

    c ("aut", "ctb", "fnd", "rev", "ths", "trl")
}

#' Statistics from DESCRIPTION files
#'
#' @inheritParams loc_stats
#' @return A `data.frame` with one row and 16 columns extracting various
#' information from the 'DESCRIPTION' file, include websites, tallies of
#' different kinds of authors and contributors, and package dependencies.
#' @family stats
#' @export
#' @examples
#' f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
#' # have to extract tarball to call function on source code:
#' path <- extract_tarball (f)
#' desc_stats (path)
desc_stats <- function (path) {

    path <- check_path (path)
    if (grepl ("\\.tar\\.gz$", path)) {
        stop ("path must be directory containing package source", call. = FALSE)
    }

    desc <- fs::path (path, "DESCRIPTION")
    d <- data.frame (
        read.dcf (desc),
        stringsAsFactors = FALSE
    )
    # The following pkgs have extra guff in version strings:
    # "X-X.X (version YY/MM/DD)"
    # gee 4.13-2
    # gee 4.13-3
    # nlme 3-1.2
    version <- gsub ("\\s+.*$", "", d$Version)
    license <- d$License
    urls <- NA_character_
    if ("URL" %in% names (d)) {
        urls <- d$URL
    }

    n_aut <- desc_authors (d)

    # Dependencies:
    dep <- extract_deps (d, "Depends")
    imp <- extract_deps (d, "Imports")
    sug <- extract_deps (d, "Suggests")
    enh <- extract_deps (d, "Enhances")
    lnk <- extract_deps (d, "LinkingTo")

    bugs <- ifelse ("BugReports" %in% names (d),
        d$BugReports,
        NA_character_
    )

    desc_date <- ifelse ("Date.Publication" %in% names (d),
        gsub ("\\sUTC$", "", d$Date.Publication),
        paste0 (file.info (desc)$mtime)
    )

    data.frame (
        package = d$Package,
        version = version,
        date = desc_date,
        license = license,
        urls = urls,
        bugs = bugs,
        n_aut,
        depends = dep,
        imports = imp,
        suggests = sug,
        enhances = enh,
        linking_to = lnk,
        stringsAsFactors = FALSE
    )
}

desc_authors <- function (d) {

    if ("Authors.R" %in% names (d)) {
        authors <- tryCatch (
            eval (parse (text = d$Authors.R)),
            error = function (e) NULL
        )
        if (is.null (authors)) {
            # some pkgs fail to parse, like surveillance_1.20.0
            authors <- lapply (
                strsplit (d$Authors.R, ")),"),
                function (i) paste0 (i, ")")
            )
        }
        # remove everything before and after square brackets
        authors <- gsub ("^.*\\[|\\].*$", "", authors)
        n_aut <- vapply (
            aut_types (), function (i) {
                length (grep (i, authors))
            },
            integer (1)
        )
    } else {
        # There is no reliable way to establish numbers of authors for packages
        # which only have an "Author" field, because these are not intended to
        # be machine-parseable.
        authors <- d$Author
        if (grepl ("\\sand\\s", authors)) {
            authors <- strsplit (authors, "\\sand\\s") [[1]]
        } else if (grepl (",", authors)) {
            authors <- strsplit (authors, ",") [[1]]
        }
        n_aut <- rep (0, length (aut_types ()))
        names (n_aut) <- aut_types ()
        n_aut [1] <- length (authors)
    }

    n_aut <- data.frame (matrix (n_aut, nrow = 1))
    names (n_aut) <- aut_types ()

    return (n_aut)
}

extract_deps <- function (d, type = "Depends") {

    res <- d [[type]]
    res <- ifelse (length (res) == 0, NA_character_, res)
    res <- strsplit (res, ",") [[1]]

    if (type == "Depends") {
        res <- res [which (!grepl ("^R(\\s|\\()", res))]
        res <- ifelse (length (res) == 0, NA_character_, res)
    }

    res <- vapply (res, function (i) {
        strsplit (i, "\\(") [[1]] [1]
    },
    character (1),
    USE.NAMES = FALSE
    )

    res <- gsub ("^\\s+|\\s+$", "", res)

    return (paste0 (res, collapse = ", "))
}

test_all <- (identical (Sys.getenv ("MPADGE_LOCAL"), "true") ||
    identical (Sys.getenv ("GITHUB_JOB"), "test-coverage"))

test_that ("archive trawl", {

    if (!test_all) {
        Sys.setenv ("PKGSTATS_CRAN_TESTS" = "true")
    }

    archive <- setup_test_archive ()

    expect_type (archive, "list")
    expect_length (archive, 3L)
    expect_identical (names (archive), c ("tarball", "path", "archive_path"))

    v <- get_pkg_version (archive$path)
    expect_type (v, "character")
    expect_length (v, 2L)
    expect_identical (v, c ("pkgstats", "9.9"))

    f <- archive_results_file_name ("aaa.txt")
    expect_equal (basename (f), "aaa.Rds")
    expect_equal (f, fs::path (fs::path_wd (), "aaa.Rds"))

    expect_error (
        pkgstats_from_archive (archive$path),
        "Assertion on 'path' failed"
    )
    expect_error (
        pkgstats_fns_from_archive (archive$path),
        "path must contain a 'tarballs' directory"
    )

    tarball_path <- setup_test_tarball (archive)

    out <- pkgstats_from_archive (tarball_path)

    expect_s3_class (out, "data.frame")
    expect_equal (nrow (out), 1L)

    out <- pkgstats_fns_from_archive (tarball_path)

    expect_s3_class (out, "data.frame")
    expect_equal (ncol (out), 3L)
    expect_identical (names (out), c ("package", "version", "fn_name"))

    unlink (archive$archive_path, recursive = TRUE)
    if (!test_all) {
        Sys.unsetenv ("PKGSTATS_CRAN_TESTS")
    }
})

test_that ("archive_trawl with save options", {

    if (!test_all) {
        Sys.setenv ("PKGSTATS_CRAN_TESTS" = "true")
    }

    archive <- setup_test_archive ()
    rds_name <- gsub ("\\.tar\\.gz$", ".Rds", archive$tarball)
    tarball_path <- setup_test_tarball (archive)

    tempfiles <-
        basename (fs::dir_ls (fs::path_temp (), type = "file", recurse = TRUE))
    expect_false (rds_name %in% tempfiles)

    out <- pkgstats_from_archive (tarball_path, save_full = TRUE)
    tempfiles <-
        basename (fs::dir_ls (fs::path_temp (), type = "file", recurse = TRUE))
    expect_true (rds_name %in% tempfiles)

    rds_path <- fs::path (fs::path_temp (), rds_name)
    expect_true (fs::file_exists (rds_path))
    dat <- readRDS (rds_path)
    expect_type (dat, "list")
    expect_length (dat, 8L)
    nms <- c (
        "loc",
        "vignettes",
        "data_stats",
        "desc",
        "translations",
        "objects",
        "network",
        "external_calls"
    )
    expect_identical (names (dat), nms)
    fs::file_delete (rds_path)

    out <- pkgstats_from_archive (tarball_path, save_ex_calls = TRUE)
    tempfiles <- basename (fs::dir_ls (fs::path_temp (), type = "file"))
    expect_true (rds_name %in% tempfiles)

    dat <- readRDS (rds_path)
    expect_s3_class (dat, "data.frame")
    nms <- c (
        "tags_line", "call", "tag", "file", "kind", "start", "end", "package"
    )
    expect_equal (names (dat), nms)
    expect_equal (ncol (dat), 8L)
    if (test_all) {
        expect_true (nrow (dat) > 500L)
    }
    fs::file_delete (rds_path)

    unlink (archive$archive_path, recursive = TRUE)
    if (!test_all) {
        Sys.unsetenv ("PKGSTATS_CRAN_TESTS")
    }

})

#' Get all \pkg{roxygen2} blocks from R directory
#'
#' This is used only to get the documentation stats for non-exported functions.
#' @noRd
roxygen_block_stats <- function (path) {

    r_files <- fs::dir_ls (fs::path (path, "R"), regexp = "\\.(R|r|q)$")
    blocks <- lapply (r_files, function (f) {
        suppressMessages (
            tryCatch (
                roxygen2::parse_file (f, env = NULL),
                error = function (e) NULL
            )
        )
    })
    names (blocks) <- r_files
    blocks <- do.call (c, blocks)

    file_names <- gsub ("[0-9]*$", "", names (blocks))
    file_names <- as.character (fs::path_rel (file_names, path))
    fn_names <- vapply (blocks, function (b) {
        as.character (b$call) [2]
    }, character (1L), USE.NAMES = FALSE)

    exported <- vapply (blocks, function (b) {
        length (roxygen2::block_get_tags (b, "export")) > 0L
    }, logical (1L))

    num_doclines <- vapply (blocks, function (b) {
        lines_b <- vapply (b$tags, function (i) i$line, numeric (1L))
        as.integer (diff (range (lines_b)))
    }, integer (1L), USE.NAMES = FALSE)
    num_param_chars <- vapply (blocks, function (b) {
        pars <- roxygen2::block_get_tags (b, "param")
        par_txt <- vapply (pars, function (p) {
            p$val$description
        }, character (1L))
        ret <- roxygen2::block_get_tags (b, "returns")
        ret_txt <- vapply (ret, function (p) {
            p$val
        }, character (1L))

        nchars <- 0
        if (length (par_txt) > 0L || length (ret_txt) > 0L) {
            nchars <- nchar (c (par_txt, ret_txt))
        }
        return (c (md = stats::median (nchars), mn = mean (nchars)))
    }, numeric (2L), USE.NAMES = FALSE)

    data.frame (
        file_name = file_names,
        fn_name = fn_names,
        exported = exported,
        param_nchars_md = num_param_chars [1, ],
        param_nchars_mn = num_param_chars [2, ],
        num_doclines = num_doclines,
        row.names = NULL
    )
}

#' Find average amount of leading white space from a vector of all values of
#' white space.
#'
#' This is done by identifying all local maxima, converting their heights to
#' change relative to preceding local maxima, and picking the first value which
#' is higher than all preceding values.
#' @param x The vector representing the frequency table of leading white space
#' values returned by `cpp_loc`
#' @return A single integer of the average leading white space = indentation
#' @noRd
average_leading_white <- function (x) {

    index <- seq_along (x) [-1]

    delta <- rep (0, length (x))
    delta [1] <- x [1] - x [2]
    delta [index] <- x [index] - x [index - 1] - x [index + 1]
    delta <- c (-1, delta)
    index <- which (delta > 0)
    delta_i <- delta [index]
    delta_ratio <- delta_i [-length (delta_i)] / delta_i [-1]
    # Select first peak which is higher than preceding peak(s):
    i <- which (delta_ratio > 1) [1]

    return (index [i])
}

#' Use internal file-type-dict.R to match files in flist to file types
#' @param flist List of files to be analysed
#' @return Modified version of the \link{file_exts} function containing one row
#' for each file in flist, and including also the path to the file.
#' @noRd
get_file_types <- function (flist) {

    exts <- vapply (
        flist, function (i) {
            utils::tail (strsplit (i, "\\.") [[1]], 1)
        },
        character (1)
    )
    ftypes <- file_exts (exts)

    # any non-parseable files (like .o, .so) then have NA extensions:
    index <- which (!is.na (ftypes$ext))
    ftypes <- ftypes [index, ]
    flist <- flist [index]

    # Remove regex-"^\\s*" from start of single-line comments
    ftypes$cmt <- gsub ("\\^\\\\s\\*", "", ftypes$cmt)
    # And reduce all comment symbols to actual symbols minus
    # regex-formatting:
    ftypes$cmt_open <- gsub ("\\\\", "", ftypes$cmt_open)
    ftypes$cmt_close <- gsub ("\\\\", "", ftypes$cmt_close)
    ftypes$cmt <- gsub ("\\\\", "", ftypes$cmt)

    ftypes$file <- flist
    rownames (ftypes) <- NULL

    return (ftypes)
}

#' Internal calculation of Lines-of-Code Statistics
#'
#' @param path Directory to source code of package being analysed
#' @return A list of statistics for each of three directories, 'R', 'src', and
#' 'inst/include', each one having 5 statistics of total numbers of lines,
#' numbers of empty lines, total numbers of white spaces, total numbers of
#' characters, and indentation used in files in that directory.
#'
#' @note NA values are returned for directories which do not exist.
#' @family stats
#' @export
#' @examples
#' f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
#' # have to extract tarball to call function on source code:
#' path <- extract_tarball (f)
#' loc_stats (path)
loc_stats <- function (path) {

    path <- expand_path (path)

    dirs <- c ("R", "src", "inst", "tests", "vignettes")
    paths <- fs::path (path, dirs)
    paths <- c (
        paths [which (fs::dir_exists (paths))],
        extra_manifest_paths (path)
    )

    flist <- lapply (paths, function (p) {
        paths_p <- fs::dir_ls (p, recurse = TRUE)
        cbind (rep (p, length (paths_p)), paths_p)
    })
    flist <- unname (do.call (rbind, flist))

    # .Rmd files generally excluded except in vignettes:
    index <- grep ("vignettes.*\\.Rmd$", flist [, 2])
    rmd_vignettes <- flist [index, ]
    flist <- flist [which (!grepl (excluded_file_ptn (), flist [, 2])), ]
    flist <- rbind (flist, rmd_vignettes)

    ftypes <- get_file_types (flist [, 2])
    flist_dir <- flist [match (ftypes$file, flist [, 2]), 1]

    s <- cpp_loc (
        ftypes$file,
        ftypes$cmt_open,
        ftypes$cmt_close,
        ftypes$cmt
    )

    nstats <- 8L # number of stats for each file, taken from src/loc.cpp

    index <- seq_along (s) [-(seq (nstats * nrow (ftypes)))]
    leading_white <- s [index [-1]] # rm 1st value
    indentation <- average_leading_white (leading_white)

    index <- seq (nstats * nrow (ftypes))
    s <- data.frame (
        t (matrix (s [index], nrow = nstats)),
        stringsAsFactors = FALSE
    )
    names (s) <- c (
        "nlines",
        "ncode",
        "ndoc",
        "nempty",
        "nspaces",
        "nchars",
        "nbrackets",
        "ntabs"
    )

    s$nbrackets [s$nbrackets < 1] <- NA_integer_
    s$language <- s$dir <- ""
    if (nrow (ftypes) > 0) { # data pkgs may have no code
        s$language <- ftypes$type
        s$dir <- fs::path_rel (flist_dir, path)
    }

    # suprress no visible binding notes:
    language <- nfiles <- nlines <- ncode <- ndoc <-
        nempty <- nspaces <- nchars <- nbrackets <- ntabs <- NULL

    # No magrittr here, plus note final renaming of nbrackets to nexpr
    xg <- dplyr::group_by (s, language, dir)
    s <- dplyr::summarise (
        xg,
        nfiles = length (which (nlines > 0)),
        nlines = sum (nlines),
        ncode = sum (ncode),
        ndoc = sum (ndoc),
        nempty = sum (nempty),
        nspaces = sum (nspaces),
        nchars = sum (nchars),
        nexpr = stats::median (nbrackets, na.rm = TRUE),
        ntabs = sum (ntabs),
        .groups = "keep"
    )
    s$indentation <- indentation

    return (s)
}

setup_test_archive <- function () {
    f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
    tarball <- basename (f)

    archive_path <- fs::path (fs::path_temp (), "archive")
    if (!fs::dir_exists (archive_path)) {
        fs::dir_create (archive_path)
    }
    path <- fs::path (archive_path, tarball)
    if (!fs::file_exists (path)) {
        fs::file_copy (f, path)
    }

    list (tarball = tarball, path = path, archive_path = archive_path)
}

setup_test_tarball <- function (archive) {
    tarball_path <- fs::path (archive$archive_path, "tarballs")
    if (!fs::dir_exists (tarball_path)) {
        fs::dir_create (tarball_path, recurse = TRUE)
    }
    if (!fs::file_exists (fs::path (tarball_path, archive$tarball))) {
        fs::file_copy (archive$path, fs::path (tarball_path, archive$tarball))
    }

    return (tarball_path)
}

#' Trawl a local CRAN archive to extract function names only from all packages
#'
#' @inheritParams pkgstats_from_archive
#' @return A `data.frame` object with one row for each function in each package
#' and the following columns:
#' \itemize{
#' \item Package name
#' \item Package version
#' \item Function name
#' }
#'
#' @family archive
#' @export
pkgstats_fns_from_archive <- function (path,
                                       archive = FALSE,
                                       prev_results = NULL,
                                       results_file = NULL,
                                       chunk_size = 1000L,
                                       num_cores = 1L,
                                       results_path = fs::path_temp ()) {

    requireNamespace ("hms", quietly = TRUE)
    requireNamespace ("parallel", quietly = TRUE)

    if (!grepl ("tarball", path)) {
        if (!fs::dir_exists (fs::path (path, "tarballs"))) {
            stop ("path must contain a 'tarballs' directory")
        }
        path <- fs::path (path, "tarballs")
    }

    if (basename (path) != "tarballs") {
        stop ("path must be a directory named 'tarballs'")
    }

    if (!fs::dir_exists (path)) {
        stop ("[", path, "] directory does not exist")
    }

    res <- NULL
    out <- prev_results

    flist <- fs::dir_ls (
        path,
        recurse = archive,
        regexpr = "\\.tar\\.gz$"
    )
    flist <- expand_path (flist)
    flist <- rm_prev_files (flist, prev_results)
    nfiles <- length (flist)

    if (nfiles > 0) {

        n <- ceiling (nfiles / chunk_size)
        n <- factor (rep (seq (n), each = chunk_size)) [seq (nfiles)]
        flist <- split (flist, f = n)

        message (
            "Starting trawl of ", nfiles,
            " files in ", length (flist), " chunks"
        )

        results_path <- expand_path (results_path)
        if (!fs::dir_exists (results_path)) {
            fs::dir_create (results_path)
        }
        results_files <- NULL

        index <- 1 # name of temporary files
        pt0 <- proc.time ()

        for (f in flist) {

            res <- parallel::mclapply (f, function (i) {

                tryCatch (
                    pkgstats::pkgstats_fn_names (i),
                    error = function (e) NULL
                )

            }, mc.cores = num_cores)

            fname <- fs::path (
                results_path,
                paste0 ("pkgstats-fn-names-results-", index, ".Rds")
            )
            saveRDS (do.call (rbind, res), fname)
            results_files <- c (results_files, fname)

            prog <- index * chunk_size / nfiles
            prog_fmt <- format (100 * prog, digits = 2)
            pt1 <- as.integer ((proc.time () - pt0) [3])
            t_per_file <- pt1 / (index * chunk_size)
            t_total <- t_per_file * nfiles
            t_rem <- hms::hms (t_total - pt1)

            ndone <- min (c (nfiles, index * chunk_size))

            message (
                "[", ndone, " / ", nfiles,
                "]  = ", prog_fmt, "%; (elapsed, remaining) = (",
                pt1, ", ", t_rem, ")"
            )

            index <- index + 1
        }

        res <- do.call (rbind, lapply (results_files, readRDS))
    }

    t_total <- hms::hms (as.integer (proc.time () [3] - pt0 [3]))
    message ("\nTotal time = ", t_total)

    out <- rbind (out, res)
    rownames (out) <- NULL

    chk <- fs::file_delete (results_files) # nolint

    if (!is.null (out) && !is.null (results_file)) {

        out <- rbind (out, base_rmd_fns_df ())
        out <- out [which (!duplicated (out)), ]

        if (!grepl (.Platform$file.sep, results_file)) {
            results_file <- fs::path (".", results_file)
        }
        results_file <- expand_path (results_file)

        results_path <- gsub (
            basename (results_file), "",
            results_file
        )
        results_path <- expand_path (results_path)
        if (!fs::dir_exists (results_path)) {
            stop ("Directory [", results_path, "] does not exist")
        }

        results_file <- basename (results_file)
        results_file <- tools::file_path_sans_ext (results_file)
        results_file <- fs::path (
            results_path,
            paste0 (results_file, ".Rds")
        )

        saveRDS (out, results_file)
    }

    invisible (out)
}

#' Update function names data from previous data and newly updated CRAN
#' packages only.
#'
#' @inheritParams pkgstats_from_archive
#' @return A `data.frame` object with one row for each function in each package
#' and the following columns:
#' \itemize{
#' \item Package name
#' \item Package version
#' \item Function name
#' }
#'
#' @family archive
#' @export

# nocov start
pkgstats_fns_update <- function (prev_results = NULL,
                                 results_file = NULL,
                                 chunk_size = 1000L,
                                 num_cores = 1L,
                                 results_path = tempdir ()) {

    requireNamespace ("hms", quietly = TRUE)
    requireNamespace ("parallel", quietly = TRUE)

    checkmate::assert_int (chunk_size, lower = 1L)
    checkmate::assert_int (num_cores, lower = 1L)
    checkmate::assert_string (results_path)
    check_prev_results (prev_results)

    res <- NULL
    out <- prev_results

    new_cran_pkgs <- list_new_cran_updates (prev_results)

    npkgs <- length (new_cran_pkgs)

    if (npkgs > 0) {

        n <- ceiling (npkgs / chunk_size)
        n <- factor (rep (seq (n), each = chunk_size)) [seq (npkgs)]
        new_cran_pkgs <- split (new_cran_pkgs, f = n)

        message (
            "Starting trawl of ", npkgs,
            " files in ", length (new_cran_pkgs), " chunks"
        )

        results_path <- expand_path (results_path)
        if (!fs::dir_exists (results_path)) {
            fs::dir_create (results_path, recurse = TRUE)
        }

        index <- 1 # name of temporary files
        pt0 <- proc.time ()

        for (f in new_cran_pkgs) {

            if (num_cores > 1L) {

                res <- parallel::mclapply (f, function (i) {

                    tryCatch (
                        pkgstats::pkgstats_fn_names (i),
                        error = function (e) NULL
                    )

                }, mc.cores = num_cores)

            } else {

                res <- lapply (f, function (i) {

                    tryCatch (
                        pkgstats::pkgstats_fn_names (i),
                        error = function (e) NULL
                    )
                })
            }

            fname <- fs::path (
                results_path,
                paste0 ("pkgstats-results-", index, ".Rds")
            )
            saveRDS (do.call (rbind, res), fname)
            results_files <- c (results_files, fname)

            archive_trawl_progress_message (index, chunk_size, npkgs, pt0)
            index <- index + 1
        }

        res <- do.call (rbind, lapply (results_files, readRDS))
    }

    out <- rbind (out, res)

    tarball <- paste0 (out$package, "_", out$version)
    cran_pkgs <- get_cran_db ()
    cran_tarball <- paste0 (cran_pkgs$Package, "_", cran_pkgs$Version)
    index <- which (tarball %in% cran_tarball)
    out <- out [index, ]

    if (!is.null (res) && !is.null (results_file)) {

        out <- rbind (out, base_rmd_fns_df ())
        index <- which (duplicated (out))
        if (length (index) > 0L) {
            out <- out [-(index), ]
        }

        results_file <- archive_results_file_name (results_file)
        saveRDS (out, results_file)
    }

    invisible (out)
}

base_rmd_fns_df <- function () {

    pkgs <- c (
        list_base_recommend_pkgs (base = TRUE),
        list_base_recommend_pkgs (base = FALSE)
    )
    fns <- do.call (rbind, lapply (pkgs, list_recommended_pkg_fns))

    pkgs <- pkgs [which (!pkgs %in% fns$pkg)]

    fns_extra <- do.call (rbind, lapply (pkgs, function (p) {
        f <- ls (asNamespace (p))
        v <- as.character (utils::packageVersion (p))
        data.frame (
            pkg = rep (p, length (f)),
            vers = rep (v, length (f)),
            fn = f
        )
    }))

    fns <- rbind (fns_extra, fns)

    data.frame (
        package = fns$pkg,
        version = fns$vers,
        fn_name = fns$fn
    )
}
# nocov end

get_Rd_metadata <- utils::getFromNamespace (".Rd_get_metadata", "tools") # nolint

#' Check path is an existing root directory of an R package
#'
#' Uses `rprojroot::is_r_package` criterion:
#' "contains a file "DESCRIPTION" with contents matching "^Package: "
#' @noRd
check_path <- function (path, max_subdirs = 2L) {

    path <- fs::path_real (path)

    if (grepl ("\\.tar\\.gz$", path)) {

        checkmate::assert_file_exists (path)

    } else {

        checkmate::assert_directory_exists (path)

        proj_root <- tryCatch (
            rprojroot::find_package_root_file (path = path),
            error = function (e) NULL
        )
        if (is.null (proj_root)) {
            subdirs <- fs::dir_ls (path, type = "directory", recurse = max_subdirs)
            proj_root <- lapply (subdirs, function (d) {
                tryCatch (
                    rprojroot::find_root (rprojroot::is_r_package, path = d),
                    error = function (e) NULL
                )
            })
            proj_root <- unique (unlist (unname (proj_root)))
        }

        if (length (proj_root) != 1L) {
            stop (
                "Could not find unambiguous project root from {proj_root}",
                call. = FALSE
            )
        }

        path <- proj_root
    }

    return (path)
}

#' Expand path
#'
#' The `fs::path_real()` function does the same direct expansion, but requires
#' that path to exist. This version expands both "~" and "." without requiring
#' that full paths necessarily exist. 'fs' has a `path_expand()` function which
#' expands "~" only, but not ".". This function reverses the name of that one,
#' to avoid confusion.
#'
#' @param path Vector of one or more paths
#' @return Expanded, absolute versions of input argument.
#' @noRd
expand_path <- function (path) {
    fs::path_abs (fs::path_expand (path))
}

#' Decompose file paths into character vectors of named directories and final
#' file names
#'
#' @param f One of more file paths with system-dependent file separators
#' @return List of equivalent character vectors from which paths can be
#' reconstructed with \link{file.path}
#' @noRd
decompose_path <- function (f) {

    # https://github.com/r-lib/fs/blob/4cc4b56c26b9d7f177a676fbb331133bb2584b86/R/path.R # nolint
    strsplit (f, "^(?=/)(?!//)|(?<!^)(?<!^/)/", perl = TRUE)
}

#' Error-controlled version of parse
#'
#' `parse` can fail even with enforced encoding because of multi-byte
#' characters. This is a hacky work-around to find and simply strip any such
#' characters, on the assumption that these can't actually be part of any
#' language expressions.
#'
#' @param file name of R source file
#' @return contents of `parse(file)`
#' @noRd
control_parse <- function (file) {

    # use "latin1" encoding to force re-coding of any non-latin characters:
    x <- readr::read_lines (file,
        locale = readr::locale (encoding = "latin1"),
        progress = FALSE
    )

    # note: keep.source must be TRUE as it is, for example, switched off in
    # `rmarkdown` environments, which means no parse data are returned by
    # getParseData.
    out <- tryCatch (parse (text = x, keep.source = TRUE, encoding = "UTF-8"),
        error = function (e) e
    )

    count <- 0
    nchars0 <- sum (nchar (x))
    while (methods::is (out, "simpleError") &&
        count < floor (length (x) / 10)) {

        g <- gregexpr ("\'.*\'", out$message)
        ptn <- gsub ("\'", "", regmatches (out$message, g) [[1]])
        if (length (ptn) == 0L) {
            count <- length (x)
            next
        }

        if (substring (ptn, 1, 1) == "\\" && substring (ptn, 2, 2) != "\\") {
            ptn <- paste0 ("\\", ptn)
        }

        # Then find line to implement substitution:
        g <- gregexpr ("\\\"", out$message) [[1]]
        expr <- gsub ("\"", "", substring (out$message, min (g), max (g)))
        if (grepl ("\\\\", expr)) {
            expr <- gsub ("\\", "\\\\", expr, fixed = TRUE)
        }
        index <- which (grepl (ptn, x, fixed = TRUE) &
            grepl (expr, x, fixed = TRUE))
        x [index] <- gsub (ptn, "", x [index], fixed = TRUE)

        nchars <- sum (nchar (x))

        out <- tryCatch (
            parse (text = x, keep.source = TRUE, encoding = "UTF-8"),
            error = function (e) e
        )

        if (nchars == nchars0) {
            count <- length (x)
        } else {
            count <- count + 1L
            nchars0 <- nchars
        }
    }

    return (out)
}

#' Regex to remove select kinds of files from analyses
#'
#' These file types are excluded from `loc_stats`, as well as from both `ctags`
#' and `gtags` analyses.
#' @return regex pattern
#' @noRd
excluded_file_ptn <- function () {

    exts <- c (
        "h", "rda", "rds", "Rd", "md", "Rmd", "win", "min.js",
        "png", "svg", "jpg", "gif", "json", "geojson"
    )
    paste0 ("(", paste0 ("\\.", exts, collapse = "|"), ")$")
}

extra_manifest_paths <- function (path) {

    path_src <- fs::path (path, "src")
    if (!fs::dir_exists (path_src)) {
        return (NULL)
    }
    f_manifest <- fs::dir_ls (
        path_src,
        type = "file",
        regexp = "source.*manifest",
        ignore.case = TRUE
    )
    if (length (f_manifest) == 0L) {
        return (NULL)
    }
    f_ext <- fs::path_ext (f_manifest)
    manifest_parse_fn <- paste0 ("parse_manifest_", f_ext)
    pkg_fns <- ls (envir = asNamespace ("pkgstats"), all.names = TRUE)
    index <- which (manifest_parse_fn %in% pkg_fns)
    if (length (index) == 0L) {
        return (NULL)
    }
    # If multiple versions, parse only first
    x <- do.call (manifest_parse_fn [1], list (f_manifest [1]))
    if (!"vendor_sources" %in% names (x)) {
        return (NULL)
    }
    vendor_sources <- vapply (x$vendor_sources, function (v) {
        fs::path_abs (fs::path (path_src, v))
    }, character (1L))
    vendor_sources <- vendor_sources [which (fs::dir_exists (vendor_sources))]

    return (vendor_sources)
}

parse_manifest_toml <- function (f) {
    requireNamespace ("toml", quietly = TRUE)
    toml::parse_toml (readLines (f))
}

parse_manifest_json <- function (f) {
    requireNamespace ("jsonlite", quietly = TRUE)
    jsonlite::read_json (f)
}

which_unix <- function () {

    if (!.Platform [["OS.type"]] == "unix") {
        return (NULL)
    }

    x <- utils::capture.output (
        sys::exec_wait ("lsb_release", args = "-a", std_out = TRUE)
    )
    strsplit (grep ("^Distributor ID", x, value = TRUE), "\\t") [[1]] [2]
}

rm_file_no_err <- function (f) {

    tryCatch (
        file.remove (f),
        error = function (e) NULL
    )
}

sys_is_linux <- function () {
    tolower (Sys.info () [["sysname"]]) == "linux"
}

get_processes <- function () {

    if (!sys_is_linux ()) {
        return ()
    }

    p <- system ("ps aux", intern = TRUE)
    # It's fixed-width as defined by header:
    index <- gregexpr ("[^[:space:]]", p [1]) [[1]]
    index <- index [-(which (diff (index) == 1) + 1)]
    index <- cbind (index, c (index [-1] - 1, nchar (p [1])))

    p <- apply (index, 1, function (i) substring (p, i [1], i [2]))
    p <- data.frame (p, stringsAsFactors = FALSE)
    names (p) <- p [1, ]
    p <- p [-1, ]

    return (p)
}

wait_for_process <- function (what = "ctags") {

    if (!sys_is_linux ()) {
        Sys.sleep (0.2)
        return ()
    }

    p <- get_processes ()
    count <- 1
    while (any (grep (what, p$COMMAND))) {
        Sys.sleep (0.2)
        p <- get_processes ()
        count <- count + 1
        if (count > 10) {
            break
        }
    }
}

get_num_vignettes <- function (path) {

    nv <- 0L

    path_ls <- basename (fs::dir_ls (path))

    if ("build" %in% path_ls) {

        flist <- fs::dir_ls (fs::path (path, "build"),
            full.names = TRUE
        )
        vfile <- grep ("vignette", flist, value = TRUE)
        if (length (vfile) > 0) {
            nv <- nrow (readRDS (vfile [1]))
        }
    } else if ("vignettes" %in% path_ls) {
        nv <- length (fs::dir_ls (
            fs::path (path, "vignettes"),
            regexp = "\\.[rRq]md$"
        ))
        if (nv == 0L) {
            nv <- length (fs::dir_ls (
                fs::path (path, "vignettes"),
                regexp = "\\.[rR]$"
            ))
        }
    }

    return (nv)
}

get_num_demos <- function (path) {

    nd <- 0L

    path_ls <- basename (fs::dir_ls (path))

    if ("demo" %in% path_ls) {

        dindex <- fs::path (path, "demo", "00Index")
        nd <- length (brio::read_lines (dindex))
    }

    return (nd)
}

get_data_stats <- function (path) {

    nd <- c (n = 0L, total_size = 0L, median_size = 0L)

    path_ls <- basename (fs::dir_ls (path))

    if ("data" %in% path_ls) {

        flist <- fs::dir_ls (fs::path (path, "data"),
            regexp = "\\.rd",
            ignore.case = TRUE
        )
        sizes <- fs::file_info (flist)$size

        nd [1] <- length (flist)
        nd [2] <- as.integer (sum (sizes))
        nd [3] <- as.integer (stats::median (sizes))
    }

    return (nd)
}

# https://cran.r-project.org/doc/manuals/R-exts.html#Preparing-translations
# https://cran.r-project.org/doc/manuals/R-admin.html#Localization-of-messages
# https://developer.r-project.org/TranslationTeams.html
get_translations <- function (path) {

    ll <- NA_character_

    po_dir <- expand_path (fs::path (path, "po"))

    if (fs::dir_exists (po_dir)) {

        ll <- gsub (
            "^R-|\\.po$", "",
            fs::dir_ls (po_dir, regexp = "\\.po$")
        )
        # pkgs may generate translations without having any, and then will only
        # have '.pot' files with no usable translation fields.
        if (length (ll) == 0L) {
            ll <- NA_character_
        }
    }

    return (unique (ll))
}

#' Extract tarball of a package into temp directory and return path to extracted
#' package
#'
#' @param tarball Full path to local tarball of an R package.
#' @param exdir Directory into which tarballs are to be extracted.
#' @return Path to extracted version of package (in `tempdir()`).
#'
#' @family misc
#' @export
#' @examples
#' f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
#' path <- extract_tarball (f)
extract_tarball <- function (tarball, exdir = fs::path_temp ()) {

    if (!fs::file_exists (tarball)) {
        stop ("file [", tarball, "] does not exist")
    }
    if (!checkmate::testCharacter (tarball,
        len = 1,
        pattern = "\\.tar\\.gz$"
    )) {
        stop (paste0 (
            "tarball must be a single character ",
            "specifying path to .tar.gz file"
        ))
    }

    flist <- utils::untar (tarball,
        exdir = exdir,
        list = TRUE,
        tar = "internal"
    )
    if (utils::untar (tarball, exdir = exdir, tar = "internal") != 0) {
        stop ("Unable to extract tarball to 'tempdir'")
    }

    fdir <- vapply (flist, function (i) {
        strsplit (i, .Platform$file.sep) [[1]] [1]
    },
    character (1),
    USE.NAMES = FALSE
    )
    fdir <- names (table (fdir)) [1]
    path <- fs::path_real (fs::path (exdir, fdir))

    chk <- rename_files_in_r (path)
    if (!chk) {
        warning ("Files in .R directory unable to be re-named")
    }

    return (path)
}

#' files in './R' directory may have suffixes of .q, .r, or .s
#' https://cran.r-project.org/doc/manuals/R-exts.html#Package-subdirectories
#' But parsers + cloc only recognise .R/.r, as does gtags, so rename any others
#' Discovered via BDR's
#' https://cran.r-project.org/web/packages/boot/index.html
#' The `rms` package also has loads of '.s' code, which cloc's dictionary
#' identifies as assembly
#' @noRd
rename_files_in_r <- function (path) {

    path <- fs::path_real (path)
    path_r <- fs::path (path, "R")
    if (!fs::dir_exists (path_r)) {
        return (FALSE)
    }
    f_sq <- fs::dir_ls (path_r, regexp = "\\.(s|S|q)$")

    if (length (f_sq) > 0) {

        f_r <- fs::path_ext_set (f_sq, "R")
        fs::file_move (f_sq, f_r)
    }

    return (TRUE)
}

#' Map calls from within each function to external packages
#'
#' @param path Path to package being analysed
#' @param tags_r Result of `get_ctags(dir = "R")`
#' @param pkg_name Name of package
#' @return Network of calls to external functions
#' @noRd
external_call_network <- function (tags_r, path, pkg_name) {

    calls <- extract_call_content (tags_r)

    if (length (calls) == 0L || nrow (calls) == 0L) {
        return (NULL)
    }

    calls$kind <- tags_r$kind [calls$tags_line]
    calls$start <- tags_r$start [calls$tags_line]
    calls$end <- tags_r$end [calls$tags_line]

    calls$package <- NA_character_
    pkg_fns <- unique (tags_r$tag [tags_r$kind == "function"])
    pkg_fns <- pkg_fns [which (!grepl ("^anonFunc", pkg_fns))]
    calls$package [which (calls$call %in% pkg_fns)] <- pkg_name

    calls <- add_base_recommended_pkgs (calls)

    # Calls namespaced with `::`:
    index <- grep ("::", calls$call)
    calls$package [index] <- vapply (calls$call [index], function (i) {
        strsplit (i, "::") [[1]] [1]
    },
    character (1),
    USE.NAMES = FALSE
    )
    calls$call [index] <- gsub ("^.*::", "", calls$call [index])

    calls <- add_other_pkgs_to_calls (calls, path)

    calls <- calls [which (!is.na (calls$package)), ]

    rownames (calls) <- NULL

    return (calls)
}

extract_call_content <- function (tags_r) {

    content <- gsub ("\\\"$", "", tags_r$content)
    # Remove everything within quotes - this presumes only single quotations in
    # each line, which is almost always the case.
    content <- gsub ("\\\".*\\\"", "", content)
    content <- gsub ("\\\'.*\\\'", "", content)
    index <- which (tags_r$kind == "function")

    # Remove everything within bracets of function definitions.
    br_content <- regmatches (
        content [index],
        regexpr ("\\([^\\)]+\\)", content [index])
    )
    content [index] <- vapply (
        seq_along (index), function (i) {
            gsub (br_content [i], "()",
                content [index [i]],
                fixed = TRUE
            )
        },
        character (1L)
    )

    # Then convert all symbols which are not allowed in function names to
    # spaces:
    syms <- c (
        "\\(", "\\)",
        "\\[", "\\]",
        "\\{", "\\}",
        "<\\-", "\\=",
        "\\+", "\\/", "\\*", "\\-",
        "\\|", "&", "&&",
        "\\^", "<", ">",
        ",", ";", "\\~"
    )
    content <- gsub (paste0 (syms, collapse = "|"), " ", content)

    # Then remove ...
    # isolated "$" symbols:
    content <- gsub ("^\\$|\\$$|\\s\\$\\s", "", content)
    content <- gsub ("\\b[0-9]", "", content)
    # anything after comments
    content <- gsub ("\\#.*$", "", content)
    content [which (is.na (content))] <- ""

    # Then split all around space to obtain call references
    calls <- strsplit (content, "\\s+")
    calls <- lapply (seq_along (calls), function (i) {
        if (length (calls [[i]]) > 0L) {
            cbind (
                rep (i, length (calls [[i]])),
                calls [[i]]
            )
        }
    })
    calls <- do.call (rbind, calls)
    calls <- calls [which (!calls [, 2] == ""), , drop = FALSE]

    if (length (calls) == 0L) {
        return (NULL)
    }

    calls <- data.frame (
        tags_line = as.integer (calls [, 1]),
        call = gsub ("^[[:punct:]]+", "", calls [, 2]),
        stringsAsFactors = FALSE
    )

    rm_these <- c (
        "", "TRUE", "FALSE", "NULL",
        "NA", "...", "\\", "Inf", ":", ".",
        "function"
    )
    calls <- calls [which (!calls$call %in% rm_these), , drop = FALSE]
    calls <- calls [which (!grepl ("\\$$|^\"|^\'", calls$call)), , drop = FALSE]
    calls$tag <- tags_r$tag [calls$tags_line]
    calls$file <- tags_r$file [calls$tags_line]

    # rm global variables
    globals <- unique (tags_r$tag [which (tags_r$kind == "globalVar")])
    calls <- calls [which (!calls$call %in% globals), , drop = FALSE]

    # Finally remove any functionVars (internal variables). This requires
    # matching them to the calling environment, so only those which exist in
    # each environment (function) are removed.
    fn_vars <- tags_r [which (tags_r$kind == "functionVar"), ]
    # Then make lists of fn_vars for each function
    fns <- tags_r [which (tags_r$kind == "function"), ]
    fns <- fns [which (!grepl ("^anonFunc", fns$tag)), ]

    fn_lines <- apply (
        fns [, c ("tag", "file", "start", "end")], 1,
        function (i) cbind (i [1], i [2], seq (i [3], i [4])),
        simplify = FALSE
    )

    if (length (fn_lines) == 0L) { # There are no function calls

        return (NULL)

    } else if (is.list (fn_lines)) {

        fn_lines <- do.call (rbind, fn_lines)

    } else if (ncol (fn_lines) == 1L) {

        fn_lines <- matrix (fn_lines, ncol = 3)
    }

    fn_lines <- data.frame (
        fn_name = fn_lines [, 1],
        file = fn_lines [, 2],
        lines = as.integer (fn_lines [, 3]),
        stringsAsFactors = FALSE
    )

    for (f in unique (fn_lines$fn_name)) {

        these_lines <- fn_lines$lines [fn_lines$fn_name == f]
        this_file <- fn_lines$file [these_lines [1]]
        these_vars <- fn_vars$tag [fn_vars$file == this_file &
            fn_vars$start %in% these_lines]

        index <- which (calls$tags_line %in% these_lines &
            calls$file == this_file)
        calls$call [index] [calls$call [index] %in% these_vars] <- NA_character_
    }

    calls <- calls [which (!is.na (calls$call)), ]

    calls$call <- gsub ("(\\s*\\(|\\s*)$", "", calls$call)

    rownames (calls) <- NULL

    return (calls)
}

list_base_recommend_pkgs <- function (base = TRUE) {

    if (base) {
        # namespaces of base packages are loaded, so fns can be grabbed directly
        pkgs <- c (
            "base", "stats", "graphics", "grDevices",
            "utils", "datasets", "methods"
        )
    } else {
        # recommended pkgs, which can not be (assumed to be) loaded. This
        # list from https://cran.r-project.org/src/contrib/4.1.0/Recommended/
        pkgs <- c (
            "boot",
            "class",
            "cluster",
            "codetools",
            "foreign",
            "KernSmooth",
            "lattice",
            "MASS",
            "Matrix",
            "mgcv",
            "nlme",
            "nnet",
            "rpart",
            "spatial",
            "survival"
        )
        is_installed <- vapply (pkgs, function (i) {
            tryCatch (
                find.package (i),
                error = function (e) ""
            )}, character (1L))
        pkgs <- pkgs [which (nzchar (is_installed))]
    }

    return (pkgs)
}

list_recommended_pkg_fns <- function (pkg_name) {

    df0 <- data.frame (
        pkg = character (0L),
        vers = character (0L),
        fn = character (0L)
    )
    ll <- .libPaths ()

    index <- vapply (ll, function (p) {
        fs::dir_exists (fs::path (p, pkg_name))
    }, logical (1L))
    if (!any (index)) {
        return (df0)
    }
    rpath <- fs::path (ll [which (index) [1]], pkg_name)
    if (!fs::dir_exists (rpath)) {
        return (df0)
    }

    f <- fs::path (rpath, "DESCRIPTION")
    if (!fs::file_exists (f)) {
        return (df0)
    }
    pkg_version <- read.dcf (f) [1, "Version"]

    f <- fs::path (rpath, "NAMESPACE")
    if (!fs::file_exists (f)) {
        return (df0)
    }

    n <- brio::read_lines (f)
    fns <- parse (text = n)
    fns <- gsub (
        "export\\s?\\(|\\)$", "",
        grep ("export", fns, value = TRUE)
    )
    fns <- do.call (c, strsplit (fns, ","))
    fns <- gsub ("^\\n\\s*|^\\s*", "", fns)

    data.frame (
        pkg = rep (pkg_name, length (fns)),
        vers = rep (pkg_version, length (fns)),
        fn = fns,
        stringsAsFactors = FALSE
    )
}

#' Use ctags to get list of fns from each pkg
#'
#' @param calls `data.frame` of calls constructed in `external_call_network`.
#' @return Input with additional column of "package" identifying packages
#' associated with each fucntion
#' @noRd
add_base_recommended_pkgs <- function (calls) {

    base_pkgs <- list_base_recommend_pkgs (base = TRUE)
    for (b in base_pkgs) {
        f <- ls (paste0 ("package:", b))
        calls$package [calls$call %in% f & is.na (calls$package)] <- b
    }

    rcmds <- list_base_recommend_pkgs (base = FALSE)
    pkg_calls <- do.call (rbind, lapply (rcmds, list_recommended_pkg_fns))

    if (length (pkg_calls) > 0L) {

        na_calls <- which (is.na (calls$package))
        index <- match (calls$call [na_calls], pkg_calls$fn [na_calls])
        calls$package [na_calls] <- pkg_calls$pkg [index]
    }

    return (calls)
}

#' Add names of depends or imports packages to calls
#'
#' @inheritParams add_base_recommended_pkgs
#' @noRd
add_other_pkgs_to_calls <- function (calls, path) {

    f <- fs::path (path, "NAMESPACE")
    if (!fs::file_exists (f)) {
        return (calls)
    }

    n <- tryCatch (
        parse (text = brio::read_lines (f)),
        error = function (e) NULL
    )
    imports <- gsub (
        "^importFrom\\s?\\(|\\)$", "",
        grep ("^importFrom", n, value = TRUE)
    )

    imports <- strsplit (imports, ",")
    imports <- lapply (imports, function (i) {
        pkg <- gsub ("\\s*\\\"\\s*", "", i [1])
        fns <- gsub ("^\\s*|\\\"|\\s*$", "", i [-1])
        data.frame (
            pkg = rep (pkg, length (fns)),
            fn = fns,
            stringsAsFactors = FALSE
        )    })
    imports <- do.call (rbind, imports)

    if (!is.null (imports)) {

        na_calls <- which (is.na (calls$package))
        calls$package [na_calls] <- imports$pkg [match (
            calls$call [na_calls],
            imports$fn
        )]
    }

    # Finally, manually catch any imported package calls which are not tagged by
    # ctags - this can happen for example if calls are buried inside things like
    # `tryCatch`, or if they don't form full expressions. This relies on
    # getParseData, which is less reliable than ctags, but can catch some
    # outliers.
    pkgs_not_called <- imports$pkg [which (!imports$pkg %in% calls$package)]
    if (length (pkgs_not_called) == 0L) {
        return (calls)
    }

    imports_not_called <- imports [imports$pkg %in% pkgs_not_called, ]

    r_files <- expand_path (fs::dir_ls (
        fs::path (path, "R"),
        regexp = "\\.(r|R|q|s|S)$"
    ))
    all_calls <- paste0 (imports_not_called$fn, collapse = "|")
    tokens <- c ("FUNCTION", "SYMBOL_FUNCTION_CALL", "SPECIAL")

    external_calls <- lapply (r_files, function (i) {

        pd <- utils::getParseData (control_parse (i))
        if (is.null (pd)) {
            return (NULL)
        }
        pd <- pd [grep (all_calls, pd$text), ]
        pd <- pd [which (pd$token %in% tokens), ]

        if (nrow (pd) == 0L) {
            return (NULL)
        }

        data.frame (
            tags_line = pd$text,
            call = pd$text,
            tag = pd$token,
            file = paste0 ("R/", basename (i)),
            kind = "unknown",
            start = pd$line1,
            end = pd$line2
        )
    })
    external_calls <- do.call (rbind, external_calls)

    if (is.null (external_calls)) {
        return (calls)
    }

    if (nrow (external_calls) == 0L) {
        return (calls)
    }

    external_calls$package <-
        imports_not_called$pkg [match (
            external_calls$call,
            imports_not_called$fn
        )]

    external_calls$tags_line <- seq (nrow (external_calls)) +
        max (calls$tags_line, na.rm = TRUE)

    calls <- rbind (calls, external_calls)

    return (calls)
}

# nocov start
has_gtags <- function () {

    gtags_path <- dirname (Sys.which ("gtags"))
    nzchar (gtags_path)
}

#' Install gtags on a unix system (currently only Ubuntu + Arch).
#' @param sudo Set to `FALSE` if `sudo` is not available.
#' @noRd
gtags_install <- function (sudo = TRUE) {

    if (has_gtags ()) {
        return (NULL)
    }

    if (!.Platform$OS.type == "unix") {
        return (NULL)
    }

    f <- fs::file_temp (pattern = "gtags-install-", ext = ".txt")

    unix <- which_unix ()

    if (sudo) {

        cmd <- "sudo"
        arg <- c ("apt-get", "install", "-y")
        if (unix == "Arch") {
            arg <- c ("pacman", "-Syu")
        }

        sys::exec_wait (cmd, args = c (arg, "global"), std_out = f)
    } else {
        gtags_compile ()
    }
}

gtags_compile <- function () {

    u <- "https://ftp.gnu.org/pub/gnu/global/global-6.6.8.tar.gz"
    f <- fs::path (fs::path_temp (), basename (u))
    utils::download.file (u, f)
    utils::untar (f, exdir = fs::path_temp ())

    home <- fs::path_expand ("~")
    conf_args <- c ("--prefix", home, "--disable-gtagscscope")
    withr::with_dir (fs::path (fs::path_temp (), "global-6.6.8"), {
        sys::exec_wait ("./configure", args = conf_args)
        sys::exec_wait ("make")
        sys::exec_wait ("make", args = "install")
        sys::exec_wait ("hash", args = c (
            "-p",
            fs::path (home, "bin", "ctags"),
            "ctags"
        ))
    })
}
# nocov end

#' Extract names of all functions for one R package
#'
#' @inheritParams pkgstats
#' @return A `data.frame` with three columns:
#' \itemize{
#' \item package: Name of package
#' \item version: Package version
#' \item fn_name: Name of function
#' }
#' @family misc
#' @export
#' @examples
#' # 'path' can be path to a package tarball:
#' f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
#' path <- extract_tarball (f)
#' s <- pkgstats_fn_names (path)
pkgstats_fn_names <- function (path) {

    path <- check_path (path)

    nmsp <- get_namespace_contents (path)

    desc_path <- get_desc_path (path)

    if (any (grepl ("^exportPattern", nmsp))) {

        fns <- aliases_from_rd (path, nmsp)

    } else if (!any (grepl ("^export", nmsp))) {

        # no functions exported
        return (data.frame (
            package = character (0L),
            version = character (0L),
            fn_name = character (0L),
            stringsAsFactors = FALSE
        ))

    } else {

        fns <- grep ("^export\\s?\\(", nmsp, value = TRUE)
        fns <- gsub ("^export\\s?\\(|\\)\\s?$", "", fns)
        fns <- gsub ("\\#.*$|\\\t", "", fns)
        fns <- unlist (strsplit (fns, ","))
        fns <- gsub ("^\\s*|\\s*$", "", fns)

        # same grep as for aliases_from_rd below:
        index <- grep ("method(s?|,?)$|class$|<\\-|\\[\\[|,|\\s", fns)
        if (length (index) > 0) {
            fns <- fns [-index]
        }
    }

    # Then get imports to remove re-exported fns:
    imps <- grep ("^importFrom", nmsp, value = TRUE)
    imps <- gsub ("^importFrom\\(|\\)\\s?$", "", imps)
    imps <- gsub ("^[^,]*,", "", imps)
    imps <- gsub (
        "^\\s*|\\s*$", "",
        unlist (strsplit (imps, ","))
    )

    fns <- fns [which (!fns %in% imps)]
    if (!is.null (fns)) {
        fns <- noquote (fns)
    }

    pkg <- get_pkg_name_version (desc_path)

    data.frame (
        package = rep (pkg [1], length (fns)),
        version = rep (pkg [2], length (fns)),
        fn_name = gsub ("^\\\"|\\\"$", "", fns),
        stringsAsFactors = FALSE
    )
}

get_namespace_contents <- function (path) {

    tarball <- grepl ("\\.tar\\.gz$", path)

    if (tarball) {

        flist <- utils::untar (
            path,
            list = TRUE,
            tar = "internal"
        )

        nmsp <- grep ("NAMESPACE", flist, value = TRUE)
        # Some archived pkgs do not have NAMESPACE files; these are not
        # processed here:
        if (length (nmsp) == 0L) {
            return (NULL)
        }

        chk <- utils::untar (
            path,
            files = nmsp,
            exdir = fs::path_temp ()
        )

        if (chk != 0) {
            stop ("Unable to extract tarball [", tarball, "]",
                call. = FALSE
            )
        }

        nmsp <- fs::path (fs::path_temp (), nmsp)

        on.exit ({
            fs::dir_delete (fs::path_dir (nmsp))
        })

    } else {

        nmsp <- fs::dir_ls (
            path,
            recurse = TRUE,
            regexp = "NAMESPACE"
        )
        nmsp <- expand_path (nmsp [1])

    }

    # See R source in src/library/base/R/namespace.R for reference, especially
    # the `parseNamespaceFile()` function.
    nmsp_parsed <- parse (
        nmsp,
        keep.source = FALSE,
        srcfile = NULL
    )

    return (nmsp_parsed)
}

get_desc_path <- function (path) {

    tarball <- grepl ("\\.tar\\.gz$", path)

    if (tarball) {

        flist <- utils::untar (
            path,
            exdir = fs::path_temp (),
            list = TRUE,
            tar = "internal"
        )

        desc <- grep ("DESCRIPTION", flist, value = TRUE)
        chk <- utils::untar (
            path,
            files = desc,
            exdir = fs::path_temp ()
        )
        desc <- expand_path (fs::path (fs::path_temp (), desc))

    } else {

        desc <- fs::dir_ls (
            path,
            recurse = TRUE,
            regexp = "DESCRIPTION"
        )
        desc <- expand_path (desc [1])
    }

    return (desc)
}

#' List all functions defined in R source code which match 'exportPattern'
#'
#' @noRd
exp_ptn_r_fn_names <- function (path, nmsp) {

    tarball <- grepl ("\\.tar\\.gz$", path)

    if (tarball) {

        path_dir <- extract_tarball (path)
        s1 <- loc_stats (path_dir)
        has_tabs <- any (s1$ntabs > 0L)
        tags_r <- withr::with_dir (path_dir, get_ctags ("R", has_tabs))

    } else {

        s1 <- loc_stats (path)
        has_tabs <- any (s1$ntabs > 0L)
        tags_r <- withr::with_dir (path, get_ctags ("R", has_tabs))

    }

    tags_r <- tags_r [which (tags_r$kind == "function"), ]
    exp_ptn <- gsub (
        "^exportPattern\\(|\\)$|\\\"", "",
        grep ("^exportPattern", nmsp, value = TRUE)
    )
    fns_exp_ptn <- grep (exp_ptn, tags_r$tag, value = TRUE)

    return (fns_exp_ptn)
}

#' Get all aliases from .Rd files
#'
#' These are matched to actual function names from
#' `exp_ptn_r_fn_names()`, and any class, data, method, or package definitions
#' are removed.
#'
#' Removal of 'class', 'data', and 'package' aliases is straightforward. Removal
#' of 'method' aliases simply ignores any aliases within any .Rd file that also
#' includes a 'method' section. That is not entirely reliable, and may also
#' include genuine function names.
#' @noRd
aliases_from_rd <- function (path, nmsp) {

    tarball <- grepl ("\\.tar\\.gz$", path)

    if (tarball) {

        flist <- utils::untar (
            path,
            exdir = fs::path_temp (),
            list = TRUE,
            tar = "internal"
        )

        rd <- grep ("\\.Rd$", flist, value = TRUE)

        chk <- utils::untar (
            path,
            files = rd,
            exdir = fs::path_temp ()
        )
        if (chk != 0) {
            return (NULL)
        }

        flist <- fs::path (fs::path_temp (), rd)

    } else {

        flist <- fs::dir_ls (
            path,
            regexp = "\\.Rd$",
            recurse = TRUE
        )
    }

    flist <- expand_path (flist)
    # only extract Rd files from man directory:
    is_man <- vapply (
        fs::path_split (flist),
        function (i) {
            any (i == "man")
        },
        logical (1L)
    )
    flist <- flist [which (is_man)]

    nms <- lapply (flist, function (i) {

        rd_i <- tools::parse_Rd (i)

        doc_type <- get_Rd_metadata (rd_i, "docType")
        doc_type <- ifelse (length (doc_type) == 0L, "", doc_type)

        out <- NULL

        if (!doc_type %in% c ("class", "data", "methods", "package")) {
            out <- unique (c (
                get_Rd_metadata (rd_i, "alias")
            ))
            if (length (out) == 0L) {
                get_Rd_metadata (rd_i, "name")
            }

            index <- grep ("method(s?|,?)$|class$|<\\-|\\[\\[|,|\\s", out)
            if (length (index) > 0L) {
                out <- out [-index]
            }
            methods <- unlist (lapply (rd_i, function (j) {
                get_Rd_metadata (j, "method")
            }))
            if (length (methods) > 0L) {
                out <- NULL
            }
        }

        return (out)
    })

    nms <- unique (unlist (nms))

    nms_exp_ptn <- exp_ptn_r_fn_names (path, nmsp)

    return (nms [which (nms %in% nms_exp_ptn)])
}

get_pkg_name_version <- function (desc) {

    d <- data.frame (read.dcf (desc))

    return (c (d$Package, d$Version))
}

#' Stats from '.Rd' files
#'
#' @inheritParams loc_stats
#' @return A `data.frame` of function names and numbers of parameters and lines
#' of documentation for each, along with mean and median numbers of characters
#' used to document each parameter.
#' @family stats
#' @export
#' @examples
#' f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
#' # have to extract tarball to call function on source code:
#' path <- extract_tarball (f)
#' rd_stats (path)
rd_stats <- function (path) {

    path <- check_path (path)

    rd_files <- fs::dir_ls (fs::path (path, "man"), regexp = "\\.Rd$")

    suppressWarnings (
        params <- lapply (rd_files, get_one_params)
    )
    params <- do.call (rbind, params)
    params <- params [which (!is.na (params$parameter)), ]
    params_sp <- NULL
    if (!is.null (params)) {
        params_sp <- split (params, f = factor (params$alias))
    }

    n <- vapply (
        params_sp, function (i) {
            ret <- 0L
            if (nrow (i) > 1 | i$parameter [1] != "(none)") {
                ret <- nrow (i)
            }
            return (ret)    },
        integer (1)
    )

    nchars_mn <- vapply (params_sp, function (i) {
        mean (i$nchar)
    }, numeric (1))
    nchars_md <- vapply (params_sp, function (i) {
        stats::median (i$nchar)
    }, numeric (1))

    # excluce imported fns:
    nmspc <- fs::path (path, "NAMESPACE")
    # some packages have no NAMESPACE files (like adehabitat 1.2-1)
    if (file.exists (nmspc)) {

        nmspc <- brio::read_lines (nmspc)

        imports <- grep ("^importFrom", nmspc, value = TRUE)
        imports <- unlist (lapply (
            strsplit (imports, ","),
            function (i) {
                gsub ("\\)$", "", i [2])
            }
        ))
        imports <- gsub ("\\\"", "", imports)

        if (any (imports %in% names (n))) {
            index <- which (!names (n) %in% imports)
            n <- n [index]
            nchars_mn <- nchars_mn [index]
            nchars_md <- nchars_md [index]
        }
    }

    doclines <- vapply (
        names (n), function (i) {
            params$doclines [params$alias == i] [1]
        },
        integer (1)
    )

    ret <- data.frame (
        fn_name = names (n),
        num_params = unname (n),
        num_doclines = doclines,
        param_nchars_mn = nchars_mn,
        param_nchars_md = nchars_md,
        row.names = NULL,
        stringsAsFactors = FALSE
    )

    return (ret)
}

rd_is_fn <- function (rd) {

    if (is.null (rd)) {
        return (FALSE)
    }

    tags <- vapply (rd, function (i) attr (i, "Rd_tag"), character (1))
    tags <- gsub ("\\\\", "", grep ("^\\\\", tags, value = TRUE))

    is_fn <- TRUE
    if ("docType" %in% tags) {
        is_fn <- !get_Rd_metadata (rd, "docType") %in% c ("data", "package")
    }
    return (is_fn)
}

get_one_params <- function (man_file) {

    res <- NULL

    # rd <- tools::parse_Rd (man_file)
    # Rd comments (per sec 2.1 of Extensions manual) can muck up parsing, so
    # must be removed
    x <- brio::read_lines (man_file)
    index1 <- grep ("[^0-9]%", x)
    index2 <- grep ("\\\\%", x)
    index <- index1 [which (!index1 %in% index2)]
    x [index] <- gsub ("%.*$", "", x [index])
    ptn <- paste0 ("Rdtemp-", Sys.getpid (), "-")
    f <- fs::file_temp (pattern = ptn, ext = ".Rd")
    brio::write_lines (x, f)
    rd <- tryCatch (
        tools::parse_Rd (f),
        error = function (e) NULL
    )
    chk <- fs::file_delete (f) # nolint

    if (!rd_is_fn (rd)) {
        return (res)
    }

    out <- tryCatch (
        utils::capture.output (tools::Rd2txt (rd)),
        error = function (e) NULL
    )
    if (is.null (out)) {
        # For 'stages' param, see #40
        out <- tryCatch (
            utils::capture.output (tools::Rd2txt (rd, stages = "build")),
            error = function (e) NULL
        )
    }

    doclines <- length (out [out != ""])

    aliases <- unique (c (
        get_Rd_metadata (rd, "name"),
        get_Rd_metadata (rd, "alias")
    ))

    params <- get_Rd_metadata (rd, "arguments")
    if (length (params) == 0) {

        res <- data.frame (
            parameter = "(none)",
            nchar = NA_integer_,
            alias = aliases,
            doclines = doclines,
            stringsAsFactors = FALSE
        )
    } else {
        params <- strsplit (params, "\\n") [[1]]
        # rm lines with initial Rd comments, but params may still fail to parse
        # when comments occur later in lines.
        params <- params [which (nchar (params) > 0 &
            !grepl ("^\\s?%", params))]
        params <- paste0 (params, collapse = "\n")

        params <- tryCatch (as.list (parse (text = params)),
            error = function (e) NULL
        )

        eval_one_param_item <- function (p) {
            nm <- tryCatch (eval (p), error = function (e) NULL)
            if (is.null (nm)) {
                nm <- as.character (p)
            }
            return (unlist (nm))
        }

        nms <- lapply (params, function (i) {
            p_i <- as.list (i)
            nm <- NA_character_
            desc <- NA_integer_
            if (length (p_i) >= 3) {
                nm <- eval_one_param_item (p_i [[2]])
                desc <- eval_one_param_item (p_i [[3]])
                if (is.null (nm)) {
                    nm <- "(NULL)"
                } else if (length (nm) > 1) {
                    nm <- gsub ("\\s+", " ", paste0 (nm, collapse = ""))
                }
            }
            list (
                par_name = nm,
                nchars = sum (nchar (desc))
            )
        })

        par_name <- vapply (nms, function (i) i$par_name [1], character (1))
        nchars <- vapply (nms, function (i) i$nchars, integer (1))

        res <- data.frame (
            parameter = par_name,
            nchar = nchars,
            stringsAsFactors = FALSE
        )

        if (nrow (res) > 0) {

            res <- lapply (aliases, function (i) {
                res$alias <- i
                return (res)  })
            res <- do.call (rbind, res)
            res$doclines <- doclines
        }
    }

    if (!"doclines" %in% names (res)) {
        res$doclines <- integer (0)
    }

    return (res)
}

test_all <- (identical (Sys.getenv ("MPADGE_LOCAL"), "true") ||
    identical (Sys.getenv ("GITHUB_JOB"), "test-coverage"))

skip_if (!test_all)

test_that ("plot-network", {

    path <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
    s <- pkgstats (path)

    expect_silent (
        net <- plot_network (s, plot = FALSE)
    )

    expect_s3_class (net, "pkgstats_network")
    expect_type (net, "list")
    expect_true (length (net) > 1L)

    nodes0 <- net$nodes
    edges0 <- net$edges
    expect_equal (ncol (nodes0), 8L)
    expect_equal (ncol (edges0), 6L)
    expect_gt (nrow (nodes0), 50L)
    expect_gt (nrow (edges0), 50L)

    expect_type (net$html, "character")
    expect_true (grepl ("pkgstatsNetworkPlot", net$html, fixed = TRUE))
})

#' Download latest version of 'pkgstats' data
#'
#' @param current If 'FALSE', download data for all CRAN packages ever released,
#' otherwise (default) download data only for current CRAN packages.
#' @param path Local path to download file.
#' @param quiet If `FALSE`, display progress information on screen.
#' @return (Invisibly) A `data.frame` of `pkgstats` results, one row for each
#' package.
#' @family archive
#' @export
dl_pkgstats_data <- function (current = TRUE,
                              path = tempdir (),
                              quiet = FALSE) {

    requireNamespace ("curl", quietly = TRUE)
    requireNamespace ("jsonlite", quietly = TRUE)

    cran_data_version <- "v0.1.6"

    # First get integer "release_id" of that version:
    u <- paste0 (
        "https://api.github.com/repos/",
        "ropensci-review-tools/pkgstats/",
        "releases"
    )
    res <- curl::curl_fetch_memory (u)
    res <- jsonlite::fromJSON (rawToChar (res$content))
    row_num <- match (cran_data_version, res$tag_name)
    if (is.na (row_num)) {
        stop (
            "Release for CRAN data {cran_data_version} not found.",
            call. = FALSE
        )
    }
    release_id <- res$id [row_num]

    # Then get that release:
    u <- paste0 (u, "/", release_id)
    res <- curl::curl_fetch_memory (u)
    hdrs <- curl::parse_headers (res$headers)
    http_code <- as.integer (gsub (
        "^http\\/[0-9]\\s?|\\s+$",
        "",
        hdrs [1],
        ignore.case = TRUE
    ))
    if (http_code != 200L) {
        stop (
            "Call to GitHub failed with http error code [",
            http_code, "]"
        )
    }

    res <- jsonlite::fromJSON (rawToChar (res$content))
    assets <- res$assets

    i <- ifelse (
        current,
        grep ("current", assets$name, ignore.case = TRUE),
        grep ("all", assets$name, ignore.case = TRUE)
    )
    dl_url <- assets$browser_download_url [i]
    f <- fs::path (path, basename (dl_url))

    curl::curl_download (url = dl_url, destfile = f, quiet = quiet)
    if (!quiet) {
        message ("Downloaded to [", f, "]")
    }

    invisible (readRDS (f))
}

test_all <- (identical (Sys.getenv ("MPADGE_LOCAL"), "true") ||
    identical (Sys.getenv ("GITHUB_JOB"), "test-coverage"))

skip_if (!test_all)

test_that ("ctags-test", {

    expect_message (
        x <- ctags_test (quiet = FALSE),
        "ctags installation works as expected"
    )

    expect_true (x)
})

test_all <- (identical (Sys.getenv ("MPADGE_LOCAL"), "true") ||
    identical (Sys.getenv ("GITHUB_JOB"), "test-coverage"))

test_that ("pkgstats-null-summary", {

    if (!test_all) {
        Sys.setenv ("PKGSTATS_CRAN_TESTS" = "true")
    }

    path <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
    p <- pkgstats (path)
    if (!test_all) {
        Sys.unsetenv ("PKGSTATS_CRAN_TESTS")
    }
    s1 <- pkgstats_summary (p)

    s0 <- pkgstats_summary (NULL)
    expect_equal (ncol (s0), ncol (s1))

    expect_identical (names (s0), names (s1))
})

test_that ("pkgstats-summary", {

    if (!test_all) {
        Sys.setenv ("PKGSTATS_CRAN_TESTS" = "true")
    }

    path <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")

    p <- pkgstats (path)
    if (!test_all) {
        Sys.unsetenv ("PKGSTATS_CRAN_TESTS")
    }

    expect_silent (
        s <- pkgstats_summary (p)
    )

    expect_s3_class (s, "data.frame")
    expect_equal (nrow (s), 1L)
    expect_equal (ncol (s), 95L)

    skip_if (!test_all)

    # external_calls:
    ext <- strsplit (strsplit (s$external_calls, ",") [[1]], ":")
    ext <- do.call (rbind, ext)
    expect_equal (ncol (ext), 3L)
    expect_true (nrow (ext) > 15L) # 17 packages

    ext <- data.frame (
        pkg = ext [, 1],
        n_total = as.integer (ext [, 2]),
        n_unique = as.integer (ext [, 3])
    )
    expect_true (all (ext$n_total >= ext$n_unique))
    expect_true (mean (ext$n_total) > mean (ext$n_unique))
})

#' Plot interactive visualisation of object-relationship network of package.
#'
#' @param s Package statistics obtained from \link{pkgstats} function.
#' @param plot If `TRUE`, plot the network which opens an interactive browser
#' pane.
#' @param vis_save Name of local file in which to save `html` file of network
#' visualisation (will override `plot` to `FALSE`).
#' @return (Invisibly) A `pkgstats_network` object: a list of the `nodes` and
#' `edges` used to construct the network, along with the self-contained
#' `html` document itself.
#'
#' @note Edge thicknesses are scaled to centrality within the package function
#' call network. Node sizes are scaled to numbers of times each function is
#' called from all other functions within a package. The visualisation is
#' rendered with a small bundle of \pkg{D3}-based JavaScript shipped with
#' this package (see `system.file ("js", package = "pkgstats")`), so no
#' internet connection or additional R packages are required to view it.
#'
#' @family output
#' @examples
#' f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
#' \donttest{
#' p <- pkgstats (f)
#' net <- plot_network (p, plot = FALSE)
#' }
#' # use default 'plot = TRUE' to automatically open network in browser.
#' @export
plot_network <- function (s, plot = TRUE, vis_save = NULL) {

    # suppress no visible binding notes:
    kind <- NULL

    if (!all (c (
        "loc", "vignettes", "data_stats", "desc",
        "translations", "objects", "network"
    ) %in% names (s))) {

        stop (
            "'s' must be a 'pkgstats' object obtained ",
            "from the 'pkgstats' function"
        )
    }

    pkg_title <- paste0 (s$desc$package, " network")

    if (nrow (s$network) == 0L && nrow (s$objects) == 0L) {

        return (NULL)

    } else if (nrow (s$network) == 0L) {

        # pkg has no internal fn_calls, so plot network of vertices only (#12)
        index <- which (s$objects$kind == "function" &
            !grepl ("^anonFunc", s$objects$fn_name))
        nodes <- s$objects [index, ]
        nodes <- nodes [which (!duplicated (nodes$fn_name)), ]
        nodes <- data.frame (
            id = nodes$fn_name,
            label = nodes$fn_name,
            name = nodes$fn_name,
            language = nodes$language,
            group = nodes$language,
            n = NA_integer_,
            value = 1L,
            file = nodes$file_name,
            stringsAsFactors = FALSE
        )
        edges <- data.frame (
            from = character (0),
            to = character (0),
            language = character (0),
            centrality = numeric (0),
            n = integer (0),
            width = numeric (0),
            stringsAsFactors = FALSE
        )
        subtitle <- ""

    } else {

        edges <- data.frame (
            from = s$network$from,
            to = s$network$to,
            centrality = s$network$centrality_undir,
            language = s$network$language,
            stringsAsFactors = FALSE
        )
        from <- language <- centrality <- NULL # suppress no visible binding msg
        edges <- dplyr::count (edges, from, to, language, centrality)
        edges <- edges [which (!is.na (edges$from)), ]

        obj <- dplyr::filter (s$objects, kind == "function")
        obj <- dplyr::filter (obj, !grepl ("^anonFunc", obj$fn_name))
        obj <- obj [which (!duplicated (obj$fn_name)), ]
        nodes <- data.frame (
            id = obj$fn_name,
            label = obj$fn_name,
            name = obj$fn_name,
            language = obj$language,
            file = obj$file_name,
            stringsAsFactors = FALSE
        )
        nodes$group <- nodes$language

        to <- NULL # no visible binding
        n <- dplyr::count (edges, to)
        nodes <- dplyr::left_join (nodes, n, by = c ("name" = "to"))
        nodes$value <- nodes$n
        nodes$value [is.na (nodes$value)] <- 1

        edges$width <- edges$centrality * 10 / max (edges$centrality)

        subtitle <- paste0 (
            "Edge thickness scaled to network centrality &middot; ",
            "node sizes scaled to numbers of times each fn is called"
        )
    }

    html <- pkgstats_network_html (
        nodes, edges,
        title = pkg_title, subtitle = subtitle
    )

    out <- list (nodes = nodes, edges = edges, html = html)
    class (out) <- c ("pkgstats_network", class (out))

    if (plot || !is.null (vis_save)) {

        if (!is.null (vis_save)) {
            if (!is.character (vis_save)) {
                stop ("vis_save must be a character specifying a file name")
            }
            if (length (vis_save) > 1) {
                stop ("vis_save must be a single character")
            }
            if (!fs::dir_exists (fs::path_dir (vis_save))) {
                stop (
                    "directory [", fs::path_dir (vis_save),
                    "] does not exist"
                )
            }

            vis_save <- fs::path_ext_set (vis_save, "html")
            path <- fs::path_dir (vis_save)
            if (!fs::dir_exists (path)) {
                fs::dir_create (path, recurse = TRUE)
            }
            brio::write_file (html, vis_save)

        } else {
            tmp <- fs::file_temp (ext = "html")
            brio::write_file (html, tmp)
            utils::browseURL (tmp)
        }
    }

    invisible (out)
}

#' Assemble the self-contained network-plot `html` document from a bundled
#' `D3`-based JavaScript skeleton (see `inst/js`), populated with the given
#' `nodes` and `edges`.
#'
#' @noRd
pkgstats_network_html <- function (nodes, edges, title, subtitle = "") {

    requireNamespace ("jsonlite", quietly = TRUE)

    data <- list (
        nodes = nodes [, c ("id", "label", "group", "value", "file")],
        edges = edges [, c ("from", "to", "width")]
    )
    data_json <- jsonlite::toJSON (
        data,
        dataframe = "rows",
        auto_unbox = TRUE,
        na = "null"
    )

    js_path <- function (f) system.file ("js", f, package = "pkgstats")

    template <- brio::read_file (js_path ("template.html"))
    d3_js <- brio::read_file (js_path ("d3.v7.min.js"))
    network_js <- brio::read_file (js_path ("pkgstats-network.js"))
    network_css <- brio::read_file (js_path ("pkgstats-network.css"))

    html <- gsub ("__TITLE__", title, template, fixed = TRUE)
    html <- gsub ("__SUBTITLE__", subtitle, html, fixed = TRUE)
    html <- gsub ("__CSS__", network_css, html, fixed = TRUE)
    html <- gsub ("__D3_JS__", d3_js, html, fixed = TRUE)
    html <- gsub ("__NETWORK_JS__", network_js, html, fixed = TRUE)
    html <- gsub ("__DATA__", data_json, html, fixed = TRUE)

    html
}

test_all <- (identical (Sys.getenv ("MPADGE_LOCAL"), "true") ||
    identical (Sys.getenv ("GITHUB_JOB"), "test-coverage"))

skip_if (!test_all)

test_that ("cran data update functions", {

    results_path <- fs::path (fs::path_temp (), "pkgstats-data")
    fs::dir_create (results_path)

    f <- dl_prev_data (results_path, what = "current")
    expect_type (f, "character")
    expect_length (f, 1L)
    expect_true (fs::file_exists (f))

    x <- readRDS (f)
    expect_s3_class (x, "data.frame")
    expect_equal (ncol (x), 95L)
    expect_true (nrow (x) > 20000L)

    expect_silent (
        check_prev_results (x)
    )

    new <- list_new_cran_updates (x)
    expect_type (new, "character")
    if (length (new) > 0L) {
        expect_true (all (grepl ("\\_[0-9]", new)))
    }

    fs::dir_delete (results_path)
})

test_that ("download tarball", {

    pkg <- "odbc"
    u <- paste0 ("https://cran.r-project.org/package=", pkg)
    qry <- httr2::request (u)
    resp <- httr2::req_perform (qry)
    # No 'xml' or 'rvest' here, so process as raw text:
    b <- httr2::resp_body_string (resp)
    b <- strsplit (b, ",\\s") [[1]]
    tarball <- grep ("\\.tar\\.gz", b, value = TRUE)
    ptn <- "odbc\\_[0-9](\\.[0-9]){2,}\\.tar\\.gz"
    tarball <- regmatches (tarball, regexpr (ptn, tarball))
    tarball <- gsub ("\\.tar\\.gz$", "", tarball)

    results_path <- fs::path (fs::path_temp (), "pkgstats-data")
    fs::dir_create (results_path)

    path <- dl_one_tarball (results_path, tarball)
    expect_true (fs::file_exists (path))

    fs::dir_delete (results_path)
})

# Some pkgs define an "exportPattern" in the namespace, and use that to define
# exported functions. These call the separate `aliases_from_rd()` function,
# which is covered in these tests. The selected package was identified using
# the following code:
#
# flist <- fs::dir_ls (path_to_tarballs, regexp = "\\.tar\\.gz$")
# dat <- pbapply::pblapply (flist, function (f) {
#
#     tb <- tryCatch (extract_tarball (f), error = function (e) NULL)
#     ret <- NULL
#     if (!is.null (tb)) {
#         nmsp <- get_namespace_contents (tb)
#         if (any (grepl ("exportPattern", nmsp))) {
#             ret <- c (unname (f), as.character (file.size (f)))
#         }
#         fs::dir_delete (tb)
#     }
#
#     return (ret)
# })
# dat <- do.call (rbind, dat)
# rownames (dat) <- NULL
# dat <- data.frame (pkg = basename (dat [, 1]), size = as.numeric (dat [, 2]))
# dat <- dat [order (dat$size), ]
#
test_that ("fn_names_from_export_ptn", {

    pkg <- "abodOutlier"
    u <- paste0 ("https://cran.r-project.org/package=", pkg)
    qry <- httr2::request (u)
    resp <- httr2::req_perform (qry)
    # No 'xml' or 'rvest' here, so process as raw text:
    b <- httr2::resp_body_string (resp)
    b <- strsplit (b, ",\\s") [[1]]
    tarball <- grep ("\\.tar\\.gz", b, value = TRUE)
    ptn <- "abodOutlier\\_[0-9]\\.[0-9]\\.tar\\.gz"
    tarball <- regmatches (tarball, regexpr (ptn, tarball))
    tarball <- gsub ("\\.tar\\.gz$", "", tarball)

    results_path <- fs::path (fs::path_temp (), "pkgstats-data")
    fs::dir_create (results_path)

    path <- dl_one_tarball (results_path, tarball)
    expect_true (fs::file_exists (path))

    fn_nms <- pkgstats_fn_names (path)
    expect_s3_class (fn_nms, "data.frame")
    expect_equal (ncol (fn_nms), 3L)
    expect_equal (names (fn_nms), c ("package", "version", "fn_name"))
    expect_true (nrow (fn_nms) > 0L)

    fs::dir_delete (results_path)

    pkg <- fs::path (fs::path_temp (), gsub ("\\_.*$", "", tarball))
    if (fs::dir_exists (pkg)) {
        fs::dir_delete (pkg)
    }

})

#' Analyse statistics of one R package
#'
#' @param path Either a path to a local source repository, or a local '.tar.gz'
#' file, containing code for an R package.
#' @return List of statistics and data on function call networks (or object
#' relationships in other languages). Includes the following components:
#' \enumerate{
#'   \item loc: Summary of Lines-of-Code in all package directories
#'   \item vignettes: Numbers of vignettes and "demo" files
#'   \item data_stats: Statistics of numbers and sizes of package data files
#'   \item desc: Summary of contents of 'DESCRIPTION' file
#'   \item translations: List of translations into other (human) languages
#'   (where provides)
#'   \item objects: A `data.frame` of all functions in R, and all other
#'   objects (functions, classes, structures, global variables, and more) in all
#'   other languages
#'   \item network: A `data.frame` of object references within and between all
#'   languages; in R these are function calls, but may be more abstract in other
#'   languages.
#'   \item external_calls: A `data.frame` of all calls make to all functions
#'   from all other R packages, including base and recommended as well as
#'   contributed packages.
#' }
#'
#' @family stats
#' @export
#' @examples
#' # 'path' can be path to a package tarball:
#' f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
#' @examplesIf ctags_test (noerror = TRUE)
#' s <- pkgstats (f)
#' # or to a source directory:
#' path <- extract_tarball (f)
#' s <- pkgstats (path)
pkgstats <- function (path = ".") {

    path <- check_path (path)

    tarball <- FALSE
    if (grepl ("\\.tar\\.gz$", path)) {
        tarball <- TRUE
        path <- extract_tarball (path)
    }

    s1 <- loc_stats (path)
    has_tabs <- any (s1$ntabs > 0L) # passed to tags_data below

    num_vignettes <- get_num_vignettes (path)
    num_demos <- get_num_demos (path)
    data_stats <- get_data_stats (path)
    s2 <- desc_stats (path)
    s3 <- rd_stats (path)

    fns <- all_r_functions (path)

    if (nrow (fns) > 0L) {

        fns$exported <- FALSE
        index <- which (fns$fn_name %in% s3$fn_name)
        fns$exported [which (fns$fn_name %in% s3$fn_name)] <- TRUE

        # non-dplyr left_join:
        index2 <- match (fns$fn_name [index], s3$fn_name)
        fns$num_doclines <- fns$param_nchars_mn <-
            fns$param_nchars_md <- NA_integer_
        fns$num_doclines [index] <- s3$num_doclines [index2]
        fns$param_nchars_mn [index] <- s3$param_nchars_mn [index2]
        fns$param_nchars_md [index] <- s3$param_nchars_md [index2]

        # Get stats for non-exported fns
        roxy_stats <- roxygen_block_stats (path)
        roxy_stats <- roxy_stats [which (!roxy_stats$exported), ]
        roxy_stats <- roxy_stats [which (roxy_stats$fn_name %in% fns$fn_name), ]
        index <- match (roxy_stats$fn_name, fns$fn_name)
        par_is_numeric <- vapply (
            names (roxy_stats),
            function (i) is.numeric (roxy_stats [[i]]),
            logical (1L)
        )
        var_names <- names (roxy_stats) [which (par_is_numeric)]
        for (v in var_names) {
            fns [[v]] [index] <- roxy_stats [[v]]
        }
    }

    # Running 'ctags_test()' on some CRAN machines takes too long (> 10s), so
    # this flag is used to switch off tagging routines on CRAN tests.
    if (Sys.getenv ("PKGSTATS_CRAN_TESTS") == "true") {
        tags <- dummy_tags_data ()
    } else {
        # `s2$package` mucks up local linter
        tags <- tags_data (path, has_tabs, s2 [["package"]])
    }

    translations <- get_translations (path)

    if (tarball) {
        chk <- fs::file_delete (path)
    }

    list (
        loc = s1,
        vignettes = c (
            vignettes = num_vignettes,
            demos = num_demos
        ),
        data_stats = data_stats,
        desc = s2,
        translations = translations,
        objects = add_src_to_fn_data (fns, tags$objects),
        network = tags$network,
        external_calls = tags$external_calls
    )
}

#' Get all exported and internal R functions
#' @noRd
all_r_functions <- function (path) {

    if (Sys.getenv ("PKGSTATS_CRAN_TESTS") == "true") {
        return (all_r_functions_dummy ())
    }

    r_files <- expand_path (fs::dir_ls (
        fs::path (path, "R"),
        regexp = "\\.(r|R|q|s|S)$"
    ))

    if (length (r_files) == 0L) {
        return (all_r_functions_dummy ())
    }

    eval1 <- function (f) {

        p <- control_parse (file = f)
        if (methods::is (p, "simpleError")) {
            return (NULL)
        }

        # All functions are parsed to length 3, while things like `"_PACKAGE"`
        # statements are parsed, but have parse lengths < 3.
        p <- p [which (vapply (p, length, integer (1)) > 2L)]

        loc <- vapply (
            p, function (i) {
                length (deparse (i))
            },
            integer (1)
        )
        nms <- vapply (
            p, function (i) {
                paste0 (as.list (i) [[2]], collapse = "")
            },
            character (1)
        )
        npars <- vapply (
            p, function (i) {
                dummy <- rep (NA_integer_, 2L)
                if (is.symbol (as.list (i) [[3]])) {
                    return (dummy)
                }
                call_i <- as.list (i) [[3]]
                if (length (call_i) < 2) { # not a fn
                    return (dummy)
                }
                c2 <- call_i [[2]]
                # if the call has an empty field, c2 won't necessary exist
                nms <- tryCatch (names (c2), error = function (e) NULL)
                if (is.null (nms)) {
                    return (dummy)
                }
                c (length (c2), "..." %in% nms) },
            integer (2)
        )

        data.frame (
            file_name = rep (basename (f), length (p)),
            fn_name = nms,
            loc = loc,
            npars = npars [1, ],
            has_dots = as.logical (npars [2, ]),
            stringsAsFactors = FALSE
        )
    }

    ret <- do.call (rbind, lapply (r_files, eval1))
    if (length (ret) == 0L) {
        return (all_r_functions_dummy ())
    }
    if (nrow (ret) > 0L) {
        # append "R" directory to file names:
        ret$file_name <- as.character (fs::path ("R", ret$file_name))
    }

    return (ret)
}

#' Dummy return from `all_funcitons()` function, triggered only on CRAN tests
#' which otherwise take too long because of the `parse` calls.
#' @noRd
all_r_functions_dummy <- function () {

    data.frame (
        "file_name" = character (0),
        "fn_name" = character (0),
        "loc" = integer (0),
        "npars" = integer (0),
        "has_dots" = logical (0),
        "exported" = logical (0),
        "param_nchars_md" = numeric (0),
        "param_nchars_mn" = numeric (0),
        "num_doclines" = integer (0)
    )
}

#' param src value of tags$stats from tags_data()
#' @noRd
add_src_to_fn_data <- function (fns, src) {

    if (nrow (fns) == 0L) {
        return (fns)
    }

    n <- grep ("fn_name", names (fns))

    if (length (n) == 0) { # data-only packages

        fns <- data.frame (
            file_name = NA_character_,
            fn_name = NA_character_,
            kind = NA_character_,
            language = NA_character_,
            loc = 0L,
            npars = NA_integer_,
            has_dots = FALSE,
            exported = FALSE,
            param_nchars_md = NA_integer_,
            param_nchars_mn = NA_integer_,
            num_doclines = NA_integer_,
            stringsAsFactors = FALSE
        )
    } else {

        fns <- data.frame (
            fns [, seq (n)], # file_name, fn_name
            kind = "function",
            language = "R",
            fns [, seq (ncol (fns)) [-seq (n)]],
            stringsAsFactors = FALSE
        )
    }

    if (all (is.na (src$file_name))) {

        src <- NULL

    } else {

        src$exported <- FALSE
        nms <- names (fns) [which (!names (fns) %in% names (src))]
        for (n in nms) {
            src [n] <- NA_integer_
        }
        src <- src [, match (names (fns), names (src))]
        index <- which (src$language %in% c ("R", "S4Class"))
        src_r <- src [index, ]
        src <- src [-(index), ]

        src_r <- src_r [which (src_r$fn_name %in% fns$fn_name), ]
        index <- match (src_r$fn_name, fns$fn_name)
        doclines_from_src <- src_r$num_doclines [index]
        index <- which (is.na (fns$num_doclines) | fns$num_doclines == 0L)
        fns$num_doclines [index] <- doclines_from_src [index]
    }

    out <- rbind (fns, src)

    rownames (out) <- NULL

    return (out)
}

#' Set up gtags files if not already used
#'
#' @return `TRUE` if there are no pre-existing gtags files; otherwise `FALSE`.
#' @noRd
make_gtags <- function () {

    path <- expand_path (".")
    flist <- fs::dir_ls (
        path,
        recurse = TRUE,
        regexp = "GRTAGS$|GPATH$|GTAGS$"
    )

    if (length (flist) == 0) {
        sys::exec_wait ("gtags")
        wait_for_process ("gtags")
    }

    return (length (flist) == 0)
}

get_gtags <- function () {

    f <- fs::file_temp (pattern = "global_")
    sys::exec_wait ("global", args = c ("-rx", "."), std_out = f)
    wait_for_process ("global")

    x <- brio::read_lines (f)

    # global may fail to parse files, as happens for example with "rms" package
    if (length (x) == 0) {
        return (NULL)
    }

    # these are fixed width, but only have 4 cols, so can just replace the first
    # 3 lots of space with single tab characters.
    for (i in 1:3) {
        x <- sub (" +", "\t", x)
    }

    ctypes <- list (
        readr::col_character (),
        readr::col_double (),
        readr::col_character (),
        readr::col_character ()
    )
    cnames <- c ("tag", "line", "file", "content")
    suppressWarnings (
        gtags <- tryCatch (
            readr::read_tsv (
                paste0 (x, collapse = "\n"),
                col_names = cnames,
                col_types = ctypes,
                col_select = cnames,
                progress = FALSE,
                lazy = FALSE
            ),
            error = function (e) NULL
        )
    )

    rm_file_no_err (f)
    gtags <- rm_vendored_code (gtags)
    if (is.null (gtags) || nrow (gtags) == 0L) {
        return (NULL)
    }

    gtags <- gtags [which (!grepl (excluded_file_ptn (), gtags$file)), ]

    return (gtags)
}

#' Put ctag tag objects onto the gtag references where they are called from
#' @param f one '/src' file for for which to assign ctag tags
#' @noRd
gtags_from_one_file <- function (ctags, gtags, f) {

    ctags_f <- ctags [ctags$file == f, ]
    # tags are duplicated for things like class constructors. The duplications
    # are always embedded within the main definition, so simply removing them
    # reduces line ranges to main definition only.
    ctags_f <- ctags_f [which (!duplicated (ctags_f$tag)), ]
    # end lines are not always given, as in Fortran code for which ctags works
    # but gtags does not
    if (any (is.na (ctags_f$end))) {
        return (gtags)
    }

    line_nums <- lapply (seq (nrow (ctags_f)), function (i) {
        data.frame (
            n = i,
            l = seq (
                ctags_f$start [i],
                ctags_f$end [i]
            ),
            stringsAsFactors = FALSE
        )
    })
    line_nums <- do.call (rbind, line_nums)

    index <- which (gtags$file == f)
    items <- line_nums$n [match (gtags$line [index], line_nums$l)]
    gtags$from [index] <- ctags_f$tag [items]

    return (gtags)
}

rm_gtags_files <- function (path) {

    flist <- fs::dir_ls (
        path,
        recurse = TRUE,
        regexp = "GRTAGS$|GPATH$|GTAGS$"
    )

    ret <- NULL

    if (length (flist) > 0) {

        ret <- tryCatch (fs::file_delete (flist),
            error = function (e) e
        )
    }

    return (ret)
}

#' Condense the output of `pkgstats` to summary statistics only
#'
#' @param s Output of \link{pkgstats}, containing full statistical data on one
#' package. Default of `NULL` returns a single row with `NA` values (used in
#' \link{pkgstats_from_archive}).
#' @return Summarised version of `s`, as a single row of a standardised
#' `data.frame` object
#'
#' @note Variable names in the summary object use the following abbreviations:
#' \itemize{
#'   \item "loc" = Lines-of-Code
#'   \item "fn" = Function
#'   \item "n_fns" = Number of functions
#'   \item "npars" = Number of parameters
#'   \item "doclines" = Number of documentation lines
#'   \item "nedges" = Number of edges in function call network, as a count of
#'   \emph{unique} edges, which may be less than the size of the `network`
#'   object returned by \link{pkgstats}, because that may include multiple calls
#'   between identical function pairs.
#'   \item "n_clusters" = Number of connected clusters within the function call
#'   network.
#'   \item "centrality" used as a prefix for several statistics, along with
#'   "dir" or "undir" for centrality calculated on networks respectively
#'   constructed with directed or undirected edges; "mn" or "md" for respective
#'   measures of mean or median centrality, and "no0" for measures excluding
#'   edges with zero centrality.
#' }
#' @family stats
#' @export
#' @examples
#' f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
#' @examplesIf ctags_test (noerror = TRUE)
#' p <- pkgstats (f)
#' s <- pkgstats_summary (p)
pkgstats_summary <- function (s = NULL) {

    if (is.null (s)) {
        return (null_stats ())
    }

    out <- data.frame (
        package = s$desc$package,
        version = s$desc$version,
        date = s$desc$date,
        license = s$desc$license,
        stringsAsFactors = FALSE
    )

    out <- cbind (out, loc_summary (s$loc))

    out$num_vignettes <- s$vignettes [1]
    out$num_demos <- s$vignettes [2]
    out$num_data_files <- s$data_stats [1]
    out$data_size_total <- s$data_stats [2]
    out$data_size_median <- s$data_stats [3]

    tr <- grepv ("^R\\-", basename (s$translations))
    out$translations <- paste0 (gsub ("^R\\-", "", tr), collapse = ", ")

    out <- cbind (out, desc_summary (s$desc))

    out <- cbind (out, object_summary (s$objects))

    # Append any languages not parsed by ctags:
    loc_langs <- unique (s$loc$language [s$loc$dir %in% c ("src", "inst")])
    loc_langs <- grep ("^[^R]|^html", loc_langs,
        value = TRUE, ignore.case = TRUE
    )
    langs <- unique (unlist (c (strsplit (out$languages, ",\\s"), loc_langs)))
    out$languages <- paste0 (langs, collapse = ", ")

    out <- cbind (out, network_summary (s$network))

    out <- cbind (out,
        external_calls = external_call_summary (s$external_calls)
    )

    # suffix `_pkg` because inter-package stats are calculated after initial
    # collation of all data (issue #32)
    out$cpl_instability_pkg <- coupling_instability (s)

    return (out)
}

null_stats <- function () {

    out <- data.frame (
        package = NA_character_,
        version = NA_character_,
        date = NA_character_,
        license = NA_character_,
        stringsAsFactors = FALSE
    )

    loc_nms <- c (
        "files_R", "files_src", "files_inst", "files_vignettes",
        "files_tests", "loc_R", "loc_src", "loc_inst",
        "loc_vignettes", "loc_tests", "blank_lines_R",
        "blank_lines_src", "blank_lines_inst",
        "blank_lines_vignettes", "blank_lines_tests",
        "comment_lines_R", "comment_lines_src", "comment_lines_inst",
        "comment_lines_vignettes", "comment_lines_tests", "rel_space",
        "rel_space_R", "rel_space_src", "rel_space_inst",
        "rel_space_vignettes", "rel_space_tests", "indentation",
        "nexpr"
    )
    out [loc_nms] <- NA_integer_

    out$num_vignettes <- NA_integer_
    out$num_demos <- NA_integer_
    out$num_data_files <- NA_integer_
    out$data_size_total <- NA_integer_
    out$data_size_median <- NA_integer_
    out$translations <- NA_character_

    desc_nms <- c (
        "urls", "bugs", "desc_n_aut", "desc_n_ctb", "desc_n_fnd",
        "desc_n_rev", "desc_n_ths", "desc_n_trl", "depends",
        "imports", "suggests", "enhances", "linking_to"
    )
    out [desc_nms] <- NA_character_
    out [grep ("^desc\\_", desc_nms, value = TRUE)] <- NA_integer_

    obj_nms <- c (
        "n_fns_r", "n_fns_r_exported", "n_fns_r_not_exported",
        "n_fns_src", "n_fns_per_file_r", "n_fns_per_file_src",
        "npars_exported_mn", "npars_exported_md", "loc_per_fn_r_mn",
        "loc_per_fn_r_md", "loc_per_fn_r_exp_mn",
        "loc_per_fn_r_exp_md", "loc_per_fn_r_not_exp_mn",
        "loc_per_fn_r_not_exp_md", "loc_per_fn_src_mn",
        "loc_per_fn_src_md", "languages",
        "doclines_per_fn_exp_mn", "doclines_per_fn_exp_md",
        "doclines_per_fn_not_exp_mn", "doclines_per_fn_not_exp_md",
        "doclines_per_fn_src_mn", "doclines_per_fn_src_md",
        "docchars_per_par_exp_mn", "docchars_per_par_exp_md"
    )
    out [obj_nms] <- NA_integer_

    net_nms <- c (
        "n_edges", "n_edges_r", "n_edges_src", "n_clusters",
        "centrality_dir_mn", "centrality_dir_md",
        "centrality_dir_mn_no0", "centrality_dir_md_no0",
        "centrality_undir_mn", "centrality_undir_md",
        "centrality_undir_mn_no0", "centrality_undir_md_no0",
        "num_terminal_edges_dir", "num_terminal_edges_undir",
        "node_degree_mn", "node_degree_md", "node_degree_max"
    )
    out [net_nms] <- NA_integer_

    out$external_calls <- NA_character_
    out$cpl_instability_pkg <- NA

    return (out)
}

#' @param x the 'loc` component of `pkgstats` output.
#' @noRd
loc_summary <- function (x) {

    # suprress no visible binding notes:
    nfiles <- ncode <- ndoc <- nempty <- nspaces <- nchars <- language <- NULL

    indentation <- loc_indentation (x)

    rm_dirs <- c ("inst/doc", "inst/extdata")
    x <- x [which (!x$dir %in% rm_dirs), ]
    rm_langs <- c ("HTML", "CSS", "YAML")
    x <- x [which (!x$language %in% rm_langs), ]

    test_dirs <- c ("inst/tinytest", "tests/testthat")
    x$dir [x$dir %in% test_dirs] <- "tests"

    # Presume everything else in inst is genuine code:
    x$dir [grep ("^inst\\/", x$dir)] <- "inst"

    if ("YAML" %in% x$language) {
        x <- dplyr::filter (x, language != "YAML")
    }
    xg <- dplyr::group_by (x, dir)
    x <- dplyr::summarise (
        xg,
        nfiles = sum (nfiles, na.rm = TRUE),
        ncode = sum (ncode, na.rm = TRUE),
        ndoc = sum (ndoc, na.rm = TRUE),
        nempty = sum (nempty, na.rm = TRUE),
        nspaces = sum (nspaces, na.rm = TRUE),
        nchars = sum (nchars, na.rm = TRUE),
        nexpr = stats::median (nexpr, na.rm = TRUE)
    )
    # nexpr is not truly accurate, but acceptable here

    blank <- x [1, ]
    col_nms <- names (blank) [which (!names (blank) == "dir")]
    blank [col_nms] <- 0L

    add_if_missing <- function (x, type = "src") {

        if (!type %in% x$dir) {

            src <- blank
            src$dir <- type
            x <- rbind (x, src)
        }

        return (x)
    }

    has_code <- any (x$nfiles > 0) # data-only pkgs may have no code at all

    # R shouldn't ever be missing, but maybe? (BDR's boot pkg comes close)
    if (has_code) {

        x <- add_if_missing (x, "R")
        x <- add_if_missing (x, "src")
        x <- add_if_missing (x, "inst")
        x <- add_if_missing (x, "vignettes")
        x <- add_if_missing (x, "tests")
    }

    files_R <- files_src <- files_inst <- # nolint
        files_vignettes <- files_tests <- 0L
    rel_space_R <- rel_space_src <- rel_space_inst <- # nolint
        rel_space_vignettes <- rel_space_tests <- NA

    if (has_code) {

        rel_space <- x$nspaces / x$nchars
        rel_space [rel_space == 0 | is.nan (rel_space)] <- NA

        dirs <- c ("R", "src", "inst", "vignettes", "tests")
        for (d in dirs) {

            assign (paste0 ("files_", d), x$nfiles [x$dir == d])
            assign (paste0 ("loc_", d), x$ncode [x$dir == d])
            assign (paste0 ("blank_lines_", d), x$nempty [x$dir == d])
            assign (paste0 ("comment_lines_", d), x$ndoc [x$dir == d])

            di <- which (x$dir == d)
            assign (paste0 ("rel_space_", d), rel_space [di])
        }

    }

    if (files_R == 0) { # nolint
        loc_R <- blank_lines_R <- comment_lines_R <- NA_integer_ # nolint
    } # nolint
    if (files_src == 0) {
        loc_src <- blank_lines_src <- comment_lines_src <- NA_integer_
    }
    if (files_inst == 0) {
        loc_inst <- blank_lines_inst <- comment_lines_inst <- NA_integer_
    }
    if (files_vignettes == 0) {
        loc_vignettes <-
            blank_lines_vignettes <-
            comment_lines_vignettes <- NA_integer_
    }
    if (files_tests == 0) {
        loc_tests <- blank_lines_tests <- comment_lines_tests <- NA_integer_
    }

    nexpr <- x$nexpr [which (x$dir %in% c ("inst", "R", "src") & x$nexpr > 0)]
    nexpr <- stats::median (nexpr)

    data.frame (
        files_R = files_R,
        files_src = files_src,
        files_inst = files_inst,
        files_vignettes = files_vignettes,
        files_tests = files_tests,
        loc_R = loc_R,
        loc_src = loc_src,
        loc_inst = loc_inst,
        loc_vignettes = loc_vignettes,
        loc_tests = loc_tests,
        blank_lines_R = blank_lines_R,
        blank_lines_src = blank_lines_src,
        blank_lines_inst = blank_lines_inst,
        blank_lines_vignettes = blank_lines_vignettes,
        blank_lines_tests = blank_lines_tests,
        comment_lines_R = comment_lines_R,
        comment_lines_src = comment_lines_src,
        comment_lines_inst = comment_lines_inst,
        comment_lines_vignettes = comment_lines_vignettes,
        comment_lines_tests = comment_lines_tests,
        rel_space = sum (x$nspaces) / sum (x$nchars),
        rel_space_R = rel_space_R,
        rel_space_src = rel_space_src,
        rel_space_inst = rel_space_inst,
        rel_space_vignettes = rel_space_vignettes,
        rel_space_tests = rel_space_tests,
        indentation = indentation,
        nexpr = nexpr,
        stringsAsFactors = FALSE
    )
}

#' Median indentation from `loc` object
#'
#' @param x The "loc" item returned from main `pkgstats` fn
#' @param tab_threshold If numbers of initial tab characters in relation to
#' total lines of code is greater than this, then code is considered to be
#' tab-indented.
#' @return A positive integer quantifying median indentation of code, or a value
#' of -1L for tab-indentation.
#'
#' @note Indentation metrics are not truly accurate here, but likely good
#' enough. The `loc` object has single values of median indentation per file,
#' and numbers of files in each directory. These values are converted into
#' repeated sequences to obtain a subsequent approximation of the true median.
#' @noRd
loc_indentation <- function (x, tab_threshold = 0.1) {

    index <- which (x$ncode > 0)
    if (length (index) == 0) { # data-only pkgs with no code
        return (NA_integer_)
    }

    tabs_per_line <- mean (x$ntabs [index] / x$ncode [index], na.rm = TRUE)
    if (tabs_per_line > tab_threshold) {
        return (-1L)
    }

    indents <- rep (x$indentation, times = x$nfiles)
    return (stats::median (indents))
}

#' @param x the 'desc' components of 'pkgstats' output
#' @noRd
desc_summary <- function (x) {

    data.frame (
        urls = x$urls,
        bugs = x$bugs,
        desc_n_aut = x$aut,
        desc_n_ctb = x$ctb,
        desc_n_fnd = x$fnd,
        desc_n_rev = x$rev,
        desc_n_ths = x$ths,
        desc_n_trl = x$trl,
        depends = x$depends,
        imports = x$imports,
        suggests = x$suggests,
        enhances = x$enhances,
        linking_to = x$linking_to,
        stringsAsFactors = FALSE
    )
}

#' @param x the 'objects' components of 'pkgstats' output
#' @noRd
object_summary <- function (x) {

    if (nrow (x) == 0L) {
        return (
            data.frame (
                n_fns_r = 0L,
                n_fns_r_exported = 0L,
                n_fns_r_not_exported = 0L,
                n_fns_src = 0L,
                n_fns_per_file_r = 0L,
                n_fns_per_file_src = 0L,
                npars_exported_mn = 0L,
                npars_exported_md = 0L,
                loc_per_fn_r_mn = 0,
                loc_per_fn_r_md = 0,
                loc_per_fn_r_exp_mn = 0,
                loc_per_fn_r_exp_md = 0,
                loc_per_fn_r_not_exp_mn = 0,
                loc_per_fn_r_not_exp_md = 0,
                loc_per_fn_src_mn = 0,
                loc_per_fn_src_md = 0,
                languages = NA_character_,
                doclines_per_fn_exp_mn = 0,
                doclines_per_fn_exp_md = 0,
                doclines_per_fn_not_exp_mn = 0,
                doclines_per_fn_not_exp_md = 0,
                doclines_per_fn_src_mn = 0,
                doclines_per_fn_src_md = 0,
                docchars_per_par_exp_mn = 0,
                docchars_per_par_exp_md = 0,
                stringsAsFactors = FALSE
            )
        )
    }

    fns <- x [x$kind == "function" &
        !grepl ("anonFunc", x$fn_name), ]
    fns$num_doclines [is.na (fns$num_doclines)] <- 0L
    fns_r <- fns [fns$language == "R", ]
    fns_not_r <- fns [fns$language != "R", ]

    n_fns_r <- length (which (!is.na (fns_r$fn_name)))
    n_fns_r_exported <- length (which (!is.na (fns_r$exported) &
        fns_r$exported))
    n_fns_r_not_exported <- n_fns_r - n_fns_r_exported
    repl0withNA <- function (x) { # nolint
        if (x == 0L) {
            x <- NA_integer_
        }
        return (x)
    }
    n_fns_r <- repl0withNA (n_fns_r)
    n_fns_r_exported <- repl0withNA (n_fns_r_exported)
    if (is.na (n_fns_r_exported)) {
        n_fns_r_not_exported <- repl0withNA (n_fns_r_not_exported)
    }
    # otherwise pkgs with all R fns exported should retain 0L here

    get_n_fns_per_file <- function (f) {
        index <- which (!is.na (f$file_name))
        n_files <- length (unique (f$file_name [index]))
        n_fns <- length (unique (f$fn_name [index]))
        ifelse (n_files == 0L, NA_integer_, n_fns / n_files)
    }
    n_fns_per_file_r <- get_n_fns_per_file (fns_r)
    n_fns_per_file_src <- get_n_fns_per_file (fns_not_r)

    index_exp <- which (fns_r$exported)
    index_not_exp <- which (!fns_r$exported)

    npars_exported_mn <- mean (fns_r$npars [index_exp], na.rm = TRUE)
    npars_exported_md <- stats::median (fns_r$npars [index_exp], na.rm = TRUE)
    loc_per_fn_r_mn <- mean (fns_r$loc, na.rm = TRUE)
    loc_per_fn_r_md <- stats::median (fns_r$loc, na.rm = TRUE)
    loc_per_fn_r_exp_mn <- mean (fns_r$loc [index_exp], na.rm = TRUE)
    loc_per_fn_r_exp_md <- stats::median (fns_r$loc [index_exp], na.rm = TRUE)
    loc_per_fn_r_not_exp_mn <- mean (fns_r$loc [index_not_exp], na.rm = TRUE)
    loc_per_fn_r_not_exp_md <- stats::median (fns_r$loc [index_not_exp],
        na.rm = TRUE
    )

    doclines_per_fn_exp_mn <- mean (fns_r$num_doclines [index_exp],
        na.rm = TRUE
    )
    doclines_per_fn_exp_md <- stats::median (fns_r$num_doclines [index_exp],
        na.rm = TRUE
    )
    doclines_per_fn_not_exp_mn <- mean (fns_r$num_doclines [index_not_exp],
        na.rm = TRUE
    )
    doclines_per_fn_not_exp_md <-
        stats::median (fns_r$num_doclines [index_not_exp], na.rm = TRUE)

    doclines_src <- x$num_doclines [grep ("^(src|inst)", x$file_name)]
    doclines_per_fn_src_mn <- mean (doclines_src, na.rm = TRUE)
    doclines_per_fn_src_md <- stats::median (doclines_src, na.rm = TRUE)

    # exract nchars-per-param for exported only:
    nchars_tot <- x$param_nchars_mn * x$npars
    docchars_per_par_exp_mn <- sum (nchars_tot [index_exp], na.rm = TRUE) /
        sum (x$npars [index_exp], na.rm = TRUE)
    nchar <- x$param_nchars_mn [index_exp]
    npar <- x$npars [index_exp]
    index <- which (!is.na (nchar) & !is.na (npar))
    nchars_tot <- rep (nchar [index], times = npar [index])
    docchars_per_par_exp_md <- stats::median (nchars_tot, na.rm = TRUE)

    n_fns_src <- length (which (!is.na (fns_not_r$fn_name)))
    n_fns_src <- repl0withNA (n_fns_src)
    loc_per_fn_src_mn <- mean (fns_not_r$loc, na.rm = TRUE)
    loc_per_fn_src_md <- stats::median (fns_not_r$loc, na.rm = TRUE)
    languages <- paste0 (unique (fns_not_r$language), collapse = ", ")

    data.frame (
        n_fns_r = n_fns_r,
        n_fns_r_exported = n_fns_r_exported,
        n_fns_r_not_exported = n_fns_r_not_exported,
        n_fns_src = n_fns_src,
        n_fns_per_file_r = n_fns_per_file_r,
        n_fns_per_file_src = n_fns_per_file_src,
        npars_exported_mn = npars_exported_mn,
        npars_exported_md = npars_exported_md,
        loc_per_fn_r_mn = loc_per_fn_r_mn,
        loc_per_fn_r_md = loc_per_fn_r_md,
        loc_per_fn_r_exp_mn = loc_per_fn_r_exp_mn,
        loc_per_fn_r_exp_md = loc_per_fn_r_exp_md,
        loc_per_fn_r_not_exp_mn = loc_per_fn_r_not_exp_mn,
        loc_per_fn_r_not_exp_md = loc_per_fn_r_not_exp_md,
        loc_per_fn_src_mn = loc_per_fn_src_mn,
        loc_per_fn_src_md = loc_per_fn_src_md,
        languages = languages,
        doclines_per_fn_exp_mn = doclines_per_fn_exp_mn,
        doclines_per_fn_exp_md = doclines_per_fn_exp_md,
        doclines_per_fn_not_exp_mn = doclines_per_fn_not_exp_mn,
        doclines_per_fn_not_exp_md = doclines_per_fn_not_exp_md,
        doclines_per_fn_src_mn = doclines_per_fn_src_mn,
        doclines_per_fn_src_md = doclines_per_fn_src_md,
        docchars_per_par_exp_mn = docchars_per_par_exp_mn,
        docchars_per_par_exp_md = docchars_per_par_exp_md,
        stringsAsFactors = FALSE
    )

}

#' @param x the 'network' components of 'pkgstats' output
#' @noRd
network_summary <- function (x) {

    n_clusters <- length (unique (x$cluster_dir))
    n_edges <- ifelse (is.null (x), 0L, nrow (x))

    dirs <- vapply (x$file, function (i) {
        strsplit (i, .Platform$file.sep) [[1]] [1]
    },
    character (1),
    USE.NAMES = FALSE
    )
    n_edges_r <- length (which (dirs == "R"))
    n_edges_src <- length (which (dirs != "R"))

    centrality_dir_mn <-
        centrality_dir_mn_no0 <-
        centrality_dir_md <-
        centrality_dir_md_no0 <-
        centrality_undir_mn <-
        centrality_undir_mn_no0 <-
        centrality_undir_md <-
        centrality_undir_md_no0 <- NA_integer_
    num_terminal_edges_dir <-
        num_terminal_edges_undir <- NA_integer_

    if (nrow (x) > 0) {

        centrality_dir_mn <- mean (x$centrality_dir, na.rm = TRUE)
        centrality_dir_mn_no0 <- mean (x$centrality_dir [x$centrality_dir > 0],
            na.rm = TRUE
        )
        centrality_dir_md <- stats::median (x$centrality_dir, na.rm = TRUE)
        centrality_dir_md_no0 <-
            stats::median (x$centrality_dir [x$centrality_dir > 0],
                na.rm = TRUE
            )

        cu <- x$centrality_undir [x$centrality_undir > 0]
        centrality_undir_mn <- mean (x$centrality_undir, na.rm = TRUE)
        centrality_undir_mn_no0 <- mean (cu, na.rm = TRUE)
        centrality_undir_md <- stats::median (x$centrality_undir, na.rm = TRUE)
        centrality_undir_md_no0 <- stats::median (cu, na.rm = TRUE)

        num_terminal_edges_dir <- length (which (x$centrality_dir == 0))
        num_terminal_edges_undir <- length (which (x$centrality_undir == 0))
    }

    from <- to <- NULL # suppress no visible binding notes
    node_degree <-
        node_degree_mn <-
        node_degree_md <-
        node_degree_max <- 0L

    if (nrow (x) > 0) {

        node_degree <- dplyr::group_by (x, from)
        node_degree <- dplyr::count (node_degree, to)
        node_degree_mn <- mean (node_degree$n)
        node_degree_md <- stats::median (node_degree$n)
        node_degree_max <- max (node_degree$n)
    }

    data.frame (
        n_edges = n_edges,
        n_edges_r = n_edges_r,
        n_edges_src = n_edges_src,
        n_clusters = n_clusters,
        centrality_dir_mn = centrality_dir_mn,
        centrality_dir_md = centrality_dir_md,
        centrality_dir_mn_no0 = centrality_dir_mn_no0,
        centrality_dir_md_no0 = centrality_dir_md_no0,
        centrality_undir_mn = centrality_undir_mn,
        centrality_undir_md = centrality_undir_md,
        centrality_undir_mn_no0 = centrality_undir_mn_no0,
        centrality_undir_md_no0 = centrality_undir_md_no0,
        num_terminal_edges_dir = num_terminal_edges_dir,
        num_terminal_edges_undir = num_terminal_edges_undir,
        node_degree_mn = node_degree_mn,
        node_degree_md = node_degree_md,
        node_degree_max = node_degree_max,
        stringsAsFactors = FALSE
    )
}

#' @param x the 'external_calls' components of 'pkgstats' output
#' @noRd
external_call_summary <- function (x) {

    if (is.null (x)) {
        return (NA_character_)
    }

    # summarise total number of calls to each package, plus number of distinct
    # functions from each:
    package <- NULL # suppress no visible binding note
    n <- dplyr::group_by (x, package, call)
    n <- dplyr::tally (n)
    n <- dplyr::summarise (n,
        ntot = sum (n),
        nunique = length (n)
    )

    res <- apply (n, 1, function (i) {
        paste0 (i, collapse = ":")
    })
    # tibbles put white space in there:
    res <- gsub ("\\s*", "", res)

    res <- data.frame (
        external_calls = paste0 (res, collapse = ","),
        stringsAsFactors = FALSE
    )

    return (res)
}

#' Package-internal metric of coupling instability between files of
#' the package.
#'
#' See 'https://en.wikipedia.org/wiki/Software_package_metrics'
#' Instability within a package is measured here as the proportion of all
#' function calls (or object references) within a package which are made
#' *between* different files compared with total numbers of calls. Packages with
#' relatively more calls confined to within single files are thus more stable,
#' while packages with more calls between different files are less stable.
#' Because a package is a strictly closed system, total numbers of afferent or
#' incoming calls must always equal numbers of efferent or outgoing calls.
#'
#' @inheritParams pkgstats_summary
#' @return Single value for coupling instability
#' @noRd
coupling_instability <- function (s) {

    file_to <- s$objects$file_name [match (s$network$to, s$objects$fn_name)]

    ntot <- nrow (s$network)
    nbetween <- length (which (s$network$file != file_to))

    return (nbetween / ntot)
}

fn_var_call_graph_r <- function (fns, fn_vars, path) {

    res <- NULL

    for (f in unique (fns$file)) {

        fns_f <- fns [fns$file == f, ]
        fns_f <- fns_f [order (fns_f$start), c ("tag", "start", "end")]
        fns_f <- fns_f [which (!is.na (fns_f$start) & !is.na (fns_f$end)), ]
        if (nrow (fns_f) < 1L) {
            next
        }

        fns_index <- lapply (seq (nrow (fns_f)), function (i) {
            cbind (i, seq (fns_f$start [i], fns_f$end [i]))
        })
        fns_index <- do.call (rbind, fns_index)

        f_full <- expand_path (fs::path (path, f))

        p <- control_parse (file = f_full)
        if (methods::is (p, "simpleError")) {
            next
        }

        pd <- utils::getParseData (control_parse (file = f_full))

        fn_calls <- pd [pd$text %in% fns$tag &
            pd$token == "SYMBOL_FUNCTION_CALL", ]
        index <- match (fn_calls$line1, fns_index [, 2])
        fn_calls$fns_index <- fns_index [index, 1]
        fn_calls$name <- fns_f$tag [fn_calls$fns_index]

        if (nrow (fn_calls) > 0) {

            res <- rbind (
                res,
                data.frame (
                    file = f,
                    line1 = fn_calls$line1,
                    line2 = fn_calls$line2,
                    from = fn_calls$name,
                    to = fn_calls$text,
                    language = "R",
                    stringsAsFactors = FALSE
                )
            )
        }
    }

    return (res)
}

fn_var_call_graph_src <- function (gtags) {

    data.frame (
        file = gtags$file,
        line1 = gtags$line,
        line2 = gtags$line,
        from = gtags$from,
        to = gtags$tag,
        language = gtags$language,
        stringsAsFactors = FALSE
    )
}

add_igraph_stats <- function (g, directed = TRUE) {

    g_igr <- igraph::graph_from_data_frame (g [, c ("from", "to")],
        directed = directed
    )

    cl <- igraph::components (g_igr)
    index <- match (g$from, names (cl$membership))
    if (directed) {
        nms <- c ("cluster_dir", "centrality_dir")
    } else {
        nms <- c ("cluster_undir", "centrality_undir")
    }
    g [nms [1]] <- cl$membership [index]
    btw <- igraph::betweenness (g_igr)
    g [nms [2]] <- btw [match (g$from, names (btw))]

    return (g)
}

---
title: "Package Statistics"
author: 
  - "Mark Padgham"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{Package Statistics}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set (
    collapse = TRUE,
    warning = TRUE,
    message = TRUE,
    width = 120,
    comment = "#>",
    fig.retina = 2,
    fig.path = "README-"
)
options (repos = c (
    ropenscireviewtools = "https://ropensci-review-tools.r-universe.dev",
    CRAN = "https://cloud.r-project.org"
))
library (pkgstats)
```

This vignette describes the statistics collated by `pkgstats`. The first
section provides full descriptions of all data returned by [the main
`pkgstats()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats.html), and the
second section describes the output of [the `pkgstats_summary()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats_summary.html)
which converts statistics to a single-row summary. Single-row summaries from
multiple packages can be combined to represent the statistical properties of
multiple packages in a single `data.frame` object. [The `pkgstats()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats.html) is
applied to all CRAN packages on a regular basis, with results accessible with
[the `dl_pkgstats_data()`
function](https://docs.ropensci.org/pkgstats/reference/dl_pkgstats_data.html).

## Overview of Package Statistics

The [main `pkgstats()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats.html) returns a
`list` of eight main components:

1. `"loc"` summarising "Lines of Code" in package sub-directories and languages;
2. `"vignettes"` containing counts of numbers of vignettes and demos;
3. `"data_stats"` summarising data files;
4. `"desc"` summarising the contents of the package "DESCRIPTION" file;
5. `"translations"` summarising translations into other (human) languages;
6. `"objects"`: a table of all "objects" in all languages;
7. `"network"`: a table of relationships between objects, such as function calls; and
8. `"external_calls"`: a detailed table of all calls made to all R functions.

The following sub-sections provide further detail on these components (except
the simpler components of `"vignettes"`, `"data_stats"`, and `"translations"`).
The results use the output of applying the function to the source code of this package:

```{r main-pkgstats-call, eval = FALSE}
s <- pkgstats () # run in root directory of `pkgstats` source
```
```{r pkgstats-fakey, echo = FALSE}
# These data all have to be faked because they can't be generated on CRAN
# windows machines.
loc <- tibble::tibble (
    language = c ("C++", "R", "R", "Rmd"),
    dir = c ("src", "R", "tests", "vignettes"),
    nfiles = c (3L, 24L, 7L, 2L),
    nlines = c (364L, 4727L, 300L, 347L),
    ncode = c (276L, 345L, 234L, 278L),
    nempty = c (67L, 682L, 61L, 61L),
    nspaces = c (932L, 333L, 511L, 1483L),
    nchars = c (6983L, 114334L, 5543L, 11290L),
    nexpr = c (1L, 1L, 1L, 1L),
    ntabs = c (0L, 0L, 0L, 0L),
    indentation = c (4L, 4L, 4L, 4L)
)
vignettes <- c (vignettes = 2L, demos = 0L)
data_stats <- c (n = 0L, total_size = 0L, median_size = 0L)
desc <- data.frame (
    package = "pkgstats",
    verion = "0.1.1",
    date = date (),
    license = "GPL-3",
    urls = paste0 (c (
        "https://docs.ropensci.org/pkgstats/",
        "nhttps://github.com/ropensci-review-tools/pkgstats"
    ), collapse = ",\n"),
    bugs = "https://github.com/ropensci-review-tools/pkgstats/issues",
    aut = 1L,
    ctb = 0L,
    fnd = 0L,
    rev = 0L,
    ths = 0L,
    trl = 0L,
    depends = NA,
    imports = paste0 (c (
        "brio",
        "checkmate",
        "dplyr",
        "fs",
        "igraph",
        "methods",
        "readr",
        "sys",
        "withr"
    ), collapse = ", "),
    suggests = paste0 (c (
        "curl",
        "hms",
        "jsonlite",
        "knitr",
        "parallel",
        "pkgbuild",
        "Rcpp",
        "rmarkdown",
        "roxygen2",
        "testthat",
        "visNetwork"
    ), collapse = ", "),
    enchances = NA_character_,
    linking_to = "cpp11"
)
translations <- NA_character_

s <- list (
    loc = loc,
    vignettes = vignettes,
    data_stats = data_stats,
    desc = desc,
    translations = translations
)
```


The result is a list of various data extracted from the code. All except for
`objects` and `network` represent summary data:

```{r}
s [!names (s) %in% c ("objects", "network", "external_calls")]
```

These results demonstrate that many fields use `NA` to denote values of zero.
The following sub-sections explore these various components generated by the
`pkgstats()` function in more detail.

## Lines of Code

The first item in the above list is "loc" for Lines-of-Code, which are  counted
using an internal routine specifically developed for R packages, and which
provides more accurate and R-specific information than most open source code
counting libraries. For example, the counts in `pkgstats` are able to
distinguish and separately count code chunks and text lines in `.Rmd` files. 

```{r loc}
s$loc
```

That output includes the following components, grouped by both computer
language and package directory:

1. `nfiles` = Numbers of files in each directory and language.
2. `nlines` = Total numbers of lines in all files.
3. `ncode` = Total numbers of lines of code.
4. `ndoc` = Total numbers of documentation or comment lines.
5. `nempty` = Total numbers of empty of blank lines.
6. `nspaces` = Total numbers of white spaces in all code lines, excluding
   leading indentation spaces.
7. `nchars` = Total numbers of non-white-space characters in all code lines.
8. `nexpr` = Median numbers of nested expressions in all lines which have any
   expressions (see below).
9. `ntabs` = Number of lines of code with initial tab indentation.
10. `indentation` = Number of spaces by which code is indented (with `-1`
   denoting tab-indentation).

Numbers of nested expressions are counted as numbers of brackets or braces of
any type nested on a single line. The following line has one nested bracket:

```{r nested1, eval = FALSE}
x <- myfn ()
```

while the following has four:

```{r nested4, eval = FALSE}
x <- function () { return (myfn ()) }
```

Code with fewer nested expressions per line is generally easier to read, and
this metric is provided as one indication of the general readability of
code. A second relative indication may be extracted by converting numbers of
spaces and characters to a measure of relative numbers of white spaces, noting
that the `nchars` value quantifies total characters including white spaces.

```{r rel-space}
index <- which (s$loc$dir %in% c ("R", "src")) # consider source code only
sum (s$loc$nspaces [index]) / sum (s$loc$nchars [index])
```

Finally, the `ntabs` statistic can be used to identify whether code uses tab
characters as indentation, otherwise the `indentation` statistics indicate
median numbers of white spaces by which code is indented. The `objects`,
`network`, and `external_calls` items returned by the [`pkgstats()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats.html) are
described further below.

## `"desc"`: The package "DESCRIPTION" file

The `desc` item looks like this:

```{r desc}
s$desc
```

This item includes the following components:

- Package name, version, date, and license
- Package URL(s) (`urls`)
- URL for BugReports (`bugs`)
- Number of contributors with role of *author* (`desc_n_aut`), *contributor*
  (`desc_n_ctb`), *funder* (`desc_n_fnd`), *reviewer* (`desc_n_rev`), *thesis
  advisor* (`ths`), and *translator* (`trl`, relating to translation between
  computer and not spoken languages).
- Comma-separated character entries for all `depends`, `imports`, `suggests`,
  and `linking_to` packages.

The "Date" field is taken from the "Date/Publication" field automatically
inserted by CRAN on package publication, or for non-CRAN packages to the
"mtime" value (modification time) value of the DESCRIPTION file. Note that
"date" values extracted by `pkgstats` do not use "Date" values from DESCRIPTION
files (as these are manually-entered, and potentially unreliable).

### 1.3 `"objects"`: Objects in all languages

The `objects` item contains all code objects identified by
the code-tagging library [`ctags`](https://ctags.io). For R, those are
primarily functions, but for other languages may be a variety of entities such
as class or structure definitions, or sub-members thereof. Object tables look
like this:

```{r objects-fakey, echo = FALSE}
s$objects <- data.frame (
    file_name = c (
        rep ("R/archive-trawl.R", 4L),
        "R/cpp11.R",
        "R/ctags-install.R"
    ),
    fn_name = c (
        "pkgstats_from_archive",
        "list_archive_files",
        "rm_prev_files",
        "pkgstats_fns_from_archive",
        "cpp_loc",
        "clone_ctag"
    ),
    kind = rep ("function", 6L),
    language = rep ("R", 6L),
    loc = c (89, 17, 24, 82, 3, 17),
    npars = c (9, 2, 2, 7, 4, 1),
    has_dots = rep (FALSE, 6L),
    exported = rep (c (TRUE, FALSE, FALSE), 2L),
    param_nchards_md = c (
        133,
        rep (NA_integer_, 2L),
        163,
        rep (NA_integer_, 2L)
    ),
    param_nchards_mn = c (
        159.7778,
        rep (NA_integer_, 2L),
        174.5714,
        rep (NA_integer_, 2L)
    ),
    num_doclines = c (
        77,
        rep (NA_integer_, 2L),
        50,
        rep (NA_integer_, 2L)
    )
)
```


```{r objects}
head (s$objects)
```

Objects are primarily sorted by language, with R-language objects given first.
These are mostly functions, and include statistics on:

- lines of code used to define each function (`loc`);
- numbers of parameters (`npars`);
- whether or not the function includes a "three dots" parameter (that is,
  `...`; identified by `has_dots`);
- whether or not a function is exported (`exported`);
- Mean and median numbers of character used to document each parameter
  (`param_nchards_mn` and `param_nchards_md`, respectively); and
- Total number of lines of documentation for that object / function.


## `"network"`: Relationships between objects

The `network` item details all relationships between objects, which generally
reflects one object calling or otherwise depending on another object. Each row
thus represents one edge of a "function call" network, with each entry in the
`from` and `to` columns representing the network vertices or nodes.

```{r network-fakey, echo = FALSE}
network <- data.frame (
    file = c (
        rep ("R/external_calls.R", 4L),
        rep ("R/pkgstats-summary.R", 2L)
    ),
    line1 = c (11L, 26L, 38L, 326L, 39L, 50L),
    from = c (
        rep ("external_call_network", 3L),
        "add_other_pkgs_to_calls",
        rep ("pkgstats_summary", 2L)
    ),
    to = c (
        "extract_call_content",
        "add_base_recommended_pkgs",
        "add_other_pkgs_to_calls",
        "control_parse",
        "null_stats",
        "loc_summary"
    ),
    language = rep ("R", 6L),
    cluster_dir = rep (1L, 6L),
    centrality_dir = c (9L, 9L, 9L, 1L, 11L, 11L),
    cluster_undir = rep (1L, 6L),
    centrality_undir = c (rep (230.8333, 3L), 6, rep (874, 2L))
)

nrows <- 142 # full number in result
# expand to nrows:
n <- ceiling (nrows / nrow (network))
network <- with (network, data.frame (
    file = rep (file, n),
    line1 = rep (line1, n),
    from = rep (from, n),
    to = rep (to, n),
    language = rep (language, n),
    cluster_dir = rep (cluster_dir, n),
    centrality_dir = rep (centrality_dir, n),
    cluster_undir = rep (cluster_undir, n),
    centrality_undir = rep (centrality_undir, n)
))
s$network <- network [seq (nrows), ]
```


```{r}
head (s$network)
nrow (s$network)
```

The network table includes additional statistics on the centrality of each
edge, measured as betweenness centrality assuming edges to be both directed
(`centrality_dir`) and undirected (`centrality_undir`). More central edges
reflect connections between objects that are more central to package
functionality, and vice versa. The distinct components of the network are also
represented by discrete cluster numbers, calculated both for directed and
undirected versions of the network. Each distinct cluster number represents
a distinct group of objects, internally related to other members of the same
cluster, yet independent of all objects with different cluster numbers.

The network can be viewed as an interactive [`vis.js`](https://visjs.org/)
network through passing the result of `pkgstats` -- the variable `p` in the
code above -- to the [`plot_network()`
function](https://docs.ropensci.org/pkgstats/reference/plot_network.html).

## `"external_calls"`: All calls made to all R functions

The `external_calls` item is structured similar to the `network` object, but
identifies all calls to functions from external packages. However, unlike the
`network` and `object` data, which provide information on objects and
relationships in all computer languages used within a package, the
`external_calls` object maps calls within R code only, in order to provide
insight into the use within a package of of functions from other packages,
including R's base and recommended packages. The object looks like this:

```{r external_calls-fakey, echo = FALSE}
s$external_calls <- data.frame (
    tags_line = 1:6,
    call = c (
        "c",
        rep ("character", 2L),
        "logical",
        "integer",
        "left_join"
    ),
    tag = c (
        "GTAGSLABEL",
        "file_name",
        "fn_name",
        "has_dots",
        "loc",
        "name"
    ),
    file = c (
        "R/ctags-test.R",
        rep ("R/pkgstats.R", 4L),
        "R/plot.R"
    ),
    kind = rep ("nameattr", 6L),
    start = c (109L, 185L, 186L, 189L, 187L, 89L),
    end = c (109L, 185L, 186L, 189L, 187L, 89L),
    package = c (rep ("base", 5L), "dplyr")
)
```


```{r ext-call-head}
head (s$external_calls)
```

These data are converted to a summary form by the [`pkgstats_summary()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats_summary.html),
which tabulates numbers of external calls and unique functions from each
package. These data are presented as a single character string which looks like
this:

```{r ext-call-summary, eval = FALSE}
s_summ <- pkgstats_summary (s)
print (s_summ$external_calls)
```
```{r ext-call-sumamry-fakey, echo = FALSE}
s_summ <- list (external_calls = paste0 (c (
    "base:581:84",
    "brio:11:2",
    "curl:4:3",
    "dplyr:7:4",
    "fs:4:2",
    "graphics:10:2",
    "hms:2:1",
    "igraph:3:3",
    "parallel:2:1",
    "pkgstats:126:73",
    "readr:8:5",
    "stats:19:3",
    "sys:14:1",
    "tools:3:2",
    "utils:22:7",
    "visNetwork:3:2",
    "withr:6:2"
), collapse = ","))
```

These data can be easily converted to the corresponding numeric values using
code like the following:

```{r ext-call-details, eval = TRUE}
x <- strsplit (s_summ$external_calls, ",") [[1]]
x <- do.call (rbind, strsplit (x, ":"))
x <- data.frame (
    pkg = x [, 1],
    n_total = as.integer (x [, 2]),
    n_unique = as.integer (x [, 3])
)
x$n_total_rel <- round (x$n_total / sum (x$n_total), 3)
x$n_unique_rel <- round (x$n_unique / sum (x$n_unique), 3)
print (x)
```

Those data reveal, for example, that this package makes 
`r x$n_total [x$pkg == "base"]` individual calls to
`r x$n_unique [x$pkg == "base"]` unique functions from the "base" package.

# Package Statistics

This vignette describes the statistics collated by `pkgstats`. The first
section provides full descriptions of all data returned by [the main
`pkgstats()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats.html),
and the second section describes the output of [the `pkgstats_summary()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats_summary.html)
which converts statistics to a single-row summary. Single-row summaries
from multiple packages can be combined to represent the statistical
properties of multiple packages in a single `data.frame` object. [The
`pkgstats()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats.html) is
applied to all CRAN packages on a regular basis, with results accessible
with [the `dl_pkgstats_data()`
function](https://docs.ropensci.org/pkgstats/reference/dl_pkgstats_data.html).

## Overview of Package Statistics

The [main `pkgstats()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats.html)
returns a `list` of eight main components:

1.  `"loc"` summarising “Lines of Code” in package sub-directories and
    languages;
2.  `"vignettes"` containing counts of numbers of vignettes and demos;
3.  `"data_stats"` summarising data files;
4.  `"desc"` summarising the contents of the package “DESCRIPTION” file;
5.  `"translations"` summarising translations into other (human)
    languages;
6.  `"objects"`: a table of all “objects” in all languages;
7.  `"network"`: a table of relationships between objects, such as
    function calls; and
8.  `"external_calls"`: a detailed table of all calls made to all R
    functions.

The following sub-sections provide further detail on these components
(except the simpler components of `"vignettes"`, `"data_stats"`, and
`"translations"`). The results use the output of applying the function
to the source code of this package:

``` r
s <- pkgstats () # run in root directory of `pkgstats` source
```

The result is a list of various data extracted from the code. All except
for `objects` and `network` represent summary data:

``` r
s [!names (s) %in% c ("objects", "network", "external_calls")]
#> $loc
#> # A tibble: 4 × 11
#>   language dir       nfiles nlines ncode nempty nspaces nchars nexpr ntabs
#>   <chr>    <chr>      <int>  <int> <int>  <int>   <int>  <int> <int> <int>
#> 1 C++      src            3    364   276     67     932   6983     1     0
#> 2 R        R             24   4727   345    682     333 114334     1     0
#> 3 R        tests          7    300   234     61     511   5543     1     0
#> 4 Rmd      vignettes      2    347   278     61    1483  11290     1     0
#> # ℹ 1 more variable: indentation <int>
#> 
#> $vignettes
#> vignettes     demos 
#>         2         0 
#> 
#> $data_stats
#>           n  total_size median_size 
#>           0           0           0 
#> 
#> $desc
#>    package verion                     date license
#> 1 pkgstats  0.1.1 Fri Feb 13 11:58:10 2026   GPL-3
#>                                                                                       urls
#> 1 https://docs.ropensci.org/pkgstats/,\nnhttps://github.com/ropensci-review-tools/pkgstats
#>                                                       bugs aut ctb fnd rev ths
#> 1 https://github.com/ropensci-review-tools/pkgstats/issues   1   0   0   0   0
#>   trl depends                                                        imports
#> 1   0      NA brio, checkmate, dplyr, fs, igraph, methods, readr, sys, withr
#>                                                                                          suggests
#> 1 curl, hms, jsonlite, knitr, parallel, pkgbuild, Rcpp, rmarkdown, roxygen2, testthat, visNetwork
#>   enchances linking_to
#> 1      <NA>      cpp11
#> 
#> $translations
#> [1] NA
```

These results demonstrate that many fields use `NA` to denote values of
zero. The following sub-sections explore these various components
generated by the `pkgstats()` function in more detail.

## Lines of Code

The first item in the above list is “loc” for Lines-of-Code, which are
counted using an internal routine specifically developed for R packages,
and which provides more accurate and R-specific information than most
open source code counting libraries. For example, the counts in
`pkgstats` are able to distinguish and separately count code chunks and
text lines in `.Rmd` files.

``` r
s$loc
#> # A tibble: 4 × 11
#>   language dir       nfiles nlines ncode nempty nspaces nchars nexpr ntabs
#>   <chr>    <chr>      <int>  <int> <int>  <int>   <int>  <int> <int> <int>
#> 1 C++      src            3    364   276     67     932   6983     1     0
#> 2 R        R             24   4727   345    682     333 114334     1     0
#> 3 R        tests          7    300   234     61     511   5543     1     0
#> 4 Rmd      vignettes      2    347   278     61    1483  11290     1     0
#> # ℹ 1 more variable: indentation <int>
```

That output includes the following components, grouped by both computer
language and package directory:

1.  `nfiles` = Numbers of files in each directory and language.
2.  `nlines` = Total numbers of lines in all files.
3.  `ncode` = Total numbers of lines of code.
4.  `ndoc` = Total numbers of documentation or comment lines.
5.  `nempty` = Total numbers of empty of blank lines.
6.  `nspaces` = Total numbers of white spaces in all code lines,
    excluding leading indentation spaces.
7.  `nchars` = Total numbers of non-white-space characters in all code
    lines.
8.  `nexpr` = Median numbers of nested expressions in all lines which
    have any expressions (see below).
9.  `ntabs` = Number of lines of code with initial tab indentation.
10. `indentation` = Number of spaces by which code is indented (with
    `-1` denoting tab-indentation).

Numbers of nested expressions are counted as numbers of brackets or
braces of any type nested on a single line. The following line has one
nested bracket:

``` r
x <- myfn ()
```

while the following has four:

``` r
x <- function () { return (myfn ()) }
```

Code with fewer nested expressions per line is generally easier to read,
and this metric is provided as one indication of the general readability
of code. A second relative indication may be extracted by converting
numbers of spaces and characters to a measure of relative numbers of
white spaces, noting that the `nchars` value quantifies total characters
including white spaces.

``` r
index <- which (s$loc$dir %in% c ("R", "src")) # consider source code only
sum (s$loc$nspaces [index]) / sum (s$loc$nchars [index])
#> [1] 0.01042723
```

Finally, the `ntabs` statistic can be used to identify whether code uses
tab characters as indentation, otherwise the `indentation` statistics
indicate median numbers of white spaces by which code is indented. The
`objects`, `network`, and `external_calls` items returned by the
[`pkgstats()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats.html)
are described further below.

## `"desc"`: The package “DESCRIPTION” file

The `desc` item looks like this:

``` r
s$desc
#>    package verion                     date license
#> 1 pkgstats  0.1.1 Fri Feb 13 11:58:10 2026   GPL-3
#>                                                                                       urls
#> 1 https://docs.ropensci.org/pkgstats/,\nnhttps://github.com/ropensci-review-tools/pkgstats
#>                                                       bugs aut ctb fnd rev ths
#> 1 https://github.com/ropensci-review-tools/pkgstats/issues   1   0   0   0   0
#>   trl depends                                                        imports
#> 1   0      NA brio, checkmate, dplyr, fs, igraph, methods, readr, sys, withr
#>                                                                                          suggests
#> 1 curl, hms, jsonlite, knitr, parallel, pkgbuild, Rcpp, rmarkdown, roxygen2, testthat, visNetwork
#>   enchances linking_to
#> 1      <NA>      cpp11
```

This item includes the following components:

- Package name, version, date, and license
- Package URL(s) (`urls`)
- URL for BugReports (`bugs`)
- Number of contributors with role of *author* (`desc_n_aut`),
  *contributor* (`desc_n_ctb`), *funder* (`desc_n_fnd`), *reviewer*
  (`desc_n_rev`), *thesis advisor* (`ths`), and *translator* (`trl`,
  relating to translation between computer and not spoken languages).
- Comma-separated character entries for all `depends`, `imports`,
  `suggests`, and `linking_to` packages.

The “Date” field is taken from the “Date/Publication” field
automatically inserted by CRAN on package publication, or for non-CRAN
packages to the “mtime” value (modification time) value of the
DESCRIPTION file. Note that “date” values extracted by `pkgstats` do not
use “Date” values from DESCRIPTION files (as these are manually-entered,
and potentially unreliable).

### 1.3 `"objects"`: Objects in all languages

The `objects` item contains all code objects identified by the
code-tagging library [`ctags`](https://ctags.io). For R, those are
primarily functions, but for other languages may be a variety of
entities such as class or structure definitions, or sub-members thereof.
Object tables look like this:

``` r
head (s$objects)
#>           file_name                   fn_name     kind language loc npars
#> 1 R/archive-trawl.R     pkgstats_from_archive function        R  89     9
#> 2 R/archive-trawl.R        list_archive_files function        R  17     2
#> 3 R/archive-trawl.R             rm_prev_files function        R  24     2
#> 4 R/archive-trawl.R pkgstats_fns_from_archive function        R  82     7
#> 5         R/cpp11.R                   cpp_loc function        R   3     4
#> 6 R/ctags-install.R                clone_ctag function        R  17     1
#>   has_dots exported param_nchards_md param_nchards_mn num_doclines
#> 1    FALSE     TRUE              133         159.7778           77
#> 2    FALSE    FALSE               NA               NA           NA
#> 3    FALSE    FALSE               NA               NA           NA
#> 4    FALSE     TRUE              163         174.5714           50
#> 5    FALSE    FALSE               NA               NA           NA
#> 6    FALSE    FALSE               NA               NA           NA
```

Objects are primarily sorted by language, with R-language objects given
first. These are mostly functions, and include statistics on:

- lines of code used to define each function (`loc`);
- numbers of parameters (`npars`);
- whether or not the function includes a “three dots” parameter (that
  is, `...`; identified by `has_dots`);
- whether or not a function is exported (`exported`);
- Mean and median numbers of character used to document each parameter
  (`param_nchards_mn` and `param_nchards_md`, respectively); and
- Total number of lines of documentation for that object / function.

## `"network"`: Relationships between objects

The `network` item details all relationships between objects, which
generally reflects one object calling or otherwise depending on another
object. Each row thus represents one edge of a “function call” network,
with each entry in the `from` and `to` columns representing the network
vertices or nodes.

``` r
head (s$network)
#>                   file line1                    from                        to
#> 1   R/external_calls.R    11   external_call_network      extract_call_content
#> 2   R/external_calls.R    26   external_call_network add_base_recommended_pkgs
#> 3   R/external_calls.R    38   external_call_network   add_other_pkgs_to_calls
#> 4   R/external_calls.R   326 add_other_pkgs_to_calls             control_parse
#> 5 R/pkgstats-summary.R    39        pkgstats_summary                null_stats
#> 6 R/pkgstats-summary.R    50        pkgstats_summary               loc_summary
#>   language cluster_dir centrality_dir cluster_undir centrality_undir
#> 1        R           1              9             1         230.8333
#> 2        R           1              9             1         230.8333
#> 3        R           1              9             1         230.8333
#> 4        R           1              1             1           6.0000
#> 5        R           1             11             1         874.0000
#> 6        R           1             11             1         874.0000
nrow (s$network)
#> [1] 142
```

The network table includes additional statistics on the centrality of
each edge, measured as betweenness centrality assuming edges to be both
directed (`centrality_dir`) and undirected (`centrality_undir`). More
central edges reflect connections between objects that are more central
to package functionality, and vice versa. The distinct components of the
network are also represented by discrete cluster numbers, calculated
both for directed and undirected versions of the network. Each distinct
cluster number represents a distinct group of objects, internally
related to other members of the same cluster, yet independent of all
objects with different cluster numbers.

The network can be viewed as an interactive
[`vis.js`](https://visjs.org/) network through passing the result of
`pkgstats` – the variable `p` in the code above – to the
[`plot_network()`
function](https://docs.ropensci.org/pkgstats/reference/plot_network.html).

## `"external_calls"`: All calls made to all R functions

The `external_calls` item is structured similar to the `network` object,
but identifies all calls to functions from external packages. However,
unlike the `network` and `object` data, which provide information on
objects and relationships in all computer languages used within a
package, the `external_calls` object maps calls within R code only, in
order to provide insight into the use within a package of of functions
from other packages, including R’s base and recommended packages. The
object looks like this:

``` r
head (s$external_calls)
#>   tags_line      call        tag           file     kind start end package
#> 1         1         c GTAGSLABEL R/ctags-test.R nameattr   109 109    base
#> 2         2 character  file_name   R/pkgstats.R nameattr   185 185    base
#> 3         3 character    fn_name   R/pkgstats.R nameattr   186 186    base
#> 4         4   logical   has_dots   R/pkgstats.R nameattr   189 189    base
#> 5         5   integer        loc   R/pkgstats.R nameattr   187 187    base
#> 6         6 left_join       name       R/plot.R nameattr    89  89   dplyr
```

These data are converted to a summary form by the [`pkgstats_summary()`
function](https://docs.ropensci.org/pkgstats/reference/pkgstats_summary.html),
which tabulates numbers of external calls and unique functions from each
package. These data are presented as a single character string which
looks like this:

``` r
s_summ <- pkgstats_summary (s)
print (s_summ$external_calls)
```

These data can be easily converted to the corresponding numeric values
using code like the following:

``` r
x <- strsplit (s_summ$external_calls, ",") [[1]]
x <- do.call (rbind, strsplit (x, ":"))
x <- data.frame (
    pkg = x [, 1],
    n_total = as.integer (x [, 2]),
    n_unique = as.integer (x [, 3])
)
x$n_total_rel <- round (x$n_total / sum (x$n_total), 3)
x$n_unique_rel <- round (x$n_unique / sum (x$n_unique), 3)
print (x)
#>           pkg n_total n_unique n_total_rel n_unique_rel
#> 1        base     581       84       0.704        0.426
#> 2        brio      11        2       0.013        0.010
#> 3        curl       4        3       0.005        0.015
#> 4       dplyr       7        4       0.008        0.020
#> 5          fs       4        2       0.005        0.010
#> 6    graphics      10        2       0.012        0.010
#> 7         hms       2        1       0.002        0.005
#> 8      igraph       3        3       0.004        0.015
#> 9    parallel       2        1       0.002        0.005
#> 10   pkgstats     126       73       0.153        0.371
#> 11      readr       8        5       0.010        0.025
#> 12      stats      19        3       0.023        0.015
#> 13        sys      14        1       0.017        0.005
#> 14      tools       3        2       0.004        0.010
#> 15      utils      22        7       0.027        0.036
#> 16 visNetwork       3        2       0.004        0.010
#> 17      withr       6        2       0.007        0.010
```

Those data reveal, for example, that this package makes 581 individual
calls to 84 unique functions from the “base” package.

---
title: "Installation"
author: 
  - "Mark Padgham"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{Installation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set (
    collapse = TRUE,
    warning = TRUE,
    message = TRUE,
    width = 120,
    comment = "#>",
    fig.retina = 2,
    fig.path = "README-"
)
options (repos = c (
    ropenscireviewtools = "https://ropensci-review-tools.r-universe.dev",
    CRAN = "https://cloud.r-project.org"
))
```

The `pkgstats` package [is on
CRAN](https://cran.r-project.org/package=pkgstats), so can be installed
directly with,

```{r install, eval = FALSE}
install.packages ("pkgstats")
```

The latest development version can be installed [via the associated
`r-universe`](https://ropensci-review-tools.r-universe.dev/pkgstats). As
shown there, simply enable the universe with

```{r options, eval = FALSE}
options (repos = c (
    ropenscireviewtools = "https://ropensci-review-tools.r-universe.dev",
    CRAN = "https://cloud.r-project.org"
))
```

And then call `install.packages()` the same way. Alternatively, the development
version of the package can be installed by running one of the following lines:


```{r remotes, eval = FALSE}
remotes::install_github ("ropensci-review-tools/pkgstats")
pak::pkg_install ("ropensci-review-tools/pkgstats")
```

For those who prefer not to use GitHub, the package can also be installed with
one of the following two additional options:

```{r remotes-elsewhere, eval = FALSE}
remotes::install_git ("https://codeberg.org/ropensci-review-tools/pkgstats")
remotes::install_git ("https://codefloe.com/ropensci-review-tools/pkgstats")
```


The package can then loaded for use with:

```{r library, eval = TRUE}
library (pkgstats)
```

## Installation on Linux systems

This package requires the [system libraries
`ctags-universal`](https://ctags.io) and [GNU
`global`](https://www.gnu.org/software/global/), both of which are
automatically installed along with the package on both Windows and MacOS
systems. Most Linux distributions do not include a sufficiently up-to-date
version of [`ctags-universal`](https://ctags.io), and so it must be compiled
from source. This can be done by running a single function, `ctags_install()`,
which will install both [`ctags-universal`](https://ctags.io) and [GNU
`global`](https://www.gnu.org/software/global/).

The `pkgstats` package includes a function to ensure your local installations
of `universal-ctags` and `global` work correctly. Please ensure you see the
following prior to proceeding:

```{r ctags-check, eval = FALSE}
ctags_test ()
#> [1] TRUE
```

Note that GNU `global` can be linked at installation to the Universal Ctags
plug-in parser to expand the [default 5 languages to
30](https://www.gnu.org/software/global/). This makes no difference to
`pkgstats` results, as `gtags` output is only used to trace function call
networks, which is only possible for compiled languages able to dynamically
share pointers to the same objects. This is possible with the default parser
regardless. The wealth of extra information obtained from linking `global` to
the Universal Ctags parser is ultimately discarded anyway, yet parsing may take
considerably longer. If this is the case, "default" behaviour may be recovered
by first running the following command:

```{r unsetenv, eval = FALSE}
Sys.unsetenv (c ("GTAGSCONF", "GTAGSLABEL"))
```

# Installation on macOS systems

On macOS, [GNU `global` and `universal-ctags`](https://formulae.brew.sh/formula/global) can be easily installed with Homebrew. First follow the instructions at [https://brew.sh/](https://brew.sh/) to install Homebrew if you don't have it already. Then run the following command in your terminal:

```bash
brew install global
```

This will install both `global` and a compatible version of `universal-ctags`.

::: callout-note
Do not install `ctags` via Homebrew with `brew install ctags` as the installed version will conflict with `global`'s required `universal-ctags` dependency. If you have `ctags` already installed for another purpose, see Homebrew's messages in the console on how to proceed after attempting to intall `global`, it will usually involve unlinking the existing `ctags` with `brew unlink ctags` or removing it completely with `brew rm ctags`.
:::

The `pkgstats` package [is on
CRAN](https://cran.r-project.org/package=pkgstats), so can be installed
directly with,

``` r
install.packages ("pkgstats")
```

The latest development version can be installed [via the associated
`r-universe`](https://ropensci-review-tools.r-universe.dev/pkgstats). As
shown there, simply enable the universe with

``` r
options (repos = c (
    ropenscireviewtools = "https://ropensci-review-tools.r-universe.dev",
    CRAN = "https://cloud.r-project.org"
))
```

And then call `install.packages()` the same way. Alternatively, the
development version of the package can be installed by running one of
the following lines:

``` r
remotes::install_github ("ropensci-review-tools/pkgstats")
pak::pkg_install ("ropensci-review-tools/pkgstats")
```

For those who prefer not to use GitHub, the package can also be
installed with one of the following two additional options:

``` r
remotes::install_git ("https://codeberg.org/ropensci-review-tools/pkgstats")
remotes::install_git ("https://codefloe.com/ropensci-review-tools/pkgstats")
```

The package can then loaded for use with:

``` r
library (pkgstats)
```

## Installation on Linux systems

This package requires the [system libraries
`ctags-universal`](https://ctags.io) and [GNU
`global`](https://www.gnu.org/software/global/), both of which are
automatically installed along with the package on both Windows and MacOS
systems. Most Linux distributions do not include a sufficiently
up-to-date version of [`ctags-universal`](https://ctags.io), and so it
must be compiled from source. This can be done by running a single
function, `ctags_install()`, which will install both
[`ctags-universal`](https://ctags.io) and [GNU
`global`](https://www.gnu.org/software/global/).

The `pkgstats` package includes a function to ensure your local
installations of `universal-ctags` and `global` work correctly. Please
ensure you see the following prior to proceeding:

``` r
ctags_test ()
#> [1] TRUE
```

Note that GNU `global` can be linked at installation to the Universal
Ctags plug-in parser to expand the [default 5 languages to
30](https://www.gnu.org/software/global/). This makes no difference to
`pkgstats` results, as `gtags` output is only used to trace function
call networks, which is only possible for compiled languages able to
dynamically share pointers to the same objects. This is possible with
the default parser regardless. The wealth of extra information obtained
from linking `global` to the Universal Ctags parser is ultimately
discarded anyway, yet parsing may take considerably longer. If this is
the case, “default” behaviour may be recovered by first running the
following command:

``` r
Sys.unsetenv (c ("GTAGSCONF", "GTAGSLABEL"))
```

# Installation on macOS systems

On macOS, [GNU `global` and
`universal-ctags`](https://formulae.brew.sh/formula/global) can be
easily installed with Homebrew. First follow the instructions at
<https://brew.sh/> to install Homebrew if you don’t have it already.
Then run the following command in your terminal:

``` bash
brew install global
```

This will install both `global` and a compatible version of
`universal-ctags`.

<div class="callout-note">

Do not install `ctags` via Homebrew with `brew install ctags` as the
installed version will conflict with `global`’s required
`universal-ctags` dependency. If you have `ctags` already installed for
another purpose, see Homebrew’s messages in the console on how to
proceed after attempting to intall `global`, it will usually involve
unlinking the existing `ctags` with `brew unlink ctags` or removing it
completely with `brew rm ctags`.

</div>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgstats-summary.R
\name{pkgstats_summary}
\alias{pkgstats_summary}
\title{Condense the output of \code{pkgstats} to summary statistics only}
\usage{
pkgstats_summary(s = NULL)
}
\arguments{
\item{s}{Output of \link{pkgstats}, containing full statistical data on one
package. Default of \code{NULL} returns a single row with \code{NA} values (used in
\link{pkgstats_from_archive}).}
}
\value{
Summarised version of \code{s}, as a single row of a standardised
\code{data.frame} object
}
\description{
Condense the output of \code{pkgstats} to summary statistics only
}
\note{
Variable names in the summary object use the following abbreviations:
\itemize{
\item "loc" = Lines-of-Code
\item "fn" = Function
\item "n_fns" = Number of functions
\item "npars" = Number of parameters
\item "doclines" = Number of documentation lines
\item "nedges" = Number of edges in function call network, as a count of
\emph{unique} edges, which may be less than the size of the \code{network}
object returned by \link{pkgstats}, because that may include multiple calls
between identical function pairs.
\item "n_clusters" = Number of connected clusters within the function call
network.
\item "centrality" used as a prefix for several statistics, along with
"dir" or "undir" for centrality calculated on networks respectively
constructed with directed or undirected edges; "mn" or "md" for respective
measures of mean or median centrality, and "no0" for measures excluding
edges with zero centrality.
}
}
\examples{
f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
\dontshow{if (ctags_test (noerror = TRUE)) withAutoprint(\{ # examplesIf}
p <- pkgstats (f)
s <- pkgstats_summary (p)
\dontshow{\}) # examplesIf}
}
\seealso{
Other stats:
\code{\link[=desc_stats]{desc_stats()}},
\code{\link[=loc_stats]{loc_stats()}},
\code{\link[=pkgstats]{pkgstats()}},
\code{\link[=rd_stats]{rd_stats()}}
}
\concept{stats}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cran-data-current.R
\name{pkgstats_cran_current_from_full}
\alias{pkgstats_cran_current_from_full}
\title{Reduce \code{data.frame} of full CRAN archive data to current packages only.}
\usage{
pkgstats_cran_current_from_full(prev_results, results_file = NULL)
}
\arguments{
\item{prev_results}{Result of previous call to this function, if available.
Submitting previous results will ensure that only newer packages not present
in previous result will be analysed, with new results simply appended to
previous results. This parameter can also specify a file to be read with
\code{readRDS()}.}

\item{results_file}{Can be used to specify the name or full path of a \code{.Rds}
file to which results should be saved once they have been generated. The
'.Rds' extension will be automatically appended, and any other extensions
will be ignored.}
}
\value{
A \code{data.frame} object with one row for each package containing
summary statistics generated from the \link{pkgstats_summary} function.
}
\description{
Reduce \code{data.frame} of full CRAN archive data to current packages only.
}
\seealso{
Other archive:
\code{\link[=dl_pkgstats_data]{dl_pkgstats_data()}},
\code{\link[=pkgstats_fns_from_archive]{pkgstats_fns_from_archive()}},
\code{\link[=pkgstats_fns_update]{pkgstats_fns_update()}},
\code{\link[=pkgstats_from_archive]{pkgstats_from_archive()}},
\code{\link[=pkgstats_update]{pkgstats_update()}}
}
\concept{archive}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_network}
\alias{plot_network}
\title{Plot interactive visualisation of object-relationship network of package.}
\usage{
plot_network(s, plot = TRUE, vis_save = NULL)
}
\arguments{
\item{s}{Package statistics obtained from \link{pkgstats} function.}

\item{plot}{If \code{TRUE}, plot the network which opens an interactive browser
pane.}

\item{vis_save}{Name of local file in which to save \code{html} file of network
visualisation (will override \code{plot} to \code{FALSE}).}
}
\value{
(Invisibly) A \code{pkgstats_network} object: a list of the \code{nodes} and
\code{edges} used to construct the network, along with the self-contained
\code{html} document itself.
}
\description{
Plot interactive visualisation of object-relationship network of package.
}
\note{
Edge thicknesses are scaled to centrality within the package function
call network. Node sizes are scaled to numbers of times each function is
called from all other functions within a package. The visualisation is
rendered with a small bundle of \pkg{D3}-based JavaScript shipped with
this package (see \code{system.file ("js", package = "pkgstats")}), so no
internet connection or additional R packages are required to view it.
}
\examples{
f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
\donttest{
p <- pkgstats (f)
net <- plot_network (p, plot = FALSE)
}
# use default 'plot = TRUE' to automatically open network in browser.
}
\concept{output}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ctags-test.R
\name{ctags_test}
\alias{ctags_test}
\title{test a 'ctags' installation}
\usage{
ctags_test(quiet = TRUE, noerror = FALSE)
}
\arguments{
\item{quiet}{If \code{TRUE}, display on screen whether or not 'ctags' is correctly
installed.}

\item{noerror}{If \code{FALSE} (default), this function will error if either
'ctags' or 'gtags' are not installed. If \code{TRUE}, the function will complete
without erroring, and issue appropriate messages regarding required but
non-installed system libraries.}
}
\value{
'TRUE' or 'FALSE' respectively indicating whether or not 'ctags' is
correctly installed.
}
\description{
This uses the example from
\url{https://github.com/universal-ctags/ctags/blob/master/man/ctags-lang-r.7.rst.in}
and also checks the GNU global installation.
}
\examples{
# The function errors if not ctags or gtags found.
\donttest{
ctags_okay <- !is.null (tryCatch (
    ctags_test (),
    error = function (e) NULL
))
}
}
\seealso{
Other tags:
\code{\link[=ctags_install]{ctags_install()}},
\code{\link[=tags_data]{tags_data()}}
}
\concept{tags}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loc.R
\name{loc_stats}
\alias{loc_stats}
\title{Internal calculation of Lines-of-Code Statistics}
\usage{
loc_stats(path)
}
\arguments{
\item{path}{Directory to source code of package being analysed}
}
\value{
A list of statistics for each of three directories, 'R', 'src', and
'inst/include', each one having 5 statistics of total numbers of lines,
numbers of empty lines, total numbers of white spaces, total numbers of
characters, and indentation used in files in that directory.
}
\description{
Internal calculation of Lines-of-Code Statistics
}
\note{
NA values are returned for directories which do not exist.
}
\examples{
f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
# have to extract tarball to call function on source code:
path <- extract_tarball (f)
loc_stats (path)
}
\seealso{
Other stats:
\code{\link[=desc_stats]{desc_stats()}},
\code{\link[=pkgstats]{pkgstats()}},
\code{\link[=pkgstats_summary]{pkgstats_summary()}},
\code{\link[=rd_stats]{rd_stats()}}
}
\concept{stats}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rd-stats.R
\name{rd_stats}
\alias{rd_stats}
\title{Stats from '.Rd' files}
\usage{
rd_stats(path)
}
\arguments{
\item{path}{Directory to source code of package being analysed}
}
\value{
A \code{data.frame} of function names and numbers of parameters and lines
of documentation for each, along with mean and median numbers of characters
used to document each parameter.
}
\description{
Stats from '.Rd' files
}
\examples{
f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
# have to extract tarball to call function on source code:
path <- extract_tarball (f)
rd_stats (path)
}
\seealso{
Other stats:
\code{\link[=desc_stats]{desc_stats()}},
\code{\link[=loc_stats]{loc_stats()}},
\code{\link[=pkgstats]{pkgstats()}},
\code{\link[=pkgstats_summary]{pkgstats_summary()}}
}
\concept{stats}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cran-data-update.R
\name{pkgstats_update}
\alias{pkgstats_update}
\title{Update pkgstats` data on GitHub release}
\usage{
pkgstats_update(upload = TRUE)
}
\arguments{
\item{upload}{If \code{TRUE}, upload updated results to GitHub release.}
}
\value{
Local path to directory containing updated results.
}
\description{
This function is intended for internal rOpenSci use only. Usage by any
unauthorized users will error and have no effect unless run with \code{upload = FALSE}, in which case updated data will be created in the sub-directory
"pkgstats-results" of R's current temporary directory.
}
\seealso{
Other archive:
\code{\link[=dl_pkgstats_data]{dl_pkgstats_data()}},
\code{\link[=pkgstats_cran_current_from_full]{pkgstats_cran_current_from_full()}},
\code{\link[=pkgstats_fns_from_archive]{pkgstats_fns_from_archive()}},
\code{\link[=pkgstats_fns_update]{pkgstats_fns_update()}},
\code{\link[=pkgstats_from_archive]{pkgstats_from_archive()}}
}
\concept{archive}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgstats-function.R
\name{pkgstats}
\alias{pkgstats}
\title{Analyse statistics of one R package}
\usage{
pkgstats(path = ".")
}
\arguments{
\item{path}{Either a path to a local source repository, or a local '.tar.gz'
file, containing code for an R package.}
}
\value{
List of statistics and data on function call networks (or object
relationships in other languages). Includes the following components:
\enumerate{
\item loc: Summary of Lines-of-Code in all package directories
\item vignettes: Numbers of vignettes and "demo" files
\item data_stats: Statistics of numbers and sizes of package data files
\item desc: Summary of contents of 'DESCRIPTION' file
\item translations: List of translations into other (human) languages
(where provides)
\item objects: A \code{data.frame} of all functions in R, and all other
objects (functions, classes, structures, global variables, and more) in all
other languages
\item network: A \code{data.frame} of object references within and between all
languages; in R these are function calls, but may be more abstract in other
languages.
\item external_calls: A \code{data.frame} of all calls make to all functions
from all other R packages, including base and recommended as well as
contributed packages.
}
}
\description{
Analyse statistics of one R package
}
\examples{
# 'path' can be path to a package tarball:
f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
\dontshow{if (ctags_test (noerror = TRUE)) withAutoprint(\{ # examplesIf}
s <- pkgstats (f)
# or to a source directory:
path <- extract_tarball (f)
s <- pkgstats (path)
\dontshow{\}) # examplesIf}
}
\seealso{
Other stats:
\code{\link[=desc_stats]{desc_stats()}},
\code{\link[=loc_stats]{loc_stats()}},
\code{\link[=pkgstats_summary]{pkgstats_summary()}},
\code{\link[=rd_stats]{rd_stats()}}
}
\concept{stats}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function-names.R
\name{pkgstats_fn_names}
\alias{pkgstats_fn_names}
\title{Extract names of all functions for one R package}
\usage{
pkgstats_fn_names(path)
}
\arguments{
\item{path}{Either a path to a local source repository, or a local '.tar.gz'
file, containing code for an R package.}
}
\value{
A \code{data.frame} with three columns:
\itemize{
\item package: Name of package
\item version: Package version
\item fn_name: Name of function
}
}
\description{
Extract names of all functions for one R package
}
\examples{
# 'path' can be path to a package tarball:
f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
path <- extract_tarball (f)
s <- pkgstats_fn_names (path)
}
\seealso{
Other misc:
\code{\link[=extract_tarball]{extract_tarball()}}
}
\concept{misc}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cran-data-archive.R
\name{pkgstats_from_archive}
\alias{pkgstats_from_archive}
\title{Trawl a local CRAN archive and extract statistics from all packages}
\usage{
pkgstats_from_archive(
  path,
  archive = TRUE,
  prev_results = NULL,
  results_file = NULL,
  chunk_size = 1000L,
  num_cores = 1L,
  save_full = FALSE,
  save_ex_calls = FALSE,
  results_path = fs::path_temp()
)
}
\arguments{
\item{path}{Path to local archive of R packages, either as source
directories, or '.tar.gz' files such as in a CRAN mirror.}

\item{archive}{If \code{TRUE}, extract statistics for all packages in the
\verb{/Archive} sub-directory, otherwise only statistics for main directory (that
is, current packages only).}

\item{prev_results}{Result of previous call to this function, if available.
Submitting previous results will ensure that only newer packages not present
in previous result will be analysed, with new results simply appended to
previous results. This parameter can also specify a file to be read with
\code{readRDS()}.}

\item{results_file}{Can be used to specify the name or full path of a \code{.Rds}
file to which results should be saved once they have been generated. The
'.Rds' extension will be automatically appended, and any other extensions
will be ignored.}

\item{chunk_size}{Divide large archive trawl into chunks of this size, and
save intermediate results to local files. These intermediate files can be
combined to generate a single \code{prev_results} file, to enable jobs to be
stopped and re-started without having to recalculate all results. These files
will be named \code{pkgstats-results-N.Rds}, where "N" incrementally numbers each
file.}

\item{num_cores}{Number of machine cores to use in parallel, defaulting to
single-core processing.}

\item{save_full}{If \code{TRUE}, full \link{pkgstats} results are saved for each
package to files in \code{results_path}.}

\item{save_ex_calls}{If \code{TRUE}, the results of the \code{external_calls} component
are saved for each package to files in \code{results_path} (only if \code{save_full = FALSE}).}

\item{results_path}{Path to save intermediate files generated by the
\code{chunk_size} parameter described above.}
}
\value{
A \code{data.frame} object with one row for each package containing
summary statistics generated from the \link{pkgstats_summary} function.
}
\description{
Trawl a local CRAN archive and extract statistics from all packages
}
\examples{
# Create fake archive directory with single tarball:
f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
tarball <- basename (f)

archive_path <- file.path (tempdir (), "archive")
if (!dir.exists (archive_path)) {
    dir.create (archive_path)
}
path <- file.path (archive_path, tarball)
file.copy (f, path)
tarball_path <- file.path (archive_path, "tarballs")
dir.create (tarball_path, recursive = TRUE)
file.copy (path, file.path (tarball_path, tarball))
out <- pkgstats_from_archive (tarball_path)
}
\seealso{
Other archive:
\code{\link[=dl_pkgstats_data]{dl_pkgstats_data()}},
\code{\link[=pkgstats_cran_current_from_full]{pkgstats_cran_current_from_full()}},
\code{\link[=pkgstats_fns_from_archive]{pkgstats_fns_from_archive()}},
\code{\link[=pkgstats_fns_update]{pkgstats_fns_update()}},
\code{\link[=pkgstats_update]{pkgstats_update()}}
}
\concept{archive}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgstats-package.R
\docType{package}
\name{pkgstats-package}
\alias{pkgstats-package}
\title{pkgstats: Metrics of R Packages}
\description{
Static code analyses for R packages using the external code-tagging libraries 'ctags' and 'gtags'. Static analyses enable packages to be analysed very quickly, generally a couple of seconds at most. The package also provides access to a database generating by applying the main function to the full 'CRAN' archive, enabling the statistical properties of any package to be compared with all other 'CRAN' packages.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://docs.ropensci.org/pkgstats/}
  \item \url{https://github.com/ropensci-review-tools/pkgstats}
  \item Report bugs at \url{https://github.com/ropensci-review-tools/pkgstats/issues}
}

}
\author{
\strong{Maintainer}: Mark Padgham \email{mark.padgham@email.com} (\href{https://orcid.org/0000-0003-2172-5265}{ORCID})

Authors:
\itemize{
  \item Mark Padgham \email{mark.padgham@email.com} (\href{https://orcid.org/0000-0003-2172-5265}{ORCID})
}

Other contributors:
\itemize{
  \item Michael Sumner (\href{https://orcid.org/0000-0002-2471-7511}{ORCID}) [contributor]
  \item Jeffrey Hollister (\href{https://orcid.org/0000-0002-9254-9740}{ORCID}) [contributor]
  \item Egor Kotov (\href{https://orcid.org/0000-0001-6690-5345}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract-tarball.R
\name{extract_tarball}
\alias{extract_tarball}
\title{Extract tarball of a package into temp directory and return path to extracted
package}
\usage{
extract_tarball(tarball, exdir = fs::path_temp())
}
\arguments{
\item{tarball}{Full path to local tarball of an R package.}

\item{exdir}{Directory into which tarballs are to be extracted.}
}
\value{
Path to extracted version of package (in \code{tempdir()}).
}
\description{
Extract tarball of a package into temp directory and return path to extracted
package
}
\examples{
f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
path <- extract_tarball (f)
}
\seealso{
Other misc:
\code{\link[=pkgstats_fn_names]{pkgstats_fn_names()}}
}
\concept{misc}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cran-data-fn-names.R
\name{pkgstats_fns_update}
\alias{pkgstats_fns_update}
\title{Update function names data from previous data and newly updated CRAN
packages only.}
\usage{
pkgstats_fns_update(
  prev_results = NULL,
  results_file = NULL,
  chunk_size = 1000L,
  num_cores = 1L,
  results_path = tempdir()
)
}
\arguments{
\item{prev_results}{Result of previous call to this function, if available.
Submitting previous results will ensure that only newer packages not present
in previous result will be analysed, with new results simply appended to
previous results. This parameter can also specify a file to be read with
\code{readRDS()}.}

\item{results_file}{Can be used to specify the name or full path of a \code{.Rds}
file to which results should be saved once they have been generated. The
'.Rds' extension will be automatically appended, and any other extensions
will be ignored.}

\item{chunk_size}{Divide large archive trawl into chunks of this size, and
save intermediate results to local files. These intermediate files can be
combined to generate a single \code{prev_results} file, to enable jobs to be
stopped and re-started without having to recalculate all results. These files
will be named \code{pkgstats-results-N.Rds}, where "N" incrementally numbers each
file.}

\item{num_cores}{Number of machine cores to use in parallel, defaulting to
single-core processing.}

\item{results_path}{Path to save intermediate files generated by the
\code{chunk_size} parameter described above.}
}
\value{
A \code{data.frame} object with one row for each function in each package
and the following columns:
\itemize{
\item Package name
\item Package version
\item Function name
}
}
\description{
Update function names data from previous data and newly updated CRAN
packages only.
}
\seealso{
Other archive:
\code{\link[=dl_pkgstats_data]{dl_pkgstats_data()}},
\code{\link[=pkgstats_cran_current_from_full]{pkgstats_cran_current_from_full()}},
\code{\link[=pkgstats_fns_from_archive]{pkgstats_fns_from_archive()}},
\code{\link[=pkgstats_from_archive]{pkgstats_from_archive()}},
\code{\link[=pkgstats_update]{pkgstats_update()}}
}
\concept{archive}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ctags-install.R
\name{ctags_install}
\alias{ctags_install}
\title{Install 'ctags' from a clone of the 'git' repository}
\usage{
ctags_install(bin_dir = NULL, sudo = TRUE)
}
\arguments{
\item{bin_dir}{This parameter only has an effect on *nix-type operating
systems (such as Linux), on which it's a prefix to pass to the
\code{autoconf} configure command defining location to install the binary, with
default of \verb{/usr/local}.}

\item{sudo}{Set to \code{FALSE} if \code{sudo} is not available, in which case a
value for \code{bin_dir} will also have to be explicitly specified, and be a
location where a binary is able to be installed without \code{sudo} privileges.}
}
\value{
Nothing; the function will fail if installation fails, otherwise
returns nothing.
}
\description{
'ctags' is installed with this package on both Windows and macOS systems;
this is an additional function to install from source on Unix systems.
}
\examples{
\dontrun{
ctags_install (bin_dir = "/usr/local") # default Linux location.
}
}
\seealso{
Other tags:
\code{\link[=ctags_test]{ctags_test()}},
\code{\link[=tags_data]{tags_data()}}
}
\concept{tags}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgstats-data-download.R
\name{dl_pkgstats_data}
\alias{dl_pkgstats_data}
\title{Download latest version of 'pkgstats' data}
\usage{
dl_pkgstats_data(current = TRUE, path = tempdir(), quiet = FALSE)
}
\arguments{
\item{current}{If 'FALSE', download data for all CRAN packages ever released,
otherwise (default) download data only for current CRAN packages.}

\item{path}{Local path to download file.}

\item{quiet}{If \code{FALSE}, display progress information on screen.}
}
\value{
(Invisibly) A \code{data.frame} of \code{pkgstats} results, one row for each
package.
}
\description{
Download latest version of 'pkgstats' data
}
\seealso{
Other archive:
\code{\link[=pkgstats_cran_current_from_full]{pkgstats_cran_current_from_full()}},
\code{\link[=pkgstats_fns_from_archive]{pkgstats_fns_from_archive()}},
\code{\link[=pkgstats_fns_update]{pkgstats_fns_update()}},
\code{\link[=pkgstats_from_archive]{pkgstats_from_archive()}},
\code{\link[=pkgstats_update]{pkgstats_update()}}
}
\concept{archive}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag-data.R
\name{tags_data}
\alias{tags_data}
\title{use ctags and gtags to parse call data}
\usage{
tags_data(path, has_tabs = NULL, pkg_name = NULL)
}
\arguments{
\item{path}{Path to local repository}

\item{has_tabs}{A logical flag indicating whether or not the code contains
any tab characters. This can be determined from \link{loc_stats}, which has a
\code{tabs} column. If not given, that value will be extracted from internally
calling that function.}

\item{pkg_name}{Only used for external_call_network, to label
package-internal calls.}
}
\value{
A list of three items:
\itemize{
\item "network" A \code{data.frame} of relationships between objects, generally as
calls between functions in R, but other kinds of relationships in other
source languages. This is effectively an edge-based network representation,
and the data frame also include network metrics for each edge, calculated
through representing the network in both directed (suffix "_dir") and
undirected (suffix "_undir") forms.
\item "objects" A \code{data.frame} of statistics on each object (generally
functions in R, and other kinds of objects in other source languages),
including the kind of object, the language, numbers of lines-of-code,
parameters, and lines of documentation, and a binary flag indicating whether
or not R functions accept "three-dots" parameters (\code{...}).
\item "external_calls" A \code{data.frame} of every call from within every R
function to any external R package, including base and recommended packages.
The location of each calls is recorded, along with the external function and
package being called.
}
}
\description{
use ctags and gtags to parse call data
}
\examples{
f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
# have to extract tarball to call function on source code:
path <- extract_tarball (f)
\dontshow{if (ctags_test (noerror = TRUE)) withAutoprint(\{ # examplesIf}
tags <- tags_data (path)
\dontshow{\}) # examplesIf}
}
\seealso{
Other tags:
\code{\link[=ctags_install]{ctags_install()}},
\code{\link[=ctags_test]{ctags_test()}}
}
\concept{tags}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cran-data-fn-names.R
\name{pkgstats_fns_from_archive}
\alias{pkgstats_fns_from_archive}
\title{Trawl a local CRAN archive to extract function names only from all packages}
\usage{
pkgstats_fns_from_archive(
  path,
  archive = FALSE,
  prev_results = NULL,
  results_file = NULL,
  chunk_size = 1000L,
  num_cores = 1L,
  results_path = fs::path_temp()
)
}
\arguments{
\item{path}{Path to local archive of R packages, either as source
directories, or '.tar.gz' files such as in a CRAN mirror.}

\item{archive}{If \code{TRUE}, extract statistics for all packages in the
\verb{/Archive} sub-directory, otherwise only statistics for main directory (that
is, current packages only).}

\item{prev_results}{Result of previous call to this function, if available.
Submitting previous results will ensure that only newer packages not present
in previous result will be analysed, with new results simply appended to
previous results. This parameter can also specify a file to be read with
\code{readRDS()}.}

\item{results_file}{Can be used to specify the name or full path of a \code{.Rds}
file to which results should be saved once they have been generated. The
'.Rds' extension will be automatically appended, and any other extensions
will be ignored.}

\item{chunk_size}{Divide large archive trawl into chunks of this size, and
save intermediate results to local files. These intermediate files can be
combined to generate a single \code{prev_results} file, to enable jobs to be
stopped and re-started without having to recalculate all results. These files
will be named \code{pkgstats-results-N.Rds}, where "N" incrementally numbers each
file.}

\item{num_cores}{Number of machine cores to use in parallel, defaulting to
single-core processing.}

\item{results_path}{Path to save intermediate files generated by the
\code{chunk_size} parameter described above.}
}
\value{
A \code{data.frame} object with one row for each function in each package
and the following columns:
\itemize{
\item Package name
\item Package version
\item Function name
}
}
\description{
Trawl a local CRAN archive to extract function names only from all packages
}
\seealso{
Other archive:
\code{\link[=dl_pkgstats_data]{dl_pkgstats_data()}},
\code{\link[=pkgstats_cran_current_from_full]{pkgstats_cran_current_from_full()}},
\code{\link[=pkgstats_fns_update]{pkgstats_fns_update()}},
\code{\link[=pkgstats_from_archive]{pkgstats_from_archive()}},
\code{\link[=pkgstats_update]{pkgstats_update()}}
}
\concept{archive}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/desc-stats.R
\name{desc_stats}
\alias{desc_stats}
\title{Statistics from DESCRIPTION files}
\usage{
desc_stats(path)
}
\arguments{
\item{path}{Directory to source code of package being analysed}
}
\value{
A \code{data.frame} with one row and 16 columns extracting various
information from the 'DESCRIPTION' file, include websites, tallies of
different kinds of authors and contributors, and package dependencies.
}
\description{
Statistics from DESCRIPTION files
}
\examples{
f <- system.file ("extdata", "pkgstats_9.9.tar.gz", package = "pkgstats")
# have to extract tarball to call function on source code:
path <- extract_tarball (f)
desc_stats (path)
}
\seealso{
Other stats:
\code{\link[=loc_stats]{loc_stats()}},
\code{\link[=pkgstats]{pkgstats()}},
\code{\link[=pkgstats_summary]{pkgstats_summary()}},
\code{\link[=rd_stats]{rd_stats()}}
}
\concept{stats}
